## Khm

Khm is a 2D framework with Kha and Phaser backends, oriented to provide fast and simple work with tree-based rendering. Main features:
- Display list implementation, with simple way to use classic g2 Kha rendering inside of game objects
- Tree-based events system, unified Pointer events and nice object dragging
- Scenes, containers, sprites, particles, tweens, ttf and canvas texts, animations and more
- Generate texture atlases in cli, just by moving textures between named folders, preview and use frames with `Res.atlasName.fileName` in IDE
- Inspect elements on screen with integrated debug tools, click on them to open game object declaration line in VSCode
- Tested between both backends on many mini-games to prevent unwanted behavior and stay stable
- Kha: watch mode with live-reload, multitexturing, and more exclusive features in future

### Setup:
- Install Haxe 4.3+
- Install NodeJS

```
npm i light-server -g
npm i free-tex-packer-core -g --legacy-peer-deps
npm i ttf2woff -g
haxelib install all
haxelib install hxnodejs
haxelib git haxe-phaser https://github.com/RblSb/haxe-phaser.git
```

### Assets folder and atlas generator

Use `assets/atlases` folder for original textures, folder structure for atlas generation is:
- /assets/atlases/png-files -> will be generated as `res/atlases/res.png` (and `res.json`)
- /assets/atlases/popup/png-files -> will be generated as `res/atlases/res_popup.png`
- /assets/atlases/otherAtlasName/png-files -> will be generated as `res/atlases/res_otherAtlasName.png`

Use `Run Task > Generate atlases` to generate all atlases and update `build-resources.hxml` automatically.
Generator will cache folder metadata and will skip folders if they are not changed.
You can force atlas generation if `res/res_atlasName.png` will be not found.

Separate (mostly `jpg`) images should be inside `res/` and have `bg` prefix in names, to be added to `build-resources.hxml` on atlas generation.

For temp files you can use `/assets/atlases/_someName/` folder names, generator will skip folders with underscore prefix (and subfolders too).
Haxe compilation server will be restarted after successful atlas generation to update fields completion, and `build-resources.hxml` will be updated, so you only need to recompile project.

### Configuration

You can create and customize `/lazyr.json` file with these properties:

```json
{
	"texpacker": {
		"tinifyKey": "your_key",
		"padding": 1,
		"trimMode": "trim"
	}
}
```

`tinifyKey` can be requested [here](https://tinypng.com/developers).
`trimMode` can be "trim" or "crop".

You can control `trimMode` behaviour with `.trimMode-crop` / `.trimMode-trim` empty file in any atlas folder.

### Diving in

To start working with khm, it is recommended to use VSCode with [this project template](https://github.com/RblSb/khmTemplate). First, you need to call from main:

```haxe
App.init({
	onInit: () -> new Game()
});
```

It will init all event listeners and rendering loop. Now you need to create `Game` scene.

```haxe
import khm.Res.res;
import khm.Scene;

class Game extends Scene {
	public function new() {
		super();
		final logo = addSprite(res.logo);
	}
}
```

`addSprite` mostly is a shortcut to:
```haxe
final logo = new Sprite(res.logo);
add(logo); // adds object to this scene/container
```
`res.logo` is `{texture:String, frame:String}` structure, that can be generated dynamically, for example, if you want to iterate some frames with counter.

To have `res.logo` field in `khm.Res`, you need to generate `res` atlas. Move your `logo.png` texture to `assets/` folder, open command palette in VSCode and select `Run Task > Generate atlases`.

You can also add things to containers, to move/scale/rotate them relatively:
```haxe
final logoC = addContainer();
logoC.addSprite(res.logo);
// center content inside and `setSize(...)` for container
// based on total children bounds
logoC.shrink();
logoC.x = 50;
logoC.y = 50;
logoC.scale = 2; // shortcut of changing scaleX/scaleY
logoC.angle = 45;
```

Instead of `shrink()`, Container can be created with a specific size with `addContainer({size: {width: 50, height: 50}})`. This will make container local space in such bounds:
`x: -25, y: -25, width: 50, height: 50`.

With container top left corner on `-25,-25`. If you want to use `0,0` as top-left cords, you can set container `originX/originY` to `0`, same approach for sprites.

You can get container local space top left position with `container.topLeftLocal()`, and most flat objects, like sprites, also have `sprite.topLeft()` method, instead of calculating cords like `sprite.x - sprite.displayWidth * sprite.originX`. You can also get global screen space of object rendering with `sprite.getGlobalPosition()`.

Scene is a superset of Container, but with these differences:
- `show()`/`hide()` methods to add scenes to screen, and some additional fields
- Origin point is `0,0` by default, so rendering starts from top left angle
- Scene `width`/`height` always resizes to `App.w` / `App.h` on every frame by default. If you want to change that, you can override `onResize` Scene function or remove and set new `Resize` callback:
```haxe
this.off(Resize);
this.on(Resize, e -> {
	width = 50;
	height = 50;
});
```
Be aware than `on` and `off` is static extension functions, so you need to call them with `this.on` (or `obj.on` in other cases).

Every game object implements `IGameObject` interface with basic transformation functions and `parent:IContainer` field.

Containers implement `IContainer`, which has `add/addAt/remove/removeAll` methods and `children` array. Everything rendered by the tree has `parent`, except for scenes, they added to `App.scenes` array on `show()` call. To combine tree rendering with g2, you can override `render` function in any `IGameObject`, for example in Scene:
```haxe
#if kha
override function render(g:kha.g2.Graphics) {
	super(g); // render scene tree first
	g.color = Red;
	// render all pressed pointers (mouse/touch)
	for (p in App.pointers)
		if (p.isActive) g.fillRect(p.x - 2, p.y - 2, 4, 4)
}
#end
```

To add tweens, you can use `new Tween({targets: arrOrObj, x: 123})` or another handy static extension:
```haxe
logo.addTween({ // duration is 1000ms by default
	scale: 1.2,
	yoyo: true, // after scaling logo sprite to 1.2, scale back to 1.0
	repeat: -1 // repeat endless times
});
```
If you set `duration: 0` in tween config, tween will be called immediately, which is nice to skip some animations.

Khm has a lot of static extension methods in `khm.tools`, to keep objects in memory small, and they are used with `import.hx` by default. For example, there is `arr.last()`, `obj.log()` and `obj.findParent(parent -> true)`.

Most events, except `Resize`, called from top child to bottom scene. For example, if child and scene have `PointerDown` events, click on child will stop passing event to scene. But if child has `visible = false`, `alpha = 0`, `isInteractive = false` or `event.propagate = true` in listener, then scene event will be fired. You can enable propagation by default with `Event.propagateEvents = true`.

You can also debug event listeners code positions / event order with `Event.debugEventKind = PointerDown` and checkout console.
