package khm;

import haxe.Timer;
import khm.Event.EventKind;
import khm.Event;
import khm.Graphics;
import khm.Pointer.PointerButtonKind;
import khm.ResItem;
import khm.Types.Point;
import khm.audio.Audio;
import khm.math.FastMatrix3;
import khm.utils.Observer;

using khm.tools.MathTools;
using khm.tools.PhaserTools;

#if kha
import kha.Assets;
import kha.Canvas;
import kha.Font;
import kha.Framebuffer;
import kha.Scheduler;
import kha.System;
import kha.SystemImpl;
import kha.graphics2.Graphics;
import kha.graphics2.ImageScaleQuality;
import kha.input.Gamepad;
import kha.input.KeyCode;
import kha.input.Keyboard;
import kha.input.Mouse;
import kha.input.Surface;
#end
#if js
import js.Browser.document;
import js.Browser.window;
#end

#if phaser
private typedef Framebuffer = {
	g2:Graphics
}

private typedef Font = String;
#end

@:forward(keys)
abstract KeyMap(Map<KeyCode, Bool>) {
	public inline function new() {
		this = [];
	}

	@:arrayAccess
	public inline function get(key:KeyCode):Bool {
		return this[key] == true;
	}

	@:arrayAccess public inline function set(k:KeyCode, v:Bool):Bool {
		this.set(k, v);
		return v;
	}

	public inline function remove(k:KeyCode):Bool {
		return this.remove(k);
	}

	public inline function toMap():Map<KeyCode, Bool> {
		return this;
	}
}

@:structInit
class AppSets {
	/** Initial scene that will called with `scene.show()` automatically after initialization. **/
	public var onStart:() -> Scene;
	public var isTouch:Null<Bool> = null;
	// public var scene:Scene;
}

class App {
	#if kha
	public static var _scene = {
		sound: {
			pauseOnBlur: false
		}
	};
	#elseif phaser
	public static var phaserGame:phaser.Game;
	public static var _scene:phaser.Scene;
	// public static var _tempScene:phaser.Scene;
	public static var canvasElement:js.html.CanvasElement;
	#end

	public static var fps = 60;
	public static var w(default, null) = 0; // for resize event
	public static var h(default, null) = 0;
	public static var isTouch(default, null) = false;
	public static var deltaTime = 1 / fps;
	public static var gamepadDeadZone = 0.1;
	public static var debugBounds = false;
	public static var debugBoundsMode = 0;
	public static var debugInputOrder = false;
	public static var debugFont:Font = null;
	public static var defaultScaleQuality:ImageScaleQuality = High;
	public static var bgColor(default, set):Null<Color> = null;

	inline static function set_bgColor(bgColor:Color):Color {
		App.bgColor = bgColor;
		#if phaser
		App._scene.cameras.main.setBackgroundColor(bgColor.toRgb());
		#end
		return bgColor;
	}

	static var taskId = -1;
	static var isInited = false;

	#if kha
	static var gamepads:Array<Bool> = [for (i in 0...4) false];
	static final metrics = new Metrics();
	#end
	public static final keys = new KeyMap();
	public static final pointers:Array<Pointer> = [];

	public static final scenes:Array<Scene> = [];
	public static var audio:Audio;
	#if js
	static var activeJsEvent:js.html.Event;
	#end
	static var logListeners:Array<(args:Array<Dynamic>) -> Void> = [];

	/** Init systems and show first scene from `config.onStart` callback **/
	public static function init(config:AppSets):Void {
		if (isInited) return;
		if (config.onStart == null)
			throw "onStart callback is null, but should return new Scene";
		// if (config.scene == null) throw "scene cannot be null";

		#if kha
		App.init2(config);
		#elseif phaser
		function getSize():{width:Int, height:Int} {
			final getSize:() -> Any = (window : Dynamic).getSize;
			if (getSize != null) return getSize();
			return App.getWindowSize();
		}
		final size = getSize();
		final config:phaser.types.core.GameConfig = {
			#if canvas_only
			type: untyped Phaser.CANVAS,
			#end
			parent: "canvas-game",
			scale: {
				mode: untyped Phaser.Scale.NONE,
				width: size.width * window.devicePixelRatio,
				height: size.height * window.devicePixelRatio,
				autoRound: true
			},
			roundPixels: false,
			autoFocus: true,
			// autoMobilePipeline: true,
			canvasStyle: "display: block; width: 100%; height: 100%;",
			// resolution: untyped window.devicePixelRatio || 1,
			#if debug
			render: cast {
				clearBeforeRender: false,
				powerPreference: 'low-power',
				defaultPipeline: "MobilePipeline"
			},
			#end
			// callbacks: {
			// 	postBoot: () -> {
			// 		trace("postBoot");
			// 		Main.onWindowResize();
			// 	}
			// },
			scene: new GlobalScene(_scene -> {
				App._scene = _scene;
				App.init2(config);
			})
		}
		phaserGame = new phaser.Game(config);
		#end
		Observer.connect();
	}

	static function init2(config:AppSets):Void {
		#if kha_html5
		isTouch = js.Syntax.code('"ontouchstart" in window');
		#elseif (kha_android || kha_ios)
		isTouch = true;
		#end
		if (config.isTouch != null) isTouch = config.isTouch;

		updateScreenSize();

		audio = new Audio();

		#if kha
		final firstFont = Assets.fonts.names[0];
		if (debugFont == null && firstFont != null) {
			Assets.loadFont(firstFont, font -> {
				if (debugFont != null) return;
				debugFont = font;
			});
		}
		Gamepad.notifyOnConnect(id -> {
			if (gamepads[id]) return;
			gamepads[id] = true;
			trace('Gamepad $id connected');
			Gamepad.get(id).notify(_onGamepadAxis, _onGamepadButton);
		}, id -> {
			if (!gamepads[id]) return;
			gamepads[id] = false;
			trace('Gamepad $id disconnected');
			Gamepad.get(id).remove(_onGamepadAxis, _onGamepadButton);
		});
		for (id in 0...4) {
			if (gamepads[id]) continue;
			final gamepad = Gamepad.get(id);
			if (gamepad == null) continue;
			gamepads[id] = true;
			gamepad.notify(_onGamepadAxis, _onGamepadButton);
		}

		taskId = Scheduler.addTimeTask(_onUpdate, 0, 1 / fps);
		lastTime = Scheduler.time();
		System.notifyOnFrames(_onRender);

		if (Keyboard.get() != null) Keyboard.get().notify(_onKeyDown, _onKeyUp, _onKeyPress);
		Keyboard.disableSystemInterventions(Custom(code -> {
			return switch code {
				case Escape, Tab, Backspace, Space: true;
				case Up, Down, Left, Right: true;
				case _: false;
			}
		}));

		if (isTouch && Surface.get() != null) {
			Surface.get().notify(_onTouchDown, _onTouchUp, _onTouchMove);
		} else if (Mouse.get() != null) {
			Mouse.get().notify(
				(button, x, y) -> _onMouseDown(cast button, x, y),
				(button, x, y) -> _onMouseUp(cast button, x, y),
				_onMouseMove,
				onMouseWheel, onMouseLeave
			);
		}
		#if kha_html5
		SystemImpl.khanvas.focus();
		#end
		#elseif phaser
		final debugUpdateSlowly = false;
		var skipFrames = 0;

		_scene.events.on("update", (time:Float, delta:Float) -> {
			if (debugUpdateSlowly) {
				skipFrames++;
				if (skipFrames % 100 != 0) return;
				skipFrames = 0;
			}
			App.deltaTime = delta;
			// _onUpdate();
		});
		_scene.time.addEvent({
			delay: 1000 / 60, // 1 second
			callback: () -> {
				_onUpdate();
			},
			loop: true // Loop the timer
		});
		final framebuffers:Array<Framebuffer> = [{
			g2: new Graphics()
		}];
		_scene.events.on("render", () -> {
			_onRender(framebuffers);
		});
		_scene.events.on("postupdate", () -> {
			onBeforeRender();
		});

		_scene.input.addPointer(9);
		_scene.input.on("pointerdown", (p:phaser.input.Pointer) -> {
			var btn:PointerButtonKind = switch p.button {
				case 2: Right;
				case 0, _: Left;
			}
			final e:js.html.MouseEvent = p.event;
			if (e != null) {
				// when keyup happens with unfocused window
				keys.set(Shift, e.shiftKey);
				keys.set(Control, e.ctrlKey);
				keys.set(Alt, e.altKey);
				keys.set(Meta, e.metaKey);
			}
			_onMouseDown(btn, Std.int(p.x), Std.int(p.y));
		});
		_scene.input.on("pointermove", (p:phaser.input.Pointer) -> {
			_onMouseMove(Std.int(p.x), Std.int(p.y), 0, 0);
		});
		_scene.input.on("pointerup", (p:phaser.input.Pointer) -> {
			var btn:PointerButtonKind = switch p.button {
				case 2: Right;
				case 0, _: Left;
			}
			_onMouseUp(btn, Std.int(p.x), Std.int(p.y));
		});
		_scene.input.on("wheel", (pointer, obj, deltaX, deltaY, deltaZ) -> {
			if (deltaY == 0) return;
			final value = deltaY < 0 ? -1 : 1;
			onMouseWheel(value);
		});

		function fixedKeyCode(event:js.html.KeyboardEvent):KeyCode {
			return switch (event.keyCode) {
				case 91, 93: Meta; // left/right in Chrome
				case 186: Semicolon;
				case 187: Equals;
				case 189: HyphenMinus;
				default:
					cast event.keyCode;
			}
		}
		_scene.input.keyboard.on('keydown', (event:js.html.KeyboardEvent) -> {
			event.stopPropagation();
			if (event.repeat) {
				event.preventDefault();
				return;
			}
			activeJsEvent = event;
			defaultKeyBlock(event);
			_onKeyDown(fixedKeyCode(event));
		});
		_scene.input.keyboard.on('keyup', (event:js.html.KeyboardEvent) -> {
			event.stopPropagation();
			if (event.repeat) {
				event.preventDefault();
				return;
			}
			activeJsEvent = event;
			defaultKeyBlock(event);
			_onKeyUp(fixedKeyCode(event));
		});
		js.Browser.window.addEventListener("keypress", (event:js.html.KeyboardEvent) -> {
			if (event.which == 0) return;
			activeJsEvent = event;
			defaultKeyBlock(event);
			final char = String.fromCharCode(event.which);
			_onKeyPress(char);
		});

		_scene.scale.on("resize", (gameSize, baseSize, displaySize, prevW, prevH) -> {
			_onResize();
		});

		#if debug
		_scene.input.mouse.disableContextMenu();
		#end
		#end
		final origTrace = haxe.Log.trace;

		function extractPosArg(pos:haxe.PosInfos):haxe.PosInfos {
			final args = pos.customParams ?? return pos;
			final item = args[args.length - 1] ?? return pos;
			if (item.fileName != null && item.lineNumber != null) {
				args.remove(item);
				return item;
			}
			return pos;
		}

		var inLogCallback = false;
		haxe.Log.trace = (msg:Dynamic, ?infos:haxe.PosInfos) -> {
			final pos = extractPosArg(infos);
			final fileData = '${pos.fileName}:${pos.lineNumber}';
			var args:Array<Dynamic> = [fileData, msg];
			if (infos.customParams != null) args = args.concat(infos.customParams);
			#if js
			js.Browser.window.console.log(
				...haxe.Rest.of(args)
			);
			#else
			pos.customParams = infos.customParams;
			origTrace(msg, pos);
			#end
			if (inLogCallback) return;
			inLogCallback = true;
			for (callback in logListeners) callback(args);
			inLogCallback = false;
		}

		isInited = true;

		final scene = config.onStart();
		scene.show();
		Lazyr.sendAnalytics("Start");
		function sendOnExit():Void {
			Lazyr.sendAnalytics("Time Spent");
			Lazyr.sendAnalytics("All Click", totalClicks);
			Lazyr.sendAnalytics("Exit");
		}
		#if js
		window.onblur = () -> {
			sendOnExit();
		}
		document.addEventListener("visibilitychange", () -> {
			if (document.visibilityState == VISIBLE) return;
			sendOnExit();
		});
		window.addEventListener("beforeunload", () -> {
			sendOnExit();
		});
		#end

		#if phaser
		_scene.add.existing(framebuffers[0].g2.phaser);
		framebuffers[0].g2.phaser.setDepth(1);
		#end
	}

	public static function showScene(scene:Scene):Void {
		scenes.push(scene);
		final e:ScenesChangeEvent = {
			kind: ScenesChange,
			target: null,
			scene: scene,
			type: Added
		};
		Event.emit(e);
		#if phaser
		_scene.add.existing(scene.phaser);
		#end
		Event.sortObjectEvents(scene);
		scene.children.flatIter(obj -> {
			Event.sortObjectEvents(obj);
		});
	}

	public static function hideScene(scene:Scene):Void {
		scenes.remove(scene);
		final e:ScenesChangeEvent = {
			kind: ScenesChange,
			target: null,
			scene: scene,
			type: Removed
		};
		Event.emit(e);
		#if phaser
		_scene.scene.remove(scene.phaser);
		// scene.phaser.destroy();
		#end
	}

	public static inline function isAttachedScene(obj:IGameObject):Bool {
		return obj.type == SceneType && (cast obj : Scene).isAttached;
	}

	public static function setTouchMode(state:Bool):Void {
		isTouch = state;
	}

	/** Add custom callback with `trace` value/pos, that will be fired after `trace` execution in code. **/
	public static function onLog(callback:(args:Array<Dynamic>) -> Void):Void {
		logListeners.push(callback);
	}

	/** Remove custom `trace` callback, added with `onLog`. **/
	public static function offLog(callback:(args:Array<Dynamic>) -> Void):Void {
		logListeners.remove(callback);
	}

	#if phaser
	static function defaultKeyBlock(e:js.html.KeyboardEvent):Void {
		final keyCode:KeyCode = cast e.keyCode;
		switch keyCode {
			case Escape:
				e.preventDefault();
			case _:
		}
	}
	#end

	public static function preventDefault():Void {
		#if js
		#if kha
		SystemImpl.activeKeyEvent?.preventDefault();
		#end
		activeJsEvent?.preventDefault();
		#end
	}

	#if js
	public static function getWindowSize():{width:Int, height:Int} {
		var w = document.documentElement.clientWidth;
		var h = document.documentElement.clientHeight;
		if (w == 0 || h == 0) {
			w = window.innerWidth;
			h = window.innerHeight;
		}
		return {width: w, height: h};
	}
	#end

	static function _onUpdate():Void {
		#if kha
		final updateTime = Scheduler.realTime();
		if (System.windowWidth() != w || System.windowHeight() != h) _onResize();
		#end

		onUpdate();

		#if kha
		metrics.addUpdateTime(Scheduler.realTime() - updateTime);
		metrics.update();
		#end
	}

	// remove after `--library hxnodejs` will be removed from legacy hxmls
	public static function stamp():Float {
		#if (hxnodejs && !utest)
		#error "remove `--library hxnodejs` from build.hxml"
		#end
		#if js
		return js.lib.Date.now();
		#else
		return haxe.Timer.stamp();
		#end
	}

	static function _onResize():Void {
		updateScreenSize();
		onResize();
	}

	static function updateScreenSize() {
		#if kha
		w = System.windowWidth();
		h = System.windowHeight();
		#elseif phaser
		w = Std.int(_scene.cameras.main.width);
		h = Std.int(_scene.cameras.main.height);
		#end
	}

	static var lastTime = 0.0;

	static function _onRender(framebuffers:Array<Framebuffer>):Void {
		#if kha
		final now = Scheduler.realTime() * 1000;
		deltaTime = now - lastTime;
		if (deltaTime > 1000) deltaTime = 1000 / 60;
		lastTime = now;
		final renderTime = Scheduler.realTime();
		onBeforeRender();
		#end

		final frame = framebuffers[0];
		final g = frame.g2;
		g.begin(true, bgColor);
		g.opacity = 1;
		#if kha
		g.imageScaleQuality = defaultScaleQuality;
		for (obj in scenes) {
			obj.render(g);
		}
		#end
		if (debugBounds || debugBoundsMode > 0) {
			for (scene in scenes) {
				drawAllContainerBounds(g, scene);
			}
		}
		// trace(countObjects(), countPhaserObjects());

		if (debugInputOrder) {
			final drawedObjects = [];
			for (groupId => group in Event.events) {
				final groupId:EventKind<Any> = cast groupId;
				if (groupId != Drag) continue;
				for (i => event in group) {
					if (drawedObjects.contains(event.object)) continue;
					drawInputOrderNumbers(g, event.object, groupId, i);
					drawedObjects.push(event.object);
				}
			}
			final firstDrag = Event.dragObjects.inlineFind(item -> item != null);
			if (firstDrag != null) {
				g.drawString("drag: " + firstDrag.object.log(), 10, 10);
				// final text = g.drawString("drag: " + firstDrag.object.log(), 10, 10);
				// text.originX = 0;
				// text.originY = 0;
				g.drawString("drag: " + firstDrag.object.log(), 10, 10);
			}
		}
		#if kha
		metrics.addFrame();
		metrics.addRenderTime(Scheduler.realTime() - renderTime);
		if (true) metrics.render(g);
		#end
		g.end();
	}

	public static function countObjects():Int {
		var count = 0;
		function calc(obj:Container):Void {
			count++;
			for (child in obj.children) {
				if (child.type != ContainerType) {
					count++;
					continue;
				}
				calc(cast child);
			}
		}
		for (scene in scenes) calc(scene);
		return count;
	}

	#if phaser
	public static function countPhaserObjects():Int {
		var count = 0;
		function countObjects(parent:phaser.gameobjects.Container):Void {
			count++;
			for (child in parent.list) {
				final child:phaser.gameobjects.Container = cast child;
				if (child.list == null) {
					count++;
					continue;
				}
				countObjects(cast child);
			}
		}
		for (child in _scene.children.getChildren()) {
			final child:phaser.gameobjects.Container = cast child;
			if (child.list == null) continue;
			countObjects(child);
		}
		return count;
	}
	#end

	static function drawAllContainerBounds(g:Graphics, obj:IGameObject, depth = 0):Void {
		final isContainerLike = obj.type.isContainerLike();
		if (!isContainerLike && debugBoundsMode == 1) return;
		final obj:Container = cast obj;
		final p = obj.getGlobalPosition();
		final x = p.x;
		final y = p.y;
		final width = p.w;
		final height = p.h;
		g.color = {
			if (!obj.visible) 0xFF838383;
			else if (obj.alpha == 0) 0xFFCCCC00;
			else switch obj.type {
				case SceneType:
					final isFullScreen = obj.width == App.w && obj.height == App.h;
					isFullScreen ? 0x0 : 0xFF00FFD9;
				case ContainerType: 0xFF00FF00;
				case SpriteType: 0xFF00FF00;
				case _: 0xFF6F00FF;
			}
		}
		var opacity = (1 - depth / 5);
		if (opacity < 0.2) opacity = 0.2;
		// g.opacity = opacity * obj.alpha;
		g.opacity = opacity;
		g.fillRect(x, y, 10, 10);
		final lineW = depth > 1 ? 3 : 2;
		var lineScale = obj.scale;
		if (lineScale > 1) lineScale = 1;
		g.drawRect(x, y, width, height, lineW);
		g.fillRect(x + width / 2 - 1, y + height / 2 - 1, 2 / lineScale, 2 / lineScale);

		if (debugBoundsMode == 1 && depth > 1) return;
		if (!isContainerLike) return;
		depth++;
		for (obj in obj.children) {
			drawAllContainerBounds(g, obj, depth);
		}
	}

	static function drawInputOrderNumbers(g:Graphics, obj:IGameObject, groupId:EventKind<Any>, i:Int):Void {
		// if (g.font == null) {
		// 	if (App.debugFont == null) return;
		// 	g.font = App.debugFont;
		// }
		if (obj.type != ContainerType) return;
		final obj:Container = cast obj;
		// g.transformation.setFrom(obj.transformation);
		#if phaser
		var width = obj.displayWidth;
		var height = obj.displayHeight;
		final m = obj.transformation;
		var x = m._20;
		var y = m._21;
		#end

		final p = obj.topLeftLocal();
		g.color = Green;
		g.opacity = 1;
		g.fontSize = 16;
		final s = '${groupId.toString()}_$i';
		#if phaser
		g.drawString(s, x, y);
		#else
		g.drawString(s, p.x, p.y);
		#end
	}

	static function _onKeyDown(key:KeyCode):Void {
		keys[key] = true;
		onKeyDown(key);
	}

	static function _onKeyUp(key:KeyCode):Void {
		keys.remove(key);
		onKeyUp(key);
	}

	static function _onKeyPress(char:String):Void {
		onKeyPress(char);
	}

	static function createPointerIfNotExist(id:Int):Void {
		pointers[id] ??= new Pointer(id);
		#if kha_html5
		final e = SystemImpl.activeMouseEvent ?? return;
		// when keyup happens with unfocused window
		keys.set(Shift, e.shiftKey);
		keys.set(Control, e.ctrlKey);
		keys.set(Alt, e.altKey);
		keys.set(Meta, e.metaKey);
		#end
	}

	static function _onMouseDown(button:PointerButtonKind, x:Int, y:Int):Void {
		createPointerIfNotExist(0);
		final pointer = pointers[0];
		pointer.startX = x;
		pointer.startY = y;
		pointer.x = x;
		pointer.y = y;
		pointer.button = button;
		pointer.isDown = true;
		pointer.isTouch = false;
		onPointerDown(pointer);
	}

	static function _onMouseMove(x:Int, y:Int, mx:Int, my:Int):Void {
		createPointerIfNotExist(0);
		final pointer = pointers[0];
		pointer.moveX = x - pointer.x;
		pointer.moveY = y - pointer.y;
		pointer.x = x;
		pointer.y = y;
		// pointer.moveX = mx;
		// pointer.moveY = my;
		onPointerMove(pointer);
	}

	static function _onMouseUp(button:PointerButtonKind, x:Int, y:Int):Void {
		createPointerIfNotExist(0);
		final pointer = pointers[0];
		pointer.x = x;
		pointer.y = y;
		pointer.button = button;
		pointer.isDown = false;
		onPointerUp(pointer);
	}

	static function _onTouchDown(id:Int, x:Int, y:Int):Void {
		if (id > 9) return;
		createPointerIfNotExist(id);
		final pointer = pointers[id];
		pointer.startX = x;
		pointer.startY = y;
		pointer.x = x;
		pointer.y = y;
		pointer.isDown = true;
		pointer.isTouch = true;
		onPointerDown(pointer);
	}

	static function _onTouchMove(id:Int, x:Int, y:Int):Void {
		if (id > 9) return;
		createPointerIfNotExist(id);
		final pointer = pointers[id];
		pointer.moveX = x - pointer.x;
		pointer.moveY = y - pointer.y;
		pointer.x = x;
		pointer.y = y;
		pointer.isDown = true;
		onPointerMove(pointer);
	}

	static function _onTouchUp(id:Int, x:Int, y:Int):Void {
		if (id > 9) return;
		createPointerIfNotExist(id);
		final pointer = pointers[id];
		pointer.x = x;
		pointer.y = y;
		pointer.isDown = false;
		onPointerUp(pointer);
	}

	#if kha
	static function _onGamepadAxis(id:Int, value:Float):Void {
		onGamepadAxis(id, value);
	}

	static function _onGamepadButton(id:Int, value:Float):Void {
		onGamepadButton(id, value);
	}

	static function onGamepadAxis(axis:Int, value:Float):Void {
		final dz = gamepadDeadZone;
		if (value < dz && value > -dz) value = 0;
		if (value > 1 - dz) value = 1;
		if (value < -1 + dz) value = -1;
		trace('onGamepadAxis $axis $value');
		if (axis == 0) { // x
			if (value < 0) _onKeyDown(Left);
			else if (value > 0) _onKeyDown(Right);
			else {
				if (keys[Right]) _onKeyUp(Right);
				if (keys[Left]) _onKeyUp(Left);
			}
		}
		if (axis == 1) { // y
			if (value < 0) _onKeyDown(Up);
			else if (value > 0) _onKeyDown(Down);
			else {
				if (keys[Down]) _onKeyUp(Down);
				if (keys[Up]) _onKeyUp(Up);
			}
		}
	}

	static function onGamepadButton(id:Int, value:Float):Void {
		final dz = gamepadDeadZone;
		if (value < dz) value = 0;
		if (value > 1 - dz) value = 1;
		trace('onGamepadButton $id $value');
		switch (id) {
			case 0: // Down (B)
				value > 0 ? _onKeyDown(X) : _onKeyUp(X);
			case 1: // Right (A)
				value > 0 ? _onKeyDown(E) : _onKeyUp(E);
			case 2: // Left (Y)
				value > 0 ? _onKeyDown(Space) : _onKeyUp(Space); // Z
			case 3: // Up (X)
				value > 0 ? _onKeyDown(Q) : _onKeyUp(Q);
			case 4: // L1
			case 5: // R1
				value > 0 ? _onKeyDown(X) : _onKeyUp(X);
			case 6: // L2
			case 7: // R2
				value > 0 ? _onKeyDown(R) : _onKeyUp(R);
			case 8: // Select
				value > 0 ? _onKeyDown(Escape) : _onKeyUp(Escape);
			case 9: // Start
				value > 0 ? _onKeyDown(Return) : _onKeyUp(Return);
			case 12:
				onGamepadAxis(1, -(value > 0 ? value : 0));
			case 13:
				onGamepadAxis(1, (value > 0 ? value : 0));
			case 14:
				onGamepadAxis(0, -(value > 0 ? value : 0));
			case 15:
				onGamepadAxis(0, (value > 0 ? value : 0));
			default:
		}
	}
	#end

	static function onResize():Void {
		Event.emit({
			kind: Resize,
			target: null,
		});
		Event.emit({
			kind: AfterResize,
			target: null
		});
	}

	static function onUpdate():Void {
		Event.emit({
			kind: Update,
			target: null,
		});
	}

	static function onBeforeRender():Void {
		Event.emit({
			kind: BeforeRender,
			target: null,
		});
	}

	static function onKeyDown(key:KeyCode):Void {
		final event:KeyEvent = {
			kind: KeyDown,
			target: null,
			key: key,
		};
		Event.emit(event);
	}

	static function onKeyUp(key:KeyCode):Void {
		final event:KeyEvent = {
			kind: KeyUp,
			target: null,
			key: key,
		};
		Event.emit(event);
	}

	static function onKeyPress(char:String):Void {
		final event:KeyCharEvent = {
			kind: KeyPress,
			target: null,
			char: char,
		};
		Event.emit(event);
	}

	static var isFirstInteraction = false;
	static var totalClicks = 0;
	static var inaction:Tween;
	static var inactionTime = 0;

	static function onPointerDown(p:Pointer):Void {
		final event:PointerEvent = {
			kind: PointerDown,
			target: null,
			pointer: p,
			localX: 0,
			localY: 0,
		};
		Event.emit(event);
		if (!isFirstInteraction) {
			isFirstInteraction = true;
			Lazyr.sendAnalytics("First Interaction");
		}
		totalClicks++;
		inaction?.stop();
		inactionTime = 0;
		inaction = scenes[0].addTimeEvent({
			delay: 5000,
			callback: () -> {
				inactionTime += 5;
				Lazyr.sendAnalytics("Inaction", inactionTime);
			}
		});
	}

	static function onPointerMove(p:Pointer):Void {
		final event:PointerEvent = {
			kind: PointerMove,
			target: null,
			pointer: p,
			localX: 0,
			localY: 0,
		};
		Event.emit(event);
	}

	static function onPointerUp(p:Pointer):Void {
		final event:PointerEvent = {
			kind: PointerUp,
			target: null,
			pointer: p,
			localX: 0,
			localY: 0,
		};
		Event.emit(event);
	}

	static var wheelEventCount = 0;
	static var wheelEventCountStartTime = 0.0;
	static var wheelIsTrackpad = false;

	static function detectTrackpadSpam():Bool {
		if (wheelEventCount == 0) {
			wheelEventCountStartTime = Timer.stamp();
		}
		wheelEventCount++;
		final currentTime = Timer.stamp();
		if (currentTime - wheelEventCountStartTime > 0.02) {
			wheelIsTrackpad = wheelEventCount >= 2;
			wheelEventCount = 0;
		}
		return wheelIsTrackpad;
	}

	static function onMouseWheel(delta:Int):Void {
		createPointerIfNotExist(0);
		static var count = 0;
		count++;
		if (count > 99) count = 0;

		final isTrackpad = detectTrackpadSpam();
		final event:MouseWheelEvent = {
			kind: MouseWheel,
			target: null,
			pointer: pointers[0],
			delta: delta,
			isTrackpad: isTrackpad,
			eventCount: count,
		};
		Event.emit(event);
	}

	static function onMouseLeave():Void {
		for (scene in scenes) {
			scene.onMouseLeave();
		}
	}

	static function main():Void {} // lib hxml completion point
}

#if kha
private class Metrics {
	public var fps(default, null) = 0;
	public var renderTimeCache(default, null) = 0.0;
	public var updateTimeCache(default, null) = 0.0;
	public var deltaTimeCache(default, null) = 0.0;
	public var eventsCache(default, null) = 0.0;

	var time = 0.0;
	var lastTime = 0.0;
	var frames = 0;
	var uframes = 0;
	var renderTime = 0.0;
	var updateTime = 0.0;

	public function new() {}

	public function update():Void {
		var deltaTime = Scheduler.realTime() - lastTime;
		lastTime = Scheduler.realTime();
		time += deltaTime;
		if (time < 1) return;

		fps = frames;
		frames = 0;
		time = 0;

		if (renderTime < 0.001) renderTime = 0.001;
		if (updateTime < 0.001) updateTime = 0.001;
		renderTimeCache = Std.int(renderTime / fps * App.fps * 100);
		updateTimeCache = Std.int(updateTime * 100);
		deltaTimeCache = App.deltaTime.toFixed(2);
		eventsCache = Event.eventCount();
		renderTime = 0;
		updateTime = 0;
		uframes = 0;
		// final txt = '$fps | ${App.w}x${App.h}';
		// final txt2 = '${renderTimeCache}% ${updateTimeCache}%';
		// trace(txt + " | " + txt2);
	}

	public function render(g:Graphics):Void {
		if (g.font == null) {
			if (App.debugFont == null) return;
			g.font = App.debugFont;
		}
		g.transformation.setFrom(FastMatrix3.identity());
		g.opacity = 1;
		g.color = 0xFFFFFFFF;
		g.fontSize = 24;
		// final w = kha.System.windowWidth();
		// final h = kha.System.windowHeight();
		final txt = '$fps | ${App.w}x${App.h}';
		final x = App.w - g.font.width(g.fontSize, txt);
		// final y = App.h - g.font.height(g.fontSize);
		final y = App.scenes[0]?.y ?? 0;
		g.drawString(txt, x, y);

		final txt2 = '$eventsCache $renderTimeCache% $updateTimeCache%';
		final x = App.w - g.font.width(g.fontSize, txt2);
		final y = y + g.font.height(g.fontSize);
		g.drawString(txt2, x, y);
	}

	public inline function addFrame():Void {
		frames++;
	}

	public inline function addRenderTime(time:Float):Void {
		renderTime += time;
	}

	public inline function addUpdateTime(time:Float):Void {
		uframes++;
		updateTime += time;
	}
}
#end

#if phaser
private class GlobalScene extends phaser.Scene {
	var callback:(_scene:GlobalScene) -> Void;

	override public function new(callback:(_scene:GlobalScene) -> Void) {
		super({key: "globalScene"});
		this.callback = callback;
	}

	@:keep function preload() {}

	@:keep function create() {
		callback(this);
	}
}

enum abstract ImageScaleQuality(Int) {
	var Low;
	var High;
}
#end
