package khm;

import khm.Event.EventKind;
import khm.Event.ScenesChangeEventType;
import khm.IContainer.Children;
import khm.IGameObject.GameObjectType;
import khm.Particles;
import khm.ResItem;
import khm.Text.TextStyle;
import khm.Types.Point;
import khm.Types.Rect;
import khm.internal.BaseObjectUtils;
import khm.math.Degrees;
import khm.math.FastMatrix3;

using khm.tools.PhaserTools;

#if kha
import kha.graphics2.Graphics;
import kha.graphics2.ImageScaleQuality;
#end

class Container implements IContainer implements IGameObject {
	static inline var MAX_FLOAT = 999999.0;

	#if kha
	public var phaser:Dynamic = {};
	#elseif phaser
	public var phaser:phaser.gameobjects.Container;
	#end
	public var type:GameObjectType = ContainerType;
	public var parent:IContainer = null;
	public var transformation(get, null) = FastMatrix3.identity();
	public var isInteractive(default, set) = true;
	/** Current container `width` **/
	@:isVar public var width(get, set) = 0.0;
	/** Current container `height` **/
	@:isVar public var height(get, set) = 0.0;

	@:noCompletion function get_width():Float {
		return this.width;
	}

	@:noCompletion function set_width(width:Float):Float {
		return this.width = width;
	}

	@:noCompletion function get_height():Float {
		return this.height;
	}

	@:noCompletion function set_height(height:Float):Float {
		return this.height = height;
	}

	public var x(default, set) = 0.0;
	public var y(default, set) = 0.0;
	public var alpha(default, set) = 1.0;
	public var angle(default, set):Degrees = 0.0;
	public var visible(default, set) = true;
	public var scaleX(default, set) = 1.0;
	public var scaleY(default, set) = 1.0;
	public var originX(default, set) = 0.5;
	public var originY(default, set) = 0.5;
	#if kha
	public var scaleQuality:ImageScaleQuality;
	#end
	@:inheritDoc(IContainer.children)
	public final children:Children = [];
	public var state(default, never):ObjectState;

	public var scale(get, set):Float;
	public var name:Null<String>;

	@:keep @:noCompletion function get_scale():Float {
		return (scaleX + scaleY) / 2;
	}

	@:keep @:noCompletion function set_scale(scale:Float):Float {
		scaleX = scale;
		scaleY = scale;
		return scale;
	}

	public var displayWidth(get, set):Float;
	public var displayHeight(get, set):Float;

	@:noCompletion inline function get_displayWidth():Float
		return width * scaleX;

	@:noCompletion inline function get_displayHeight():Float
		return height * scaleY;

	@:noCompletion inline function set_displayWidth(dw:Float):Float {
		scaleX = dw / width;
		return dw;
	}

	@:noCompletion inline function set_displayHeight(dh:Float):Float {
		scaleY = dh / height;
		return dh;
	}

	public var first(get, never):Null<IGameObject>;

	@:noCompletion inline function get_first():Null<IGameObject> {
		return children[0];
	}

	public var last(get, never):Null<IGameObject>;

	@:noCompletion inline function get_last():Null<IGameObject> {
		return children[children.length - 1];
	}

	public function new() {
		#if kha
		scaleQuality = App.defaultScaleQuality;
		#elseif phaser
		phaser = new phaser.gameobjects.Container(App._scene);
		#end
		#if debug
		if (name == null) {
			final classPath = Type.getClassName(Type.getClass(this));
			final className = classPath.split(".").pop();
			if (className != "Container") name = className;
		}
		#end
	}

	public static function sized(width:Float, ?height:Float) {
		final container = new Container();
		container.width = width;
		container.height = height ?? width;
		return container;
	}

	public function setSize(w:Float, h:Float):Void {
		if (w < 0 || h < 0) throw 'Container size ${w}x${h} cannot be smaller 0x0';
		width = w;
		height = h;
		#if phaser
		phaser.setSize(w, h);
		#end
	}

	public function on<T>(event:EventKind<T>, callback:T, ?pos:DebugPosInfos):Void {
		Event.on(this, event, callback, pos);
	}

	public function off<T>(?event:EventKind<T>, ?callback:T):Void {
		Event.off(this, event, callback);
	}

	@:inheritDoc(IContainer.add)
	public function add(obj:IGameObject):Void {
		if (obj.parent != null) obj.parent.remove(obj);
		obj.parent = this;
		children.push(obj);
		sortEvents(obj);
		#if phaser
		phaser.add(obj.phaserObject());
		#elseif kha
		updateChildTransformation(obj);
		#end
		onContentChanged(Added, obj);
	}

	#if kha
	function updateChildTransformation(obj:IGameObject):Void {
		switch obj.type {
			case SceneType, ContainerType, WidgetType:
				final obj:Container = cast obj;
				obj.updateTransformation();
			case SpriteType:
				final obj:Sprite = cast obj;
				obj.updateTransformation();
			case TweenType:
			case TextType:
			case GraphicsType:
			case ParticlesType:
		}
	}
	#end

	@:inheritDoc(IContainer.addAt)
	public function addAt(index:Int, obj:IGameObject):Void {
		if (obj.parent != null) obj.parent.remove(obj);
		obj.parent = this;
		children.insert(index, obj);
		sortEvents(obj);
		#if phaser
		phaser.addAt(obj.phaserObject(), index);
		#elseif kha
		updateChildTransformation(obj);
		#end
		onContentChanged(Added, obj);
	}

	public dynamic function onContentChanged(type:ScenesChangeEventType, child:Null<IGameObject>):Void {}

	function sortEvents(obj:IGameObject):Void {
		if (App.isAttachedScene(this) || this.findScene()?.isAttached) {
			Event.sortObjectEvents(obj);
			if (obj is IContainer) {
				final obj:IContainer = cast obj;
				obj.children.flatIter(obj -> {
					Event.sortObjectEvents(obj);
				});
			}
		}
	}

	@:inheritDoc(IContainer.remove)
	public function remove(obj:IGameObject):Void {
		final isRemoved = children.remove(obj);
		if (!isRemoved) return;
		#if phaser
		final objPhaser = obj.phaserObject();
		phaser.remove(objPhaser);
		if (objPhaser.removeFromDisplayList != null && objPhaser.scene != null) {
			objPhaser.removeFromDisplayList();
		}
		#elseif kha
		updateChildTransformation(obj);
		#end
		onContentChanged(Removed, obj);
	}

	/** Remove all children from the children list. **/
	public function removeAll():Void {
		for (_ in 0...children.length) {
			remove(children[0]);
			onContentChanged(RemovedAll, children[0]);
		}
	}

	#if kha
	public function render(g:Graphics):Void {
		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);
		for (obj in children)
			obj.render(g);
		popContainerSpace(g);
	}

	@:noCompletion var prevAlpha:Float;
	@:noCompletion var prevScaleQuality:ImageScaleQuality;

	function pushContainerSpace(g:Graphics):Void {
		prevAlpha = g.opacity;
		prevScaleQuality = g.imageScaleQuality;
		g.opacity = g.opacity * alpha;
		g.imageScaleQuality = scaleQuality;
		updateTransformation();
		g.pushTransformation(transformation);
	}

	function popContainerSpace(g:Graphics):Void {
		g.imageScaleQuality = prevScaleQuality;
		g.opacity = prevAlpha;
		g.popTransformation();
	}
	#end

	public var isPixelPerfect = false;

	public function updateTransformation():Void {
		transformation.setFrom(FastMatrix3.identity());
		transformation.setFrom(transformation.multmat(parent.transformation));
		if (isPixelPerfect) {
			BaseObjectUtils.apply2DTransformationPixelPerfect(transformation, this);
		} else {
			BaseObjectUtils.apply2DTransformation(transformation, this);
		}
	}

	/**
		```
		final example = addSprite(Res.atlas.frameName, {
			container: exampleC,
			props: {
				x: 100
			}
		});
		```
		Same `frame` example: `{frame: "frameName", texture: "atlas"}`
	**/
	public function addSprite(frame:ResItem, ?s:{
		?container:Container,
		?onResize:(sprite:Sprite) -> Void,
		?props:{}
	}, ?pos:DebugPosInfos):Sprite {
		frame ??= cast {frame: ""};
		s ??= {};

		final sprite = new Sprite(frame);
		sprite.setPosInfos(pos);
		if (s.container != null) s.container.add(sprite);
		else add(sprite);
		if (sprite.parent.state.debug.offX != null) {
			// shrink is called, offset is not needed
			sprite.state.debug.useParentOffset = false;
		}
		if (s.props != null) setObjectProps(sprite, s.props);
		if (s.onResize != null) {
			sprite.on(Lazyr.resizeKind, e -> s.onResize(sprite), pos);
			s.onResize(sprite);
		}
		return sprite;
	}

	/**
		```
		final exampleC = addContainer({
			onResize: sprite -> {
				posContainer(sprite, {
					portrait: {
						x: 0.5,
						y: 0.5,
						w: 0.95,
						h: 0.9
					},
					landscape: {
						x: 0.5,
						y: 0.5,
						w: 0.9,
						h: 0.9
					},
				});
			}
		});
		```
	**/
	public function addContainer(?s:{
		?container:Container,
		?onResize:(sprite:Container) -> Void,
		?size:{width:Float, height:Float},
		?props:{},
	}, ?pos:DebugPosInfos):Container {
		s ??= {};
		final sprite = new Container();
		sprite.setPosInfos(pos);
		if (s.size != null) sprite.setSize(s.size.width, s.size.height);

		if (s.container != null) s.container.add(sprite);
		else add(sprite);
		if (s.props != null) setObjectProps(sprite, s.props);
		if (s.onResize != null) {
			sprite.on(Lazyr.resizeKind, e -> {
				setPosContainerLock(sprite, false);
				s.onResize(sprite);
				setPosContainerLock(sprite, true);
			}, pos);
			s.onResize(sprite);
			setPosContainerLock(sprite, true);
		}
		#if debug
		sprite.on(BeforeRender, detectZeroSizeContainer);
		#end
		return sprite;
	}

	public function addNineSlice(id:ResItem, ?leftW:Int, ?topH:Int, ?pos:DebugPosInfos):NineSlice {
		final nineSlice = new NineSlice(id, leftW, topH, pos);
		add(nineSlice);
		return nineSlice;
	}

	static function detectZeroSizeContainer(e:Event):Void {
		final sprite:Container = cast e.target;
		if (sprite.width == 0 && sprite.height == 0) {
			final pos = sprite.state.debug.posInfos;
			final name = sprite.name == null ? "container" : 'container ${sprite.name}';
			trace(
				'$name size is 0 before first scene render,' +
				' did you forget to `setSize` or `shrink` it?',
				pos
			);
		}
		sprite.off(BeforeRender, detectZeroSizeContainer);
	}

	static function setPosContainerLock(sprite:IGameObject, isLock:Bool):Void {
		#if debug
		if (!isLock) {
			sprite.state.debug.posContainerLock = null;
			return;
		}
		if (sprite.state.debug.posContainerData == null) return;
		sprite.state.debug.posContainerLock = true;
		#end
	}

	static function setObjectProps(sprite:Any, props:Dynamic):Void {
		for (field in Reflect.fields(props)) {
			final v = Reflect.field(props, field);
			if (v == null) continue;
			Reflect.setProperty(sprite, field, v);
		}
	}

	/**
		```
		final text = addText({
			text: "hello",
			style: {
				fontFamily: mainFont2,
				fontSize: 100,
				color: 0xFFFFFFFF,
				align: Center,
				wordWrap: {
					width: 1000
				}
			}
		});
		text.setShadow(5, 5, 0xFF000000, 5);
		text.setStroke(0xFF000000, 5);
		```
	**/
	public function addText(obj:{
		?container:Container,
		text:String,
		?x:Float,
		?y:Float,
		?style:TextStyle
	}, ?pos:DebugPosInfos):Text {
		final text = new Text(obj.text, obj.style);
		text.setPosInfos(pos);
		text.x = obj.x ?? 0.0;
		text.y = obj.y ?? 0.0;
		text.setOrigin(0.5);
		if (obj.container != null) obj.container.add(text);
		else add(text);
		return text;
	}

	public function addParticles(config:ParticlesConfig, ?pos:DebugPosInfos):Particles {
		final particles = new Particles(config);
		particles.setPosInfos(pos);
		add(particles);
		return particles;
	}

	public function addGraphics(s:{
		onRender:(g:khm.Graphics) -> Void,
	}):khm.Graphics {
		final graphics = new khm.Graphics();
		graphics.on(BeforeRender, e -> {
			graphics.begin();
			s.onRender(graphics);
			graphics.end();
		});
		add(graphics);
		return graphics;
	}

	/**
		Creates empty Tween with `delay` duration that will execute `callback` on `onComplete`.
		Handy alternative to `haxe.Timer.delay` that can be operated with other tweens.

		Created tween is attached to `this` object and can for example be removed with `obj.removeTweens()`.
	**/
	public function delayed(delay:Float, callback:() -> Void, ?pos:DebugPosInfos):Tween {
		if (callback == null) throw "delayed: callback is null";
		return new Tween({
			targets: this,
			ease: Linear,
			duration: delay,
			onComplete: callback
		}, pos);
	}

	public inline function sort(fn:(obj:IGameObject) -> Float):Void {
		final start = 0;
		final length = children.length;
		if (start == length) return;
		var pos = start;
		var maxValue = fn(children[pos++]) ?? throw 'sort value is null (for children[$pos])';
		while (pos < length) {
			final child = children[pos];
			final value = fn(child) ?? throw 'sort value is null (for children[$pos])';
			if (value < maxValue) {
				var p = pos - 1;
				while (p >= start) {
					final child2 = children[p];
					final value2 = fn(child2) ?? throw 'sort value is null (for children[$p])';
					if (value >= value2) break;
					children[p + 1] = child2;
					Event.sortObjectEvents(child2);
					#if phaser
					phaser.list[p + 1] = child2.phaserObject();
					#end
					p--;
				}
				children[p + 1] = child;
				Event.sortObjectEvents(child);
				#if phaser
				phaser.list[p + 1] = child.phaserObject();
				#end
			} else {
				maxValue = value;
			}
			pos++;
		}
	}

	/** Calculates AABB total children size and moves all children to center of container (0,0). **/
	public function shrink():Void {
		final rect = inline getBounds();
		final w = rect.w;
		final h = rect.h;

		if (w <= 0 || h <= 0) {
			var posText = this.posInfosLine() ?? "";
			if (posText.length != 0) posText = '$posText: ';
			throw '${posText}Calculated zero size of container: ${this.log()}';
		}
		setSize(w, h);
		final offX = -rect.x - rect.w / 2;
		final offY = -rect.y - rect.h / 2;
		if (this.state.debug.offX == null) this.state.debug.offX = offX;
		if (this.state.debug.offY == null) this.state.debug.offY = offY;
		for (child in children) {
			child.x += offX;
			child.y += offY;
		}
		Event.emit({
			kind: Lazyr.resizeKind,
			target: this,
		});
	}

	public function fit():Void {
		var maxChildX = -MAX_FLOAT;
		var maxChildY = -MAX_FLOAT;
		for (s in children) {
			final originX = s.originX;
			final originY = s.originY;
			final dw = s.displayWidth;
			final dh = s.displayHeight;
			if (s.x + dw * (1 - originX) > maxChildX) maxChildX = s.x + dw * (1 - originX);
			if (s.y + dh * (1 - originY) > maxChildY) maxChildY = s.y + dh * (1 - originY);
		}
		if (children.length == 0) {
			maxChildX = maxChildY = 0;
		}
		setSize(maxChildX, maxChildY);
		Event.emit({
			kind: Lazyr.resizeKind,
			target: this,
		});
	}

	public function getBounds():Rect {
		var minChildX = MAX_FLOAT;
		var maxChildX = -MAX_FLOAT;
		var minChildY = MAX_FLOAT;
		var maxChildY = -MAX_FLOAT;
		for (s in children) {
			final originX = s.originX;
			final originY = s.originY;
			final dw = s.displayWidth;
			final dh = s.displayHeight;
			if (s.x - dw * originX < minChildX) minChildX = s.x - dw * originX;
			if (s.x + dw * (1 - originX) > maxChildX) maxChildX = s.x + dw * (1 - originX);
			if (s.y - dh * originY < minChildY) minChildY = s.y - dh * originY;
			if (s.y + dh * (1 - originY) > maxChildY) maxChildY = s.y + dh * (1 - originY);
		}
		if (children.length == 0) {
			minChildX = minChildY = maxChildX = maxChildY = 0;
		}
		final w = maxChildX - minChildX;
		final h = maxChildY - minChildY;
		return {
			x: minChildX,
			y: minChildY,
			w: w,
			h: h,
		};
	}

	public function addBitmapMask(maskId:ResItem):Void {
		#if phaser
		final maskSprite = addSprite(maskId);
		maskSprite.visible = false; // enable for debug
		final mask = maskSprite.phaser.createBitmapMask();
		maskSprite.on(BeforeRender, e -> {
			final m = this.getWorldTransformMatrix();
			maskSprite.x = m.getX(0, 0);
			maskSprite.y = m.getY(0, 0);
			maskSprite.scaleX = m.scaleX;
			maskSprite.scaleY = m.scaleY;
			maskSprite.angle = m.rotationNormalized.toDeg();
		});
		phaser.setMask(mask);
		#end
	}

	#if phaser
	public function addGeometryMask(graphics:phaser.gameobjects.Graphics):Void {
		graphics.visible = false;
		#if debug
		// graphics.visible = true;
		// graphics.alpha = 0.2;
		#end
		final mask = graphics.createGeometryMask();
		on(BeforeRender, e -> {
			final m = this.getWorldTransformMatrix();
			graphics.x = m.getX(0, 0);
			graphics.y = m.getY(0, 0);
			graphics.scaleX = m.scaleX;
			graphics.scaleY = m.scaleY;
			graphics.angle = m.rotationNormalized.toDeg().toFloat();
		});
		phaser.setMask(mask);
	}
	#end

	@:noCompletion function set_isInteractive(v) {
		isInteractive = v;
		return v;
	}

	@:noCompletion function set_x(v) {
		posContainerCheck("x", v);
		return phaser.x = this.x = v;
	}

	@:noCompletion function set_y(v) {
		posContainerCheck("y", v);
		return phaser.y = this.y = v;
	}

	@:noCompletion
	inline function posContainerCheck(prop:String, v:Float):Void {
		#if debug
		if (Math.isNaN(v)) throw v;
		if (state.debug.posContainerLock == null) return;
		var posText = this.posInfosLine() ?? "";
		if (posText.length != 0) posText = ' ($posText)';
		final name = name ?? "container";
		trace('Hint: set `$name.state.debug.posContainerLock = null` if you want ignore this error');
		throw 'Cannot set `$prop` to ${this.log()} with `posContainer`$posText';
		#end
	}

	@:noCompletion function set_alpha(v)
		return phaser.alpha = this.alpha = v;

	@:noCompletion function set_angle(v) {
		this.angle = v;
		return phaser.angle = this.angle.toFloat();
	}

	@:noCompletion function set_visible(v)
		return phaser.visible = this.visible = v;

	@:noCompletion function set_scaleX(v) {
		posContainerCheck("scaleX", v);
		return phaser.scaleX = this.scaleX = v;
	}

	@:noCompletion function set_scaleY(v) {
		posContainerCheck("scaleY", v);
		return phaser.scaleY = this.scaleY = v;
	}

	@:noCompletion function set_originX(v)
		return this.originX = v;

	@:noCompletion function set_originY(v)
		return this.originY = v;

	#if kha
	@:noCompletion function get_transformation() {
		return transformation;
	}
	#elseif phaser
	var tempMatrix = new phaser.gameobjects.components.TransformMatrix();

	@:noCompletion function get_transformation() {
		final m = phaser.getWorldTransformMatrix(tempMatrix);
		transformation._00 = m.a;
		transformation._01 = m.b;
		transformation._10 = m.c;
		transformation._11 = m.d;
		transformation._20 = m.e;
		transformation._21 = m.f;
		return transformation;
	}
	#end
}
