package khm;

import khm.Pointer;
import khm.Types;
import khm.utils.macro.EnumAbstractUtils;

enum abstract EventKind<T>(Int) to Int {
	/** Count of all event kinds. **/
	public static inline var LENGTH = 16;
	var Update:EventKind<(e:Event) -> Void>;
	var BeforeRender:EventKind<(e:Event) -> Void>;
	var Resize:EventKind<(e:Event) -> Void>;
	var AfterResize:EventKind<(e:Event) -> Void>;
	var PointerDown:EventKind<(e:PointerEvent) -> Void>;
	var PointerMove:EventKind<(e:PointerEvent) -> Void>;
	var PointerUp:EventKind<(e:PointerEvent) -> Void>;
	var PointerOver:EventKind<(e:PointerEvent) -> Void>;
	var PointerOut:EventKind<(e:PointerEvent) -> Void>;
	var Drag:EventKind<(e:DragEvent) -> Void>;
	var DragEnd:EventKind<(e:DragEvent) -> Void>;
	var KeyDown:EventKind<(e:KeyEvent) -> Void>;
	var KeyUp:EventKind<(e:KeyEvent) -> Void>;
	var KeyPress:EventKind<(e:KeyCharEvent) -> Void>;
	var MouseWheel:EventKind<(e:MouseWheelEvent) -> Void>;
	var ScenesChange:EventKind<(e:ScenesChangeEvent) -> Void>;

	public function toString():String {
		return EnumAbstractUtils.toString(abstract);
	}
}

private typedef EventItem = {
	object:IGameObject,
	callbacks:Array<(e:Event) -> Void>,
	onces:Array<Bool>,
	pos:Array<DebugPosInfos>
}

private typedef DragObjectItem = {
	object:IGameObject,
	localX:Float,
	localY:Float,
	isDragged:Bool
}

@:structInit
@:allow(khm.Event)
class PointerEvent extends Event {
	public final pointer:Pointer;

	/**
		Local x-cord in `target` object local space, that depends on `width` and `originX`.
		Object transformations like `scale` does not affect this value. For example:

		With `originX = 0.5`, value will be in range `-width / 2 ... width / 2`.

		With `originX = 0`, value will be in range `0 ... width`.
	**/
	public var localX(default, null):Float;

	/**
		Local y-cord in `target` object local space that depends on `height` and `originY`.
		Object transformations like `scale` does not affect this value. For example:

		With `originY = 0.5`, value will be in range `-height / 2 ... height / 2`.

		With `originY = 0`, value will be in range `0 ... height`.
	**/
	public var localY(default, null):Float;
	/** Shortcut to `e.pointer`. **/
	public var p(get, never):Pointer;

	inline function get_p():Pointer
		return this.pointer;
}

@:structInit
class DragEvent extends Event {
	public final pointer:Pointer;
	@:inheritDoc(khm.Event.PointerEvent.localX)
	public final localX:Float;
	@:inheritDoc(khm.Event.PointerEvent.localY)
	public final localY:Float;
	/**
		Pointer x-cord relative to `target` parent, so you can drag `target`
		to pointer position simply by setting `target.x = e.dragX`.
	**/
	public final dragX:Float;
	/**
		Pointer y-cord relative to `target` parent, so you can drag `target`
		to pointer position simply by setting `target.y = e.dragY`.
	**/
	public final dragY:Float;
	@:inheritDoc(khm.Event.PointerEvent.p)
	public var p(get, never):Pointer;

	inline function get_p():Pointer
		return this.pointer;
}

@:structInit
class KeyEvent extends Event {
	public final key:KeyCode;
}

@:keep
@:structInit
class KeyCharEvent extends Event {
	public final char:String;
}

@:structInit
class MouseWheelEvent extends Event {
	public final pointer:Pointer;
	public final delta:Int;
	public final isTrackpad:Bool;
	/**
		Cyclic wheel event counter. Can be used to skip trackpad wheel event spam:
		```
		if (e.isTrackpad && e.eventCount % 5 != 0) return;
		final speedY = e.delta * 20; // can be used for trackpad / mouse wheel
		```
	**/
	public final eventCount:Int;
	@:inheritDoc(khm.Event.PointerEvent.p)
	public var p(get, never):Pointer;

	inline function get_p():Pointer
		return this.pointer;
}

@:structInit
class ScenesChangeEvent extends Event {
	public final scene:Scene;
	public final type:ScenesChangeEventType;
}

enum abstract ScenesChangeEventType(Int) {
	var Added;
	var Removed;
	var RemovedAll;
	var Moved;
}

@:structInit
class Event {
	public static var debugObject:Null<IGameObject> = null;
	/** Set the default `event.propagate` mode. **/
	public static var propagateEvents = false;
	/** Enable this to log all events (or use `debugEmitKind` to debug specific event kind). **/
	public static var debugEmitAll = false;
	/** Set event type to this variable to log all events of this kind. **/
	public static var debugEmitKind:Null<EventKind<Any>> = null;

	/**
		2D array for every `EventKind` with game objects with events.
		Order of objects depends on `EventKind`:
		- `Update`, `BeforeRender`, `AfterResize`, `ScenesChange` order is independent of display tree:
		- - Order of such events depends on when `obj.on(Update, ...)` is called in code
		- For `Pointer`, `Key` and other input events, order is reversed:
		- - Top children first, then all children parents, then second top children with parents and so on
	**/
	public static final events:Array<Array<EventItem>> = [
		for (i in 0...EventKind.LENGTH) []
	];
	/**
		Map of `object => Array<EventItem>` items.
		Every object event group has single `EventItem` with all `EventKind` callbacks
		(for example `objectGroups[obj][PointerDown]`).
	**/
	public static final objectGroups:Map<IGameObject, Array<EventItem>> = [];
	static final overObjects:Array<IGameObject> = [];
	static final checkMoveObjects:Array<EventItem> = [];

	/** Array map of pointer id => dragged object and local click x/y. **/
	public static final dragObjects:Array<Null<DragObjectItem>> = [];

	public var target(default, null):Null<IGameObject>;
	public var kind(default, null):EventKind<Any>;
	/**
		Propagate the event to parent objects in display list (`false` by default).
		Propagation is inversed for `Resize` events (from parents to children).
	**/
	// public var propagate = Event.propagateEvents;
	public var propagate(default, set):Bool = Event.propagateEvents;

	function set_propagate(value) {
		// if (propagate != value) {
		// 	if (debugEmitKind != null) {
		// 		trace('propagation: $value');
		// 	}
		// }
		propagate = value;
		return value;
	}

	/** Same as `propagate = value`, but will also log changes with `Event.debugEmitKind`. **/
	public function propagation(value:Bool, ?pos:DebugPosInfos):Bool {
		if (propagate != value) {
			if (debugEmitKind != null) {
				trace('propagation: $value', pos);
			}
		}
		propagate = value;
		return value;
	}

	var immediatePropagation = true;

	/**
		Stops event propagation in `e.target` listeners array of this `e.kind`.

		So, if you have two PointerDown listeners, and call `e.stopImmediatePropagation()`
		in first one, second listener will not be executed.
	**/
	public inline function stopImmediatePropagation():Void {
		immediatePropagation = false;
	}

	public static function on<T>(object:IGameObject, event:EventKind<T>, callback:T, once = false, ?pos:DebugPosInfos):Void {
		listen(object, event, callback, once, pos);
		final event:EventKind<Any> = event;
		if (event == Drag) { // enables dragging
			if (!hasEventCallback(object, PointerDown, onDragActivator)) {
				listen(object, PointerDown, onDragActivator, once, pos);
			}
		}
	}

	static function onDragActivator(e:Event):Void {
		e.propagate = true;
	}

	static function listen<T>(object:IGameObject, event:EventKind<T>, callback:T, once:Bool, ?pos:DebugPosInfos):Void {
		final item = getGroupItem(object, event);
		if (item != null) {
			item.callbacks.push(cast callback);
			item.onces.push(once);
			item.pos.push(pos);
			return;
		}
		final item:EventItem = {
			object: object,
			callbacks: [cast callback],
			onces: [once],
			pos: [pos]
		};
		objectGroups[object] ??= [];
		if (objectGroups[object][event] == null) {
			objectGroups[object][event] = item;
		}
		if (isUnorderedEventKind(event)) {
			final group = events[event];
			group.unshift(item);
		} else {
			insertObjectByDepth(event, item);
		}
	}

	static inline function isUnorderedEventKind(kind:EventKind<Any>):Bool {
		return switch kind {
			case Update, BeforeRender, AfterResize, ScenesChange: true;
			case _: false;
		}
	}

	static function getGroupItem(object:IGameObject, event:EventKind<Any>):Null<EventItem> {
		if (objectGroups[object] == null) return null;
		return objectGroups[object][event];
	}

	public static function off<T>(object:IGameObject, ?event:EventKind<T>, ?callback:T):Void {
		if (objectGroups[object] == null) return;
		// remove all event listeners
		if (event == null) {
			for (i => item in objectGroups[object]) {
				if (item == null) continue;
				offEventGroup(object, cast i, callback);
			}
			objectGroups.remove(object);
			return;
		}

		if (objectGroups[object][event] == null) return;
		final isGroupFullyRemoved = offEventGroup(object, event, callback);
		if (isGroupFullyRemoved) {
			objectGroups[object][event] = null;
			if (isObjectGroupEmpty(object)) objectGroups.remove(object);
		}

		final event:EventKind<Any> = event;
		if (event == Drag) {
			final count = Event.eventCount(Drag, object);
			if (count == 0) {
				final isGroupFullyRemoved = offEventGroup(object, PointerDown, onDragActivator);
				if (isGroupFullyRemoved) {
					objectGroups[object][PointerDown] = null;
					if (isObjectGroupEmpty(object)) objectGroups.remove(object);
				}
			}
		}
	}

	/** Detect nulls instead of event kind arrays to clean `objectGroups` `obj` references. **/
	static function isObjectGroupEmpty(obj:IGameObject):Bool {
		return !objectGroups[obj].inlineExists(item -> item != null);
	}

	/** Returns `true` if `object` event `group` is fully removed (has no callbacks anymore) **/
	static function offEventGroup(object:IGameObject, kind:EventKind<Any>, callback:Null<Any>):Bool {
		final item = objectGroups[object][kind];
		// final item = groups[kind] ?? return true;
		if (callback == null) {
			item.callbacks.resize(0);
			item.onces.resize(0);
			item.pos.resize(0);
			events[kind].remove(item);
			return true;
		}
		final i = item.callbacks.indexOf(callback);
		if (i == -1) return item.callbacks.length == 0;

		item.callbacks.splice(i, 1);
		item.onces.splice(i, 1);
		item.pos.splice(i, 1);
		final isEmpty = item.callbacks.length == 0;
		if (isEmpty) events[kind].remove(item);
		return isEmpty;
	}

	/**
		Emit event object with `kind`, `target` and event-specific data fields.

		If `target` is `null`, event will be fired for all tree objects.
	**/
	public static function emit<T>(event:Event):Void {
		emitEvent(event);
	}

	/** Emit `Resize` event, this is shortcut to `Event.emit({target: obj, kind:Resize})`. **/
	public static function emitResize(obj:Null<IGameObject>, ?propagate:Bool) {
		emitEvent({
			target: obj,
			kind: Resize,
			propagate: propagate,
		});
	}

	/**
		Sort all event listeners in their object's tree order, or for specific `event` kind.

		If you want to resort events for some added object, you should call `sortObjectEvents(obj)` instead.
	**/
	public static function sortAllObjectEvents<T>(?event:EventKind<T>):Void {
		if (event != null) {
			sortGroupObjects(event);
			return;
		}
		for (i in 0...events.length) {
			if (i == AfterResize || i == ScenesChange) continue;
			sortGroupObjects(cast i);
		}
	}

	/**
		Sort all event listeners for `obj` in tree order, or for specific `event` kind.

		If `obj` does not have event listeners of any or specific `event` kind, then sorting can be skipped,
		so it's recommended to only call resort when you add new object to container and pass it as `obj`.
	**/
	public static function sortObjectEvents<T>(obj:IGameObject, ?event:EventKind<T>):Void {
		if (event != null) {
			sortObjectsInGroupByDepth(obj, event);
			return;
		}
		if (objectGroups[obj] == null) return;
		for (i => group in objectGroups[obj]) {
			final kind:EventKind<Any> = cast i;
			if (isUnorderedEventKind(kind)) continue;
			sortObjectsInGroupByDepth(obj, kind);
		}
	}

	static function sortObjectsInGroupByDepth<T>(obj:IGameObject, event:EventKind<T>):Void {
		// if (obj.findScene() == null) return;
		// if there is already listeners item for current object
		final item = getGroupItem(obj, event);
		// we don't need to sort event group, because `obj` does not have listeners of this kind
		if (item == null) return;
		sortGroupObjects(event, item);
	}

	static function sortGroupObjects<T>(event:EventKind<T>, ?addedItem:EventItem):Void {
		final group = events[event];
		if (addedItem != null) {
			final has = group.remove(addedItem);
			if (has) insertObjectByDepth(event, addedItem);
			#if debug
			if (!group.contains(addedItem)) {
				trace(addedItem);
				throw addedItem;
			}
			#end
			return;
		}
		final groupCopy = group.copy();
		group.resize(0);
		for (item in groupCopy) {
			insertObjectByDepth(event, item);
		}
	}

	/**
		If `item` is after `array item` in common parent node,
		then insert it before `array item` input group
	**/
	static function insertObjectByDepth<T>(event:EventKind<T>, item:EventItem):Void {
		final group = events[event];
		final level = getObjectLevel(item.object);
		for (i => item2 in group) {
			// cast should be safe for `.parent` access,
			// common parent will be a container or null
			var obj:IContainer = cast item.object;
			var obj2:IContainer = cast item2.object;
			var lvl = level;
			var lvl2 = getObjectLevel(obj2);
			var parent = obj;
			var parent2 = obj2;
			// search common parent for both objects
			while (lvl > lvl2) {
				obj = parent;
				parent = parent.parent;
				lvl--;
			}
			while (lvl2 > lvl) {
				obj2 = parent2;
				parent2 = parent2.parent;
				lvl2--;
			}
			while (parent != parent2) {
				obj = parent;
				obj2 = parent2;
				parent = parent.parent;
				parent2 = parent2.parent;
			}
			// scene or object without parent
			if (parent == null) {
				// non-scene object will be resorted after added to parent, can be added to group end
				if (obj.type != SceneType) break;
				// if scene is added after `obj2.scene`, add scene before it
				if (App.scenes.indexOf(cast obj) > App.scenes.indexOf(cast obj2)) {
					group.insert(i, item);
					return;
				}
			} else {
				// add object to event group after obj2 index
				if (parent.children.indexOf(obj) > parent.children.indexOf(obj2)) {
					group.insert(i, item);
					return;
				}
			}
		}
		// add first event group, or event from obj without parent
		group.push(item);
	}

	static function getObjectLevel(obj:IGameObject):Int {
		var level = 0;
		var container = obj.parent;
		while (container != null) {
			container = container.parent;
			level++;
		}
		return level;
	}

	static function isObjectInteractive(obj:IGameObject):Bool {
		while (obj != null) {
			if (!obj.isInteractive) return false;
			if (!obj.visible) return false;
			if (obj.alpha == 0) return false;
			obj = obj.parent;
		}
		return true;
	}

	static function whyObjectIsNotInteractive(obj:IGameObject):String {
		var name = "obj";
		while (obj != null) {
			if (!obj.isInteractive) return '$name.isInteractive == false';
			if (!obj.visible) return '$name.visible == false';
			if (obj.alpha == 0) return '$name.alpha == 0';
			obj = obj.parent;
			name += ".parent";
		}
		return "unknown reason";
	}

	static function getEventItemGroup<T>(e:EventItem):Null<EventKind<T>> {
		if (objectGroups[e.object] == null) return null;
		for (i => group in objectGroups[e.object]) {
			if (group == e) return cast i;
		}
		return null;
	}

	public static function getEventItems(target:Null<IGameObject>, event:EventKind<Any>):Array<EventItem> {
		final items:Array<EventItem> = [];
		for (item in events[event]) {
			if (target != null && item.object != target) continue;
			items.push(item);
			if (target != null) break;
		}
		return items;
	}

	public static function getCallbacks(object:IGameObject, kind:EventKind<Any>):Array<(e:Event) ->
		Void> {
		final groups = objectGroups[object] ?? return [];
		final group = groups[kind] ?? return [];
		return group.callbacks.copy();
	}

	public static function getKindsOf(object:IGameObject):Array<EventKind<Any>> {
		final kinds:Array<EventKind<Any>> = [];
		final groups = objectGroups[object] ?? return kinds;
		for (i => item in groups) {
			if (item != null) kinds.push(cast i);
		}
		return kinds;
	}

	static function inputCallback(e:EventItem, event:Event):Bool {
		if (!isObjectInteractive(e.object)) {
			if (App.keys.get(Shift)) {
				final reason = whyObjectIsNotInteractive(e.object);
				trace('Not interactive ($reason): ' + e.object.log(), e.object);
			}
			return false;
		}
		return eventCallback(e, event);
	}

	static function eventCallback(e:EventItem, event:Event):Bool {
		event.propagate = propagateEvents;
		event.immediatePropagation = true;
		event.target = e.object ?? throw "event.target is null";
		logEventItem(e);
		var i = 0;
		while (i < e.callbacks.length) {
			final callback = e.callbacks[i];
			final once = e.onces[i];
			if (!event.immediatePropagation) break;
			event.propagate = propagateEvents;
			callback(event);
			if (once) {
				e.callbacks.splice(i, 1);
				e.onces.splice(i, 1);
				e.pos.splice(i, 1);
				continue;
			}
			i++;
		}

		final group:Array<EventItem> = events[event.kind];
		final isEmpty = e.callbacks.length == 0;
		if (isEmpty) group.remove(e);

		return true;
	}

	static function logEventItem(e:EventItem):Void {
		if (debugEmitAll || debugEmitKind != null) {
			final kind = getEventItemGroup(e);
			// final name = Type.getClassName(Type.getClass(e.object));
			if (debugEmitAll) {
				if (kind != null) trace(kind.toString(), e.object.log());
			} else if (debugEmitKind != null) {
				if (kind == debugEmitKind) {
					for (i => pos in e.pos) {
						final n = e.pos.length == 1 ? "" : ' ($i)';
						trace(kind.toString() + n, e.object.log(), pos);
					}
				}
			}
		}
	}

	static function emitDragEvent(kind:EventKind<Any>, p:Pointer, obj:DragObjectItem):Void {
		final hasDragEvent = getGroupItem(obj.object, kind) != null;
		if (!hasDragEvent) return;
		var dragX = p.x - obj.localX;
		var dragY = p.y - obj.localY;
		if (obj.object.parent != null) {
			final sx = obj.object.transformation._00;
			final sy = obj.object.transformation._11;
			final point = p.localPointTo(
				obj.object.parent.transformation,
				-obj.localX * sx,
				-obj.localY * sy,
			);
			dragX = point.x;
			dragY = point.y;
		}
		final event:DragEvent = {
			kind: kind,
			target: obj.object,
			pointer: p,
			localX: obj.localX,
			localY: obj.localY,
			dragX: dragX,
			dragY: dragY,
		};
		emitEvent(event);
	}

	static function emitEvent<T>(event:Event):Void {
		final group:Array<EventItem> = events[event.kind];
		final toObject = event.target;
		inline function isNonTarget(object:IGameObject):Bool {
			return toObject != null && toObject != object;
		}
		switch event.kind {
			case Update, BeforeRender, AfterResize, ScenesChange:
				var i = group.length;
				while (i-- > 0) {
					// if event callback removed multiple items from group
					final e = group[i] ?? continue;
					if (isNonTarget(e.object)) continue;
					// we dont need interactive check for this event (for tweens)
					eventCallback(e, event);
				}
			case Resize:
				// reverse order for resize events: parents first
				var i = group.length;
				while (i-- > 0) {
					final e = group[i] ?? continue;
					if (isNonTarget(e.object)) continue;
					eventCallback(e, event);
				}
			case PointerDown, PointerUp:
				final event:PointerEvent = cast event;
				final p = event.p;
				final dragObj = dragObjects[p.id];
				if (event.kind == PointerUp && dragObj != null && dragObj.isDragged) {
					emitDragEvent(DragEnd, p, dragObj);
				}
				// clear pointer drag obj before PointerDown / after PointerUp
				dragObjects[p.id] = null;
				for (e in group) {
					if (isNonTarget(e.object)) continue;
					if (!hitTest(p, e.object)) continue;
					final x = localPoint.x;
					final y = localPoint.y;
					event.localX = x;
					event.localY = y;
					final hasInput = inputCallback(e, event);
					if (hasInput) {
						// mark object for Drag event in PointerMove
						if (event.kind == PointerDown && dragObjects[p.id] == null) {
							final hasDragEvent = getGroupItem(e.object, Drag) != null;
							if (hasDragEvent) {
								dragObjects[p.id] = {
									object: e.object,
									localX: x,
									localY: y,
									isDragged: false
								};
							}
						}
					}
					if (hasInput && !event.propagate) return;
				}
			case PointerMove:
				final event:PointerEvent = cast event;
				final p = event.p;

				for (e in events[Drag]) {
					final obj = dragObjects.inlineFind(obj -> {
						return obj != null && obj.object == e.object;
					});
					// final obj = dragObjects[p.id]; // not sure
					if (obj == null) continue;
					obj.isDragged = true;
					emitDragEvent(Drag, p, obj);
					event.propagate = false;
					break;
				}

				checkMoveObjects.resize(0);
				for (e in events[PointerOver]) {
					checkMoveObjects.push(e);
				}
				for (e in events[PointerOut]) {
					if (checkMoveObjects.contains(e)) continue;
					checkMoveObjects.push(e);
				}

				inline function checkOver(e:EventItem):Void {
					if (!overObjects.contains(e.object)) {
						overObjects.push(e.object);
						final event:PointerEvent = {
							kind: PointerOver,
							target: e.object,
							pointer: p,
							localX: localPoint.x,
							localY: localPoint.y,
						};
						emitEvent(event);
					}
				}
				inline function checkOut(e:EventItem):Void {
					if (overObjects.contains(e.object)) {
						overObjects.remove(e.object);
						final event:PointerEvent = {
							kind: PointerOut,
							target: e.object,
							pointer: p,
							localX: localPoint.x,
							localY: localPoint.y,
						};
						emitEvent(event);
						// emitEvent(e.object, PointerOut, p, localPoint.x, localPoint.y);
					}
				}

				for (e in group) {
					if (isNonTarget(e.object)) continue;
					if (hitTest(p, e.object)) {
						// object is hovered
						event.localX = localPoint.x;
						event.localY = localPoint.y;
						checkOver(e);
						final hasInput = inputCallback(e, event);
						if (hasInput && !event.propagate) break;
						event.propagate = false;
					} else {
						// no hit and object in hovered before
						checkOut(e);
					}
					checkMoveObjects.remove(e);
				}

				// over/out objects without move event
				for (e in checkMoveObjects) {
					if (hitTest(p, e.object)) {
						checkOver(e);
					} else {
						checkOut(e);
					}
				}

			case PointerOver, PointerOut:
				for (e in group) {
					if (isNonTarget(e.object)) continue;
					final hasInput = inputCallback(e, event);
					if (hasInput && !event.propagate) return;
				}

			case Drag, DragEnd:
				for (e in group) {
					if (isNonTarget(e.object)) continue;
					final hasInput = inputCallback(e, event);
					if (hasInput && !event.propagate) return;
				}

			case KeyDown, KeyUp:
				// final key:KeyCode = arg;
				for (e in group) {
					final hasInput = inputCallback(e, event);
					if (hasInput && !event.propagate) return;
				}
			case KeyPress:
				// final char:String = arg;
				for (e in group) {
					final hasInput = inputCallback(e, event);
					if (hasInput && !event.propagate) return;
				}
			case MouseWheel:
				final event:MouseWheelEvent = cast event;
				final p = event.pointer;
				for (e in group) {
					if (isNonTarget(e.object)) continue;
					if (hitTest(p, e.object)) {
						final hasInput = inputCallback(e, event);
						if (hasInput && !event.propagate) return;
					}
				}
		}
	}

	static final localPoint:Point = {x: 0, y: 0};

	/**
		This will also can test objects that are hidden
		and haven't updated yet their transformation matrix in `render`.
	**/
	public static function hitTest(p:Pointer, obj:IGameObject):Bool {
		// if (obj.parent == null) return false;
		if (obj.scaleX == 0 || obj.scaleY == 0) return false;
		if (obj.transformation.scaleX * obj.transformation.scaleY < 0.00001) return false;
		#if debug
		final localP = try {
			p.localPointTo(obj.transformation);
		} catch (e) {
			trace(obj.log(), obj);
			throw e;
		}
		#else
		final localP = p.localPointTo(obj.transformation);
		#end
		localPoint.x = localP.x;
		localPoint.y = localP.y;
		if (obj.type == SceneType) {
			final obj:Scene = cast obj;
			if (obj.isUnlimitedSize) return true;
		}
		final topLeft = obj.topLeftLocal();
		final result = khm.utils.Collision.aabb({
			x: topLeft.x,
			y: topLeft.y,
			w: obj.width,
			h: obj.height
		}, {
			x: localPoint.x,
			y: localPoint.y,
			w: 1,
			h: 1
		});
		return result;
	}

	public static function eventCount(?kind:EventKind<Any>, ?object:IGameObject):Int {
		if (kind != null) return eventCountGroup(kind, object);
		var count = 0;
		for (i => group in events) {
			count += eventCountGroup(cast i, object);
		}
		return count;
	}

	static function eventCountGroup(kind:EventKind<Any>, ?object:IGameObject):Int {
		var count = 0;
		for (item in events[kind]) {
			if (object != null && item.object != object) continue;
			count += item.callbacks.length;
			if (object != null) break;
		}
		return count;
	}

	public static function hasEventCallback<T>(object:IGameObject, kind:EventKind<T>, callback:T):Bool {
		final groups = objectGroups[object] ?? return false;
		final group = groups[kind] ?? return false;
		return group.callbacks.contains(cast callback);
	}

	public static function traceLeakedListeners():Void {
		var hasLeaks = false;
		for (i => group in events) {
			for (item in group) {
				var object = item.object;
				if (object.type == TweenType) {
					final tween:Tween = cast object;
					for (target in tween.targets) {
						final target:IGameObject = cast target;
						// if target is in tree, set it as active object to skip tracing
						if (App.isAttachedScene(target) || target.findScene()?.isAttached) {
							object = target;
							break;
						}
					}
				}
				if (App.isAttachedScene(object)) continue;
				if (object.findScene()?.isAttached) continue;
				hasLeaks = true;
				trace(object.log(), item.pos[0]);
				for (i => pos in item.pos) {
					trace(item.callbacks[i], item.pos[i]);
				}
			}
		}
		trace("Objects count: " + objectGroups.count());
		if (!hasLeaks) trace("No event leaks ≈)");
	}
}
