package khm;

import khm.Event.EventKind;
import khm.Pointer;
import khm.ResItem;
import khm.Scene;
import khm.Sound.Sound;
import khm.Sound.SoundConfig;
import khm.Tween;
import khm.Types.Point;
import khm.Types.Rect;
import khm.Types.Size;
import khm.math.Degrees;
import khm.math.FastMatrix3;
import khm.tools.GameObjectAnimTools;
import khm.utils.Clipboard;
import khm.utils.FrameUtils;

using khm.tools.PhaserTools;

private typedef TRect = {
	y:Float,
	x:Float,
	scaleY:Float,
	scaleX:Float,
	angle:Degrees
};

typedef PosObjectConfig = {
	var ?padding:Float;

	/**
		```

		TopLeft    TopCenter    TopRight
		CenterLeft Center       CenterRight
		BottomLeft BottomCenter BottomRight
		```
	**/
	var ?align:Alignement;
	var portrait:PosRect;
	var ?landscape:PosRect;
}

typedef PosRect = Rect & {
	var ?padding:Float;

	/**
		```

		TopLeft    TopCenter    TopRight
		CenterLeft Center       CenterRight
		BottomLeft BottomCenter BottomRight
		```
	**/
	var ?align:Alignement;
}

typedef TimerEventConfig = {
	/** `[delay=1000]` - The delay after which the Timer Event should fire, in milliseconds. **/
	var ?delay:Float;
	/** `[repeat=-1]` - The total number of times the Timer Event will repeat before finishing. **/
	var ?repeat:Int;
	/** `[loop=false]` - `true` if the Timer Event should repeat indefinitely. **/
	@:noCompletion var ?loop:Bool;
	/** `[callback]` - The callback which will be called when the Timer Event fires. **/
	var ?callback:() -> Void;
	/** `[timeScale=1]` - The scale of the elapsed time. **/
	// var ?timeScale:Float;
	/** `[startAt=0]` - The initial elapsed time in milliseconds. Useful if you want a long duration with repeat, but for the first loop to fire quickly. **/
	// var ?startAt:Float;
	/** `[startRatio=0.0]` - Ratio of elapsed time from 0 to 1. Useful if you want a long duration with repeat, but for the first loop to fire quickly. **/
	var ?startRatio:Float;
	/** `[paused=false]` - `true` if the Timer Event should be paused. **/
	var ?paused:Bool;
};

private typedef Config = {
	?defaultAudio:String,
	?useLegazyResize:Bool
}

class Lazyr {
	static var defaultAudio:String;
	public static var isAudioInited = false;
	public static var isSceneRestarted = false;
	public static var portraitRatio = 1 / 1;
	@:noCompletion
	public static var resizeKind = EventKind.Resize;

	public static function initLazyr(config:Config):Void {
		resizeKind = config.useLegazyResize ? AfterResize : Resize;
	}

	public static function removeObject(sprite:IGameObject):Void {
		sprite.off(resizeKind);
		sprite.parent.remove(sprite);
		#if phaser
		sprite.phaserObject().destroy();
		#end
	}

	public static function posContainer(sprite:Container, s:PosObjectConfig):Void {
		#if debug
		if (sprite.parentContainer() != null) {
			throw 'posContainer cannot be used on container with parentContainer()';
		}
		#end
		if (sprite.width == 0 || sprite.height == 0) return;
		final scene = sprite.findScene() ?? throw '${sprite.log()} is not attached to scene';
		var size = scene.isPortrait ? s.portrait : s.landscape;
		size ??= s.portrait;
		final align:Alignement = size.align ?? s.align ?? Center;
		final paddingRatio = size.padding ?? s.padding ?? 0.0;
		final padding = Math.min(scene.sceneW, scene.sceneH) * paddingRatio;
		final scaleX = scene.sceneW / sprite.width * (size.w - paddingRatio * 2);
		final scaleY = scene.sceneH / sprite.height * (size.h - paddingRatio * 2);
		sprite.scale = Math.min(scaleY, scaleX);
		final spriteW = sprite.displayWidth + padding * 2;
		final spriteH = sprite.displayHeight + padding * 2;
		sprite.x = scene.sceneW * size.x + spriteW * (0.5 - align.x);
		sprite.y = scene.sceneH * size.y + spriteH * (0.5 - align.y);
		#if debug
		sprite.state.debug.posContainerData = s;
		#end
	}

	@:noCompletion
	public static function updateContainerSize(sprite:Container, moveToCenter = true):Void {
		if (!moveToCenter) {
			sprite.fit();
			return;
		}
		sprite.shrink();
	}

	@:deprecated("use container.getBounds()")
	public static function getContainerRect(sprite:Container):Rect {
		return sprite.getBounds();
	}

	public static function setResizeCallback<T:IGameObject>(sprite:T, onResize:(sprite:T) -> Void):Void {
		removeResizeCallback(sprite);
		Event.on(sprite, resizeKind, e -> onResize(sprite));
		resizeObject(sprite);
	}

	public static function removeResizeCallback(sprite:IGameObject):Void {
		sprite.state.debug.posContainerData = null;
		sprite.state.debug.posContainerLock = null;
		Event.off(sprite, resizeKind);
	}

	public static function posTo<T:IGameObject>(
		s:TweenConfig,
		onResize:(sprite:T) -> Void,
		?onComplete:() -> Void,
		?pos:DebugPosInfos
	):Tween {
		final sprite:T = s.targets;
		#if debug
		if (sprite.parentContainer() != null) {
			throw 'posTo cannot be used on object with parentContainer()';
		}
		if (s.onComplete != null) throw "Use posTo onComplete argument instead";
		#end
		removeResizeCallback(sprite);
		final oldX = sprite.x;
		final oldY = sprite.y;
		final oldScaleX = sprite.scaleX;
		final oldScaleY = sprite.scaleY;
		onResize(sprite);
		s.x = sprite.x;
		s.y = sprite.y;
		s.scaleX = sprite.scaleX;
		s.scaleY = sprite.scaleY;
		sprite.x = oldX;
		sprite.y = oldY;
		sprite.scaleX = oldScaleX;
		sprite.scaleY = oldScaleY;
		s.onComplete = () -> {
			setResizeCallback(sprite, onResize);
			if (onComplete != null) onComplete();
		}
		return new Tween(s, pos);
	}

	public static function copyPosAndSizeTo(to:Container, from:IGameObject):Void {
		to.width = from.width;
		to.height = from.height;
		to.scaleX = from.scaleX;
		to.scaleY = from.scaleY;
		to.angle = from.angle;
		to.x = from.x;
		to.y = from.y;
	}

	public static function addTweenCounter(config:TweenConfig, ?pos:DebugPosInfos):Tween {
		return Tween.counter(config, pos);
	}

	public static function tweenTint(sprite:Sprite, obj:{
		?start:Color,
		end:Color,
		?duration:Float,
		?delay:Float,
		?repeat:Int,
		?yoyo:Bool,
		?repeatDelay:Float,
		?onComplete:() -> Void
	}):Tween {
		obj.start ??= sprite.tint;
		obj.start ??= 0xFFFFFFFF;
		obj.duration ??= 1000;
		obj.delay ??= 0;
		obj.repeat ??= 0;
		obj.yoyo ??= false;
		obj.repeatDelay ??= 0;
		final tween = Tween.counter({
			from: 0,
			to: 100,
			targets: sprite,
			duration: obj.duration,
			delay: obj.delay,
			repeat: obj.repeat,
			yoyo: obj.yoyo,
			repeatDelay: obj.repeatDelay,
			onUpdate: t -> {
				final tint = MathLib.colorLerp(obj.start, obj.end, t.ratio);
				sprite.tint = tint;
			},
			onComplete: obj.onComplete
		});
		if (obj.duration + obj.delay == 0) {
			sprite.tint = obj.end;
		}
		return tween;
	}

	public static function showSprite(sprite:Any, time = 500.0, delay = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return GameObjectAnimTools.showAnim(sprite, time, delay, onComplete, p);
	}

	public static function hideSprite(sprite:Any, time = 500.0, delay = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return GameObjectAnimTools.hideAnim(sprite, time, delay, onComplete, p);
	}

	public static function opacitySprite(sprite:Any, alpha:Float, time = 500.0, delay = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return GameObjectAnimTools.opacityAnim(sprite, alpha, time, delay, onComplete, p);
	}

	public static function popUpSprite(sprite:Any, time = 500.0, delay = 0.0, scale = 1.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return GameObjectAnimTools.popUpAnim(sprite, time, delay, scale, onComplete, p);
	}

	public static function popOutSprite(sprite:Any, time = 500.0, delay = 0.0, scale = 0.0, ?onComplete:() ->
		Void, ?p:DebugPosInfos):Tween {
		return GameObjectAnimTools.popOutAnim(sprite, time, delay, scale, onComplete, p);
	}

	public static function slideInX(
		sprite:Any, offsetX:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return
			GameObjectAnimTools.slideInAnimX(sprite, offsetX, ease, time, delay, onComplete, p);
	}

	public static function slideInY(
		sprite:Any, offsetY:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return
			GameObjectAnimTools.slideInAnimY(sprite, offsetY, ease, time, delay, onComplete, p);
	}

	public static function slideOutX(
		sprite:Any, offsetX:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return
			GameObjectAnimTools.slideOutAnimX(sprite, offsetX, ease, time, delay, onComplete, p);
	}

	public static function slideOutY(
		sprite:Any, offsetY:Float, ease = Ease.BackOut,
		time = 500.0, delay = 0.0, ?onComplete:() -> Void, ?p:DebugPosInfos
	):Tween {
		return
			GameObjectAnimTools.slideOutAnimY(sprite, offsetY, ease, time, delay, onComplete, p);
	}

	/**
		```
		spriteScaleAnimation(play_now, {
			to: 1.07,
			duration: 1500,
		});
		```
	**/
	public static function spriteScaleAnimation(obj:IGameObject, config:{
		?from:Float,
		to:Float,
		duration:Float,
		?getScale:() -> Float,
		?repeatDelay:Float,
		?delay:Float
	}, ?p:DebugPosInfos):Tween {
		return GameObjectAnimTools.pulseAnim(obj, config, p);
	}

	/**
		```
		final explosion = createAnim({
			key: Res.explosion._id,
			frames: Res.explosion._values,
			frameRate: 15,
			hideOnComplete: true
		});

		sprite.play(explosion);
		```
	**/
	public static function createAnim(obj:{
		key:String,
		frames:Array<String>,
		?sortFrames:Bool,
		?defaultTextureKey:String,
		?frameRate:Float,
		?duration:Float,
		?skipMissedFrames:Bool,
		?delay:Float,
		?repeat:Float,
		?repeatDelay:Float,
		?yoyo:Bool,
		?showOnStart:Bool,
		?hideOnComplete:Bool,
	}):Animation {
		obj.defaultTextureKey ??= obj.key;
		obj.sortFrames ??= true;
		if (obj.sortFrames) {
			obj.frames.sort(FrameUtils.numericStringSort);
			obj.sortFrames = false;
		}
		final frames:Array<ResItem> = obj.frames.map(s -> {
			texture: obj.defaultTextureKey,
			frame: s
		});
		obj.frames = cast frames;
		return cast obj;
	}

	public static function playAudioSprite(frame:String, ?config:SoundConfig, ?pos:DebugPosInfos):Sound {
		var sprite = config?.sprite ?? defaultAudio ?? frame;

		#if phaser
		final psound:phaser.sound.NoAudioSoundManager = App._scene.sound;
		if (!psound.jsonCache.exists(sprite)) return null;
		#end

		final sound = new Sound(sprite, config, pos);
		if (!isAudioInited) return sound;
		sound.play(frame);

		return sound;
	}

	public static function addAudioSprite(frame:String, ?config:SoundConfig, ?pos:DebugPosInfos):Sound {
		return playAudioSprite(frame, config, pos);
	}

	public static var devicePixelRatio(get, never):Float;

	static function get_devicePixelRatio():Float {
		#if !js
		return 1;
		#else
		return js.Browser.window.devicePixelRatio ?? 1;
		#end
	}

	public static function moveToParent(sprite:IGameObject, parent:Null<IContainer>):Void {
		if (parent == null) {
			parent = sprite.findScene() ?? throw '${sprite.log()} is not attached to scene';
		}
		parent.attach(sprite);
	}

	public static function getLocalRectTo(sprite:IGameObject, parent:IContainer):TRect {
		final m = sprite.transformation;
		final rect:TRect = {
			x: m.getX(0, 0),
			y: m.getY(0, 0),
			scaleX: m.scaleX,
			scaleY: m.scaleY,
			angle: m.rotationNormalized.toDeg()
		};
		if (parent == null) {
			return rect;
		} else {
			final m:FastMatrix3 = parent.transformation.inverse();
			// m.invert();
			final x = rect.x;
			final y = rect.y;
			rect.x = m.getX(x, y);
			rect.y = m.getY(x, y);
			rect.scaleX *= m.scaleX;
			rect.scaleY *= m.scaleY;
			rect.angle += m.rotationNormalized.toDeg();
			return rect;
		}
	}

	public static function makeLink(sprite:IGameObject, clickSound:Null<String>):Void {
		sprite.on(PointerDown, e -> {
			sendAnalytics("Install");
			#if js
			if (clickSound != null) playAudioSprite(clickSound);
			untyped window["trackClick"]();
			#else
			trace("link clicked");
			#end
			e.propagate = e.pointer.button == Right;
		});
	}

	public static function addContainerBounds(container:Container):Void {
		#if debug
		final g = container.state.debug.containerBoundsPreview ?? new Graphics();
		g.clear();
		if (container.alpha == 0) return;
		final s = container.scale;
		g.color = 0xFFFF0000;
		g.drawRect(
			container.x - container.displayWidth / 2,
			container.y - container.displayHeight / 2,
			container.displayWidth,
			container.displayHeight
		);
		g.color = 0xFF0000FF;
		g.fillRect(container.x - 2, container.y - 2, 4, 4);
		g.setDepth(100);
		if (container.parent == null) throw "container is not attached to scene";
		container.parent.add(g);
		container.state.debug.containerBoundsPreview = g;
		#end
	}

	static final matchChars = ~/[^0-9A-z_]/g;
	static final startsWithNumber = ~/^[0-9]/g;

	/** Use `useParentOffset: false` if you want to use cords after `container.shrink()` call. **/
	public static function draggableDebug(obj:IGameObject, ?useParentOffset:Bool):Void {
		#if debug
		final sprite = obj;
		// sprite.on(PointerDown, (p, x, y) -> {});
		sprite.on(Drag, e -> {
			sprite.x = e.dragX;
			sprite.y = e.dragY;
			// sprite.x -= sprite.width * (0.5 - sprite.originX);
			// sprite.y -= sprite.height * (0.5 - sprite.originY);
			// trace('${Math.round(sprite.x)}, ${Math.round(sprite.y)}');
			resizeObject(sprite.parentContainer());
		});
		sprite.on(DragEnd, e -> {
			var name = genVarName(sprite);
			sprite.x = e.dragX;
			sprite.y = e.dragY;
			var x = sprite.x;
			var y = sprite.y;
			var offX = 0.0;
			var offY = 0.0;
			final parent = sprite.parentContainer();
			useParentOffset ??= sprite.state.debug.useParentOffset ?? true;
			if (parent != null && useParentOffset) {
				// updateContainerSize(parent);
				if (parent.state.debug.offX != null) {
					offX = parent.state.debug.offX;
					offY = parent.state.debug.offY;
				}
			}
			x = Math.round(x - offX);
			y = Math.round(y - offY);
			final text = '\n${name}.x = ${x};\n${name}.y = ${y};';
			Clipboard.copy(text);
			trace('$x, $y');
		});
		#end
	}

	static function genVarName(obj:IGameObject):String {
		var sprite:Sprite = cast obj;
		if (obj.type == ContainerType) {
			final obj:Container = cast obj;
			if (obj.name != null) {
				if (!matchChars.match(obj.name)) return obj.name;
			}
			sprite = cast obj.children.find(obj -> obj.type == SpriteType) ?? cast obj;
		}
		var name = sprite.id == null ? "sprite" : sprite.id.frame;
		if (name == "__BASE" && sprite.id.texture != null) name = sprite.id.texture;
		if (name == null || name.length == 0) name = "sprite";
		name = matchChars.replace(name, "_").toLowerCase();
		if (startsWithNumber.match(name)) name = '_$name';
		if (obj.type == ContainerType) name += "C";
		return name;
	}

	/**
		Saves value from `valueCallback()` to `localStorage[valueKey]` before app is exited (or browser tab is closed).

		Returns previously saved value it it exists, or `valueCallback()` as default value.
	**/
	public static function keepDebugValue<T>(valueKey:String, valueCallback:() -> T):T {
		#if js
		#if !debug
		return valueCallback();
		#end
		final storage = js.Browser.getLocalStorage();
		js.Browser.window.addEventListener("beforeunload", event -> {
			storage.setItem(valueKey, haxe.Json.stringify(valueCallback()));
		});
		return haxe.Json.parse(storage.getItem(valueKey)) ?? valueCallback();
		#else
		return valueCallback();
		#end
	}

	public static function resizeObject(sprite:IGameObject):Void {
		Event.emit({
			kind: resizeKind,
			target: sprite,
		});
	}

	static var startTime:Float = 0.0;
	static var lastTapTime:Float = 0.0;

	public static function sendAnalytics(type:String, args:haxe.Rest<Any>):Void {
		if (type == "Start") startTime = App.stamp();
		if (type == "First Interaction") lastTapTime = App.stamp();
		if (type == "All Click") lastTapTime = App.stamp();
		var time = App.stamp() - startTime;
		if (type == "Start") time = startTime;
		final args = args.toArray();
		args.unshift(time);

		#if phaser
		final window = js.Browser.window;
		if (args.length == 0) window.console.log(type);
		else window.console.log(type, args);
		final w:Dynamic = window;
		if (w.sendEvent == null) return;
		w.sendEvent(type, args);
		#end
	}
}
