package khm;

import khm.Ease;
import khm.Lazyr.*;
import khm.ResItem;
import khm.Scene;
import khm.Text;
import khm.Tween;
import khm.Types.Point;
import khm.Types.Rect;
import khm.Types.Size;
import khm.utils.Clipboard;
import khm.utils.DebugRectOverlay;

using khm.tools.PhaserTools;

#if phaser
import phaser.curves.CubicBezier;
import phaser.curves.QuadraticBezier;
#end

private typedef PercentageTweenConfig = {
	targets:Sprite,
	?ease:Ease,
	x:() -> Float,
	y:() -> Float,
	?angle:() -> Float,
	?scale:() -> Float,
	?scaleX:() -> Float,
	duration:Float,
	?delay:Float,
	?onComplete:() -> Void
}

class LazyrTools {
	final scene:Scene;

	public function new(scene:Scene) {
		this.scene = scene;
	}

	public function addShadow(frame:ResItem, ?s:{
		?shadow:Sprite,
		?duration:Int,
		?alpha:Float,
		?delay:Int,
		?onComplete:() -> Void
	}, ?pos:DebugPosInfos):Sprite {
		s ??= {};
		if (s.shadow == null) {
			s.shadow = scene.addSprite(frame, {
				onResize: sprite -> {
					sprite.scaleX = scene.sceneW / sprite.width * 1.5;
					sprite.scaleY = scene.sceneH / sprite.height * 1.5;
					sprite.x = scene.sceneW / 2;
					sprite.y = scene.sceneH / 2;
				},
			}, pos);
			s.shadow.alpha = 0;
			DebugRectOverlay.ignoreObject(s.shadow);
		}
		s.alpha ??= 0.5;
		s.delay ??= 0;
		s.duration ??= 0;
		s.shadow.addTween({
			alpha: s.alpha,
			delay: s.delay,
			duration: s.duration,
			onComplete: s.onComplete
		});
		return s.shadow;
	}

	public function restart():Void {
		isSceneRestarted = true;
		// scene.tweens.pauseAll();
		// scene.anims.pauseAll();
		// scene.anims.removeAllListeners();
		#if phaser
		final offEvents = [
			"update", "render", "postupdate",
			"pointerdown", "pointermove", "pointerup",
			"wheel", "keydown", "keyup", "keypress",
			"resize",
		];
		for (name in offEvents) App._scene.events.off(name);
		App._scene.scene.restart();
		#end
	}

	public function addPercentageTween(s:PercentageTweenConfig):Tween {
		final sprite = s.targets;
		var parent:Point
		& {scale: Float} = {
			x: sprite?.parent.x ?? 0.0,
			y: sprite?.parent.y ?? 0.0,
			scale: sprite?.parent.scaleX ?? 1.0
		};
		// if (sprite.parentContainer() == null) parent = {x: 0, y: 0, scale: 1};
		final startRatioX = (sprite.x + parent.x) / scene.sceneW;
		final startRatioY = (sprite.y + parent.y) / scene.sceneH;
		// final endRatioX = s.x() / scene.sceneW;
		// final endRatioY = s.y() / scene.sceneH;
		final startAngle = sprite.angle;
		final startScale = sprite.scale * parent.scale;
		final startScaleX = sprite.scaleX * parent.scale;
		final ease = s.ease == null ? SineInOut : s.ease;
		final angle = s.angle == null ? () -> sprite.angle : s.angle;
		return Tween.counter({
			from: 0,
			to: 1,
			duration: s.duration,
			delay: s.delay,
			ease: ease,
			onUpdate: tween -> {
				final w = scene.sceneW;
				final h = scene.sceneH;
				final ratio = tween.getValue();
				sprite.x = ratio.lerp(startRatioX * w, s.x());
				sprite.y = ratio.lerp(startRatioY * h, s.y());
				sprite.angle = Std.int(ratio.lerp(startAngle.toFloat(), angle().toFloat()));
				if (s.scale != null) {
					sprite.scale = ratio.lerp(startScale, s.scale());
				}
				if (s.scaleX != null) {
					sprite.scaleX = ratio.lerp(startScaleX, s.scaleX());
				}
			},
			onComplete: s.onComplete
		});
	}

	public function posBg(sprite:Sprite, centerX:Float, centerY:Float) {
		sprite.originX = centerX;
		sprite.originY = centerY;
		sprite.x = scene.sceneW * centerX;
		sprite.y = scene.sceneH * centerY;
		// sprite.x -= sprite.width * (0.5 - centerX);
		// sprite.y -= sprite.height * (0.5 - centerY);
		final tl = sprite.topLeft();
		final left = tl.x;
		final right = scene.sceneW - (tl.x + sprite.width);
		final top = tl.y;
		final bottom = scene.sceneH - (tl.y + sprite.height);
		sprite.displayWidth = sprite.width + left + right;
		sprite.displayHeight = sprite.height + top + bottom;
		final maxScale = Math.max(sprite.scaleX, sprite.scaleY);
		sprite.scale = maxScale;
	}

	#if phaser
	public function draggableBezier(curve:QuadraticBezier, ?parent:Container):Void {
		#if debug
		parent ??= scene;
		final g:Dynamic = App._scene.add.graphics(); // todo port curves
		parent.add(g);
		g.clear();
		g.lineStyle(3, 0xFF0000, 1);
		curve.draw(g);
		g.fillStyle(0xffff00, 1);

		for (point in [curve.p0, curve.p1, curve.p2]) {
			var pointCircle = parent.addSprite({frame: "shadow", texture: null}, {
				props: {
					x: point.x,
					y: point.y,
				}
			});
			pointCircle.scale = 3;
			// pointCircle.setData("_curvePoint", point);

			final obj = pointCircle;
			obj.setInteractive();
			App._scene.input.setDraggable(obj.phaser);
			App._scene.input.on("drag", (pointer, sprite:Sprite, dragX, dragY) -> {
				if (sprite != obj) return;
				// final point = pointCircle.state._curvePoint;
				point.x = dragX;
				point.y = dragY;
				sprite.x = dragX;
				sprite.y = dragY;
				trace('${Math.round(sprite.x)}, ${Math.round(sprite.y)}');
				g.clear();
				g.lineStyle(3, 0xFF0000, 1);
				curve.draw(g);
				g.fillStyle(0xffff00, 1);
			});
			obj.on(PointerUp, e -> {
				final sprite:Sprite = cast obj;
				var x = sprite.x;
				var y = sprite.y;
				var offX = 0.0;
				var offY = 0.0;
				final parent = sprite.parentContainer();
				if (parent != null) {
					// updateContainerSize(parent);
					if (parent.state.debug.offX != null) {
						offX = parent.state.debug.offX;
						offY = parent.state.debug.offY;
					}
				}
				x = Math.round(x - offX);
				y = Math.round(y - offY);
				final text = '\ncast {x: ${x}, y: ${y}}';
				final text = point == curve.p2 ? text : '$text,';
				Clipboard.copy(text);
				trace('$x, $y');
			});
		}
		#end
	}
	#end

	public function addInputTimeout(timeout:Float, callback:() -> Void) {
		final inputTimeoutC = scene.addContainer({
			onResize: sprite -> {
				sprite.x = scene.sceneW / 2;
				sprite.y = scene.sceneW / 2;
				sprite.width = scene.sceneW;
				sprite.height = scene.sceneH;
			}
		});
		inputTimeoutC.name = "inputTimeoutC";
		DebugRectOverlay.ignoreObject(inputTimeoutC);
		inputTimeoutC.setDepth(100);
		var loseTimeout = scene.delayed(timeout, callback);
		inputTimeoutC.on(PointerDown, e -> {
			e.propagate = true;
			loseTimeout?.stop();
			if (timeout > 0) loseTimeout = scene.delayed(timeout, callback);
		});
	}

	public function turnOverSprite(
		sprite:Sprite, frame:ResItem, ?callback:() -> Void, delay = 0
	):Void {
		final scaleX = sprite.scaleX;
		sprite.addTween({
			scaleX: 0,
			duration: 200,
			delay: delay,
			onComplete: () -> {
				sprite.setFrame(frame);
				sprite.scaleX = 0;
				sprite.addTween({
					scaleX: scaleX,
					ease: SineOut,
					duration: 200,
					onComplete: callback
				});
			}
		});
	}

	public function wrongSelectAnim(card:Sprite, angle = 15):Void {
		function rotationTween(angle:Int, ?onComplete:() -> Void):Void {
			card.addTween({
				duration: 200,
				angle: card.angle + angle,
				onComplete: onComplete
			});
		}
		final inProgress = card.state.wrongAnim != null;
		if (inProgress) return;
		card.state.wrongAnim = true;
		final angle = card.angle > 0 ? -angle : angle;
		rotationTween(angle, () -> {
			rotationTween(-angle * 2, () -> {
				rotationTween(angle, () -> {
					card.state.wrongAnim = null;
				});
			});
		});
	}

	public function sortCardsHandTween(cards:Array<Sprite>):Void {
		for (i => card in cards) {
			final w = card.displayWidth;
			final x = w * i - w * (cards.length - 1) / 2;
			card.addTween({
				duration: 500,
				delay: 100,
				x: x,
			});
		}
	}

	public function updateHpLine(line:Sprite, value:Float, max:Float, duration = 500.0):Void {
		if (value < 0) value = 0;
		if (value > max) value = max;
		// if (!Std.isOfType(line.state, Float)) throw "line.state should be Int or Float";
		Tween.counter({
			from: line.state.value,
			to: value,
			duration: duration,
			onUpdate: t -> {
				final v = t.getValue();
				final h = line.width;
				line.setCrop(0, 0, h / max * v, line.height, false);
			}
		});
		line.state.value = value;
	}

	public function countTextAnim(count:Text, to:Int, ?textFormat:(n:Int) -> String):Void {
		final from = count.state.value ?? throw "No text.state.value";
		textFormat ??= n -> '$n';
		count.state._countTweens ??= [];
		final tweens:Array<Tween> = count.state._countTweens;
		for (tween in tweens) tween.end();
		final t1 = count.addTween({
			duration: 500,
			scale: 1.2,
			yoyo: true
		});
		final t2 = Tween.counter({
			targets: count,
			duration: 1000,
			from: from,
			to: to,
			onUpdate: t -> {
				final n = Std.int(t.getValue());
				count.setText(textFormat(n));
			}
		});
		count.state.value = to;
		count.state._countTweens = [t1, t2];
	}

	public function floatCounterAnim(counter:Text, to:Float, scale = 1.2, ?textFormat:(n:Float) -> String):Void {
		final from = counter.state.value ?? throw "No text.state.value";
		textFormat ??= n -> '$n'.substr(0, 4);
		counter.state._countTweens ??= [];
		final tweens:Array<Tween> = counter.state._countTweens;
		for (tween in tweens) tween.end();
		final t1 = counter.addTween({
			duration: 500,
			scale: scale,
			yoyo: true
		});
		final t2 = Tween.counter({
			targets: counter,
			duration: 1000,
			from: from,
			to: to,
			onUpdate: t -> {
				final n = t.getValue();
				counter.setText(textFormat(n));
			}
		});
		counter.state.value = to;
		counter.state._countTweens = [t1, t2];
	}

	public function formatTimer(n:Int, zeroPadMinutes = false):String {
		final minutes = Std.int(n / 60);
		final seconds = n - minutes * 60;
		final m = '$minutes'.lpad("0", zeroPadMinutes ? 2 : 1);
		final s = '$seconds'.lpad("0", 2);
		return '$m:$s';
	}

	public function safeZone(sprite:Any, zone:Int) {
		final sprite:Sprite = cast sprite;
		if (sprite.parentContainer() != null) {
			throw 'safeZone cannot be used on sprite with parentContainer() (frame is "${sprite.id.frame}")';
		}
		final dw = sprite.displayWidth;
		final dh = sprite.displayHeight;
		if (sprite.x < zone + dw / 2) sprite.x = zone + dw / 2;
		if (sprite.y < zone + dh / 2) sprite.y = zone + dh / 2;
		if (sprite.x > scene.sceneW - zone - dw / 2) sprite.x = scene.sceneW - zone - dw / 2;
		if (sprite.y > scene.sceneH - zone - dh / 2) sprite.y = scene.sceneH - zone - dh / 2;
	}

	public function scaleSprite(sprite:Sprite, s:{
		?padding:Float,
		portrait:Size,
		?landscape:Size,
	}):Void {
		#if debug
		if (sprite.parentContainer() != null) {
			throw 'scaleSprite cannot be used on sprite with parentContainer() (frame is "${sprite.id.frame}")';
		}
		#end
		final paddingRatio = s.padding == null ? 0 : s.padding;
		var size = scene.isPortrait ? s.portrait : s.landscape;
		if (size == null) size = s.portrait;
		final scaleX = scene.sceneW / sprite.width * (size.w - paddingRatio * 2);
		final scaleY = scene.sceneH / sprite.height * (size.h - paddingRatio * 2);
		sprite.scale = Math.min(scaleY, scaleX);
	}

	public function moveContainerOrigin(sprite:Container, x:Float, y:Float):Void {
		for (obj in sprite.children) {
			// final obj:Sprite = cast obj;
			obj.x += obj.width * (0.5 - x);
			obj.y += obj.height * (0.5 - y);
		}
	}

	public function moveSpriteOrigin(sprite:Sprite, x:Float, y:Float):Void {
		sprite.x -= sprite.width * (0.5 - x);
		sprite.y -= sprite.height * (0.5 - y);
		sprite.setOrigin(x, y);
	}

	public function centerObjectsRow(objects:Array<IGameObject>, ?objW:Float, ?length:Int, offX = 0.0):Void {
		length ??= objects.length;
		for (i => obj in objects) {
			objW ??= obj.width * 1.1;
			final x = objW * i - objW * (length - 1) / 2;
			obj.addTween({
				duration: 200,
				x: x + offX,
			});
		}
	}
}
