package khm;

import khm.Color;
import khm.Types.Point;
import khm.math.Degrees;

using khm.tools.MathTools;

class MathLib {
	public static final dist = Distance;
	public static final angle = Angle;

	/** Linear interpolation between values. `lerp(0, 100, 0.5)` will return `50`. **/
	public static inline function lerp(a:Float, b:Float, ratio:Float):Float {
		return a + ratio * (b - a);
	}

	/** Normalize `value` between `a` and `b`. `unlerp(0, 100, 50)` will return `0.5`. **/
	public static inline function unlerp(a:Float, b:Float, value:Float):Float {
		return (value - a) / (b - a);
	}

	/** Unlerp `value` from `startA/endA` to `startB/endB` ratio. **/
	public static inline function linearRemap(value:Float, startA:Float, endA:Float, startB:Float, endB:Float):Float {
		return lerp(startB, endB, unlerp(startA, endA, value));
	}

	/**
		Linear interpolation between two colors (ARGB).
	**/
	public static function colorLerp(c1:Color, c2:Color, ratio:Float):Color {
		if (ratio < 0 || ratio > 1) throw 'ratio $ratio is out of 0-1 range';
		final c1:Int = c1;
		final c2:Int = c2;
		final a1 = c1 >>> 24;
		final r1 = (c1 >> 16) & 0xFF;
		final g1 = (c1 >> 8) & 0xFF;
		final b1 = c1 & 0xFF;
		final a2 = c2 >>> 24;
		final r2 = (c2 >> 16) & 0xFF;
		final g2 = (c2 >> 8) & 0xFF;
		final b2 = c2 & 0xFF;
		final a = Std.int(a1 * (1 - ratio) + a2 * ratio);
		final r = Std.int(r1 * (1 - ratio) + r2 * ratio);
		final g = Std.int(g1 * (1 - ratio) + g2 * ratio);
		final b = Std.int(b1 * (1 - ratio) + b2 * ratio);
		return Color.fromValue((a << 24) | (r << 16) | (g << 8) | b);
	}

	public static function multiColorLerp(colors:Array<Color>, ratio:Float):Color {
		final len = colors.length;
		if (len == 0) throw 'No colors provided for interpolation';
		if (len == 1) return colors[0];

		final segmentLength = 1 / (len - 1);
		// Determine which two colors to interpolate between
		final index = Std.int(ratio / segmentLength);
		// Ensure we don't go out of bounds
		final nextIndex = Std.int(Math.min(index + 1, len - 1));
		// Calculate the local ratio between the two colors
		final localRatio = (ratio - index * segmentLength) / segmentLength;

		return colorLerp(colors[index], colors[nextIndex], localRatio);
	}

	/** A Smoother Step interpolation method. **/
	public static function smoothStep(x:Float, min:Float, max:Float):Float {
		if (x <= min) return 0;
		if (x >= max) return 1;
		x = (x - min) / (max - min);
		return x * x * (3 - 2 * x);
	}

	/** Positive modulo of `a` (if `b > 0`) **/
	public static inline function mod(a:Int, b:Int):Int {
		return (a % b + b) % b;
	}
}

private class Distance {
	public static function between(x1:Float, y1:Float, x2:Float, y2:Float):Float {
		final dx = x1 - x2;
		final dy = y1 - y2;
		return Math.sqrt(dx * dx + dy * dy);
	}

	public static inline function betweenPoints(a:Point, b:Point):Float {
		final dx = a.x - b.x;
		final dy = a.y - b.y;
		return Math.sqrt(dx * dx + dy * dy);
	}

	/**
		Calculate the snake distance between two sets of coordinates (points).
		Snake distance (rectilinear distance, Manhattan distance) is the sum of the horizontal and vertical distances.
		It's the effective distance when movement is allowed only horizontally or vertically (but not both).
	**/
	public static function snake(x1:Float, y1:Float, x2:Float, y2:Float):Float {
		return Math.abs(x1 - x2) + Math.abs(y1 - y2);
	}

	public static inline function snakePoints(a:Point, b:Point):Float {
		return Math.abs(a.x - b.x) + Math.abs(a.y - b.y);
	}

	public static function squared(x1:Float, y1:Float, x2:Float, y2:Float):Float {
		final dx = x1 - x2;
		final dy = y1 - y2;
		return dx * dx + dy * dy;
	}

	public static inline function squaredPoints(a:Point, b:Point):Float {
		final dx = a.x - b.x;
		final dy = a.y - b.y;
		return dx * dx + dy * dy;
	}

	public static function betweenAngles(ang:Degrees, toAng:Degrees):Degrees {
		var a = toAng - ang;
		if (a < -180) a += 360;
		if (a > 180) a -= 360;
		return a;
	}
}

private class Angle {
	public static function between(x1:Float, y1:Float, x2:Float, y2:Float):Degrees {
		return Math.atan2(y2 - y1, x2 - x1).toDeg();
	}

	public static inline function betweenPoints(a:Point, b:Point):Degrees {
		final dx = b.x - a.x;
		final dy = b.y - a.y;
		return Math.atan2(dy, dx).toDeg();
	}

	/** Shortest distance between two angles, can be negative to show shortest direction. **/
	public static function dist(a:Degrees, b:Degrees):Degrees {
		return clamp(a - b);
	}

	public static function clamp(a:Degrees):Degrees {
		a %= 360;
		if (a > 180) a -= 360;
		else if (a <= -180) a += 360;
		// a %= Math.PI * 2;
		// if( a > Math.PI ) a -= 2 * Math.PI;
		// else if( a <= -Math.PI ) a += 2 * Math.PI;
		return a;
	}
}
