package khm;

import khm.Event.ScenesChangeEvent;
import khm.Lazyr.TimerEventConfig;
import khm.Pointer;
import khm.internal.BaseObjectUtils;
import khm.math.FastMatrix3;

class Scene extends Container {
	public final camera:Camera;
	/** Enable pointer events outside of scene rect. **/
	public var isUnlimitedSize = false;

	/** Same as `this.width` (that is equals to `App.w` in default `onResize`). **/
	public var sceneW(get, never):Int;

	@:noCompletion function get_sceneW():Int {
		return Std.int(width);
	}

	/** Same as `this.height` (that is equals to `App.h` in default `onResize`). **/
	public var sceneH(get, never):Int;

	@:noCompletion function get_sceneH():Int {
		return Std.int(height);
	}

	/** Tweaks `isPortrait` property with `sceneH * portraitRatio > sceneW` result. **/
	public var portraitRatio = 1 / 1;
	/** If this scene `height` is bigger than `width` **/
	public var isPortrait(get, never):Bool;

	@:noCompletion function get_isPortrait():Bool {
		return sceneH * portraitRatio > sceneW;
	}

	public function new() {
		super();
		type = SceneType;
		originX = 0;
		originY = 0;
		width = App.w;
		height = App.h;
		camera = new Camera(this);
	}

	/** Adds scene to `App.scenes`, adds `onUpdate` and `onResize` event listeners, and then calls `onShow`. **/
	public function show(hideOthers = false):Scene {
		if (hideOthers) {
			var i = App.scenes.length;
			while (i-- > 0) {
				final scene = App.scenes[i];
				scene.hide();
			}
		}

		App.showScene(this);

		on(Update, e -> onUpdate());
		on(Resize, e -> onResize());
		onShow();
		Event.emitResize(this);
		return this;
	}

	/** Removes scene from `App.scenes`, removes all events from `this` scene, and then calls `onHide`. **/
	public function hide():Void {
		App.hideScene(this);

		off();
		onHide();
	}

	override function updateTransformation():Void {
		transformation.setFrom(FastMatrix3.identity());
		transformation.setFrom(
			transformation.multmat(FastMatrix3.scale(camera.scaleX, camera.scaleY))
		);
		transformation.setFrom(
			transformation.multmat(FastMatrix3.translation(
				Std.int(-camera.x + camera.width * camera.originX),
				Std.int(-camera.y + camera.height * camera.originY)
			))
		);
		// transformation.setFrom(
		// 	transformation.multmat(FastMatrix3.translation(view.x, view.y))
		// );
		// transformation.setFrom(transformation.multmat(FastMatrix3.scale(obj.scaleX, obj.scaleY)));
		// transformation.setFrom(transformation.multmat(parent.transformation));
		BaseObjectUtils.apply2DTransformation(transformation, this);
	}

	#if kha
	/**
		Default scene render loop, based on `kha.graphics2`.

		Feel free to override, copy default implementation from this method and draw other things with Kha!
	**/
	override function render(g:kha.graphics2.Graphics) {
		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);
		for (obj in children)
			obj.render(g);
		popContainerSpace(g);
	}
	#end

	/** Runned after `scene.show(), empty by default.` **/
	function onShow():Void {}

	/** Runned after `scene.hide(), empty by default.` **/
	function onHide():Void {}

	/**
		Default scene `Resize` event handler, sets scene size to screen size.
		Can be overriden, or removed with `off(Resize, onResize)` and applied back with `on(Resize, onResize)`.
	**/
	public function onResize():Void {
		width = App.w;
		height = App.h;
		// if (camera.originX == 0.5) x = -width / 2;
		// if (camera.originY == 0.5) y = -height / 2;
	}

	/**
		Default scene update cycle, just like `onResize`, but added with `this.on(Update, onUpdate)` in `scene.show()`.

		Always runs in static 60 ticks per second.
	**/
	public function onUpdate():Void {}

	/** Runned after mouse leaved canvas (not scene, use `on(PointerOut, ...)` for it). **/
	public function onMouseLeave():Void {}

	/** Moves scene at top of `App.scenes` and fires `ScenesChange` event. **/
	public function bringToTop(withEvent = false):Void {
		final has = App.scenes.remove(this);
		if (!has) throw "scene not found in list to be moved on top";
		App.scenes.push(this);
		if (withEvent) {
			final e:ScenesChangeEvent = {
				kind: ScenesChange,
				target: null,
				scene: this,
				type: Moved
			};
			Event.emit(e);
		}
		Event.sortObjectEvents(this);
	}

	/** Is scene attached to `App.scenes` array for render and events. **/
	public var isAttached(get, never):Bool;

	@:noCompletion inline function get_isAttached():Bool {
		return App.scenes.contains(this);
	}

	/** (only for debug mode) Key callback shortcut for testing something on key press. **/
	public function onDebugKey(key:khm.KeyCode, callback:() ->
		Void, ?pos:DebugPosInfos):Void {
		#if debug
		on(KeyDown, e -> {
			if (e.key == key) callback();
			else e.propagate = true;
		}, pos);
		#end
	}

	/** Creates looped timer tween. **/
	public function addTimeEvent(config:TimerEventConfig, ?pos:DebugPosInfos):Tween {
		// return cast App._scene.time.addEvent(config);
		config.startRatio ??= 0.0;
		config.delay ??= 1000;
		final tween = Tween.counter({
			targets: this,
			from: 0,
			to: 1,
			paused: config.paused,
			// duration: config.delay,
			duration: 0,
			delay: config.delay * (1 - config.startRatio),
			repeatDelay: config.delay,
			repeat: config.repeat ?? -1,
			onRepeat: () -> config.callback(),
			onComplete: () -> config.callback()
		}, pos);
		return tween;
	}
}
