package khm;

import khm.audio.AudioEffect;
#if kha
import kha.Assets;
import kha.audio1.Audio as Audio1;
import kha.audio1.AudioChannel;
// import kha.audio2.Audio1;
import khm.audio.KhaSound;
import khm.audio.effects.Filter;
#elseif phaser
import phaser.sound.WebAudioSound;
#end

/**
	Marked section of a sound represented by name, and optionally start time, duration, and config object.
**/
typedef SoundMarker = {
	/** `name` - Unique identifier of a sound marker.**/
	var name:String;
	/** `[start=0]` - Sound position offset at witch playback should start.**/
	var ?start:Float;
	/** `[duration]` - Playback duration of this marker.**/
	var ?duration:Float;
	/** `[config]` - An optional config object containing default marker settings.**/
	var ?config:SoundConfig;
}

@:structInit
class SoundConfig {
	public var sprite:String = null;
	/** `Indicates whether the sound should be muted or not. **/
	public var mute = false;
	/** A value between 0 (silence) and 1 (full volume). **/
	public var volume = 1.0;
	/** Defines the speed at which the sound should be played. **/
	public var rate = 1.0;
	/** Represents detuning of sound in [cents](https://en.wikipedia.org/wiki/Cent_%28music%29). **/
	public var detune = 0.0;
	/** Position of playback for this sound, in seconds. **/
	public var seek = 0.0;
	/** Whether or not the sound or current sound marker should loop. **/
	public var loop = false;
	/** Time, in seconds, that should elapse before the sound actually starts its playback. **/
	public var delay = 0.0;
	/** A value between -1 (full left pan) and 1 (full right pan). 0 means no pan. **/
	public var pan = 0.0;
}

@:keep
class Sound {
	#if kha
	final sound:KhaSound;
	#elseif phaser
	final sound:WebAudioSound;
	#end
	final pos:Null<DebugPosInfos>;

	var key:String;

	/**
	 * A property that holds the value of sound's actual playback rate,
	 * after its rate and detune values has been combined with global
	 * rate and detune values.
	 *
	 * @default 1
	 */
	// public var totalRate:Float;
	/** A value representing the duration, in seconds. **/
	public var duration(get, never):Float;

	inline function get_duration():Float {
		return sound.duration;
	}

	/**
	 * Object containing markers definitions.
	 *
	 * @type {Object.<String, SoundMarker>}
	 * @default {}
	 */
	// public var markers:Dynamic;
	/**
	 * Currently playing marker.
	 * 'null' if whole sound is playing.
	 *
	 * @type {SoundMarker}
	 * @default null
	 */
	// public var currentMarker:SoundMarker;
	public var pan(get, set):Float;

	inline function get_pan():Float {
		return sound.pan;
	}

	inline function set_pan(pan:Float):Float {
		return sound.pan = pan;
	}

	public var volume(get, set):Float;

	inline function get_volume():Float {
		return sound.volume;
	}

	inline function set_volume(volume:Float):Float {
		return sound.volume = volume;
	}

	public var time(get, set):Float;

	inline function get_time():Float {
		#if kha
		return sound.time;
		#elseif phaser
		return sound.seek;
		#end
	}

	inline function set_time(time:Float):Float {
		#if kha
		return sound.time = time;
		#elseif phaser
		return sound.seek = time;
		#end
	}

	public var rate(get, set):Float;

	inline function get_rate():Float {
		#if kha
		return sound.pitch;
		#elseif phaser
		return sound.rate;
		#end
	}

	inline function set_rate(rate:Float):Float {
		#if kha
		return sound.pitch = rate;
		#elseif phaser
		return sound.rate = rate;
		#end
	}

	public var loop(get, set):Bool;

	inline function get_loop():Bool {
		return sound.loop;
	}

	inline function set_loop(loop:Bool):Bool {
		return sound.loop = loop;
	}

	public var isPaused(get, never):Bool;

	inline function get_isPaused():Bool {
		return sound.isPaused;
	}

	public var isFinished(get, never):Bool;

	inline function get_isFinished():Bool {
		return sound.finished;
	}

	public function new(key:String, ?config:SoundConfig, ?pos:DebugPosInfos) {
		this.key = key;
		this.pos = pos;
		config ??= {};
		#if kha
		sound = new KhaSound(key);
		sound.loop = config.loop;
		sound.pitch = config.rate;
		sound.pan = config.pan;
		sound.time = config.seek;
		sound.volume = config.volume;
		if (config.detune != 0) trace("detune not implemented");
		#elseif phaser
		sound = App._scene.sound.asType2.addAudioSprite(key, cast config).asType2;
		#end
	}

	public function addEffect<T:AudioEffect>(effect:T):T {
		#if kha
		sound.addEffect(effect);
		#elseif phaser
		trace("audio effects not available for phaser backend");
		#end
		return effect;
	}

	public function removeEffect(effect:AudioEffect):Void {
		#if kha
		sound.removeEffect(effect);
		#end
	}

	public function removeAllEffects():Void {
		#if kha
		sound.removeAllEffects();
		#end
	}

	/**
	 * Adds a marker into the current sound. A marker is represented by name, start time, duration, and optionally config object.
	 * This allows you to bundle multiple sounds together into a single audio file and use markers to jump between them for playback.
	 *
	 * @param {SoundMarker} marker - Marker object.
	 *
	 * @return {boolean} Whether the marker was added successfully.
	 */
	public function addMarker(marker:SoundMarker):Bool {
		throw "not impl";
	}

	/**
	 * Updates previously added marker.
	 *
	 * @param {SoundMarker} marker - Marker object with updated values.
	 *
	 * @return {boolean} Whether the marker was updated successfully.
	 */
	public function updateMarker(marker:SoundMarker):Bool {
		throw "not impl";
	}

	/**
	 * Removes a marker from the sound.
	 *
	 * @param {string} markerName - The name of the marker to remove.
	 *
	 * @return {?SoundMarker} Removed marker object or 'null' if there was no marker with provided name.
	 */
	public function removeMarker(markerName:String):SoundMarker {
		throw "not impl";
	}

	/**
	 * Play this sound, or a marked section of it.
	 * It always plays the sound from the start. If you want to start playback from a specific time
	 * you can set 'seek' setting of the config object, provided to this call, to that value.
	 *
	 * @param {(string|SoundConfig)} [markerName=''] - If you want to play a marker then provide the marker name here. Alternatively, this parameter can be a SoundConfig object.
	 * @param {SoundConfig} [config] - Optional sound config object to be applied to this marker or entire sound if no marker name is provided. It gets memorized for future plays of current section of the sound.
	 */
	public function play(?markerName:String):Void {
		#if kha
		sound.play();
		#elseif phaser
		sound.play(markerName);
		#end
	}

	public function pause():Void {
		sound.pause();
	}

	public function resume():Void {
		sound.resume();
	}

	public function stop():Void {
		sound.stop();
	}
}
