package khm;

import haxe.extern.EitherType as Or;
import khm.Event;
import khm.IGameObject.GameObjectType;
import khm.Tween.TweenConfig;
import khm.Types.Point;
import khm.internal.BaseObjectUtils;
import khm.math.Degrees;
import khm.math.FastMatrix3;
import khm.text.CanvasText;
#if kha
import kha.graphics2.Graphics;
import kha.graphics2.ImageScaleQuality;
#elseif phaser
import phaser.gameobjects.Text as PText;
#end

enum abstract TextAlign(String) from String {
	var Left = "left";
	var Right = "right";
	var Center = "center";
	var Justify = "justify";
}

typedef TextMetrics = {
	var ascent:Float;
	var descent:Float;
	var fontSize:Float;
}

/** A Text Style configuration object as used by the Text Game Object. **/
@:structInit
@:publicFields
class TextStyle {
	/** `[fontFamily=Text.defaultFontFamily]` - The font the Text object will render with. This is a Canvas style font string. **/
	var fontFamily:String = Text.defaultFontFamily;
	/** `[fontSize=16]` - The font size, as Float in pixels. **/
	var fontSize:Float = 16;
	/** `[fontStyle]` - Any addition font styles, such as 'strong'. **/
	// var ?fontStyle:String;
	/** `[font]` - The font family or font settings to set. Overrides the other font settings. **/
	// var ?font:String;
	/** `[backgroundColor]` - A solid fill color that is rendered behind the Text object. Can be CSS string color such as `#ff0`. **/
	var backgroundColor:Null<Color> = null;
	/** `[color=0xFFFFFFFF]` - The color the Text is drawn in. Can be CSS string color such as `#fff` or `rgb()`. **/
	var color:Color = 0xFFFFFFFF;
	/** `[verGradient]` - Gradient color the Text is drawn in. Can be array of `Color` or of objects`. **/
	var verGradient:Array<Or<Color, {pos:Float, color:Color}>> = null;
	/** `[stroke=0xFFFFFFFF]` - The color used to stroke the Text if the `strokeThickness` property is greater than zero. **/
	var stroke:Color = 0xFFFFFFFF;
	/** `[strokeThickness=0]` - The thickness of the stroke around the Text. Set to zero for no stroke. **/
	var strokeThickness:Float = 0.0;
	/** `[shadow]` - The Text shadow configuration object. **/
	var shadow:TextShadow = null;
	/** `[padding]` - A Text Padding object. **/
	// var ?padding:TextPadding;
	/** `[align='left']` - The alignment of the Text. This only impacts multi-line text. Either `left`, `right`, `center` or `justify`. **/
	var align:TextAlign = Left;
	/** `[maxLines=0]` - The maximum number of lines to display within the Text object. **/
	var maxLines:Int = 0;
	/** `[fixedWidth=0]` - Force the Text object to have the exact width specified in this property. Leave as zero for it to change accordingly to content. **/
	var fixedWidth:Float = 0.0;
	/** `[fixedHeight=0]` - Force the Text object to have the exact height specified in this property. Leave as zero for it to change accordingly to content. **/
	var fixedHeight:Float = 0.0;
	/** `[resolution=0]` - Sets the resolution (DPI setting) of the Text object. Leave at zero for it to use the game resolution. **/
	// var ?resolution:Float;
	/** `[rtl=false]` - Set to `true` if this Text object should render from right-to-left. **/
	var rtl:Bool = false;
	/** `[testString='|MÃ‰qgy']` - This is the string used to aid Canvas in calculating the height of the font. **/
	var testString:String = "|MÃ‰qgy";
	/** `[baselineX=1.2]` - The amount of horizontal padding added to the width of the text when calculating the font metrics. **/
	// var ?baselineX:Float;
	/** `[baselineY=1.4]` - The amount of vertical padding added to the height of the text when calculating the font metrics. **/
	// var ?baselineY:Float;
	/** `[wordWrap]` - The Text Word wrap configuration object. **/
	var wordWrap:{
		/** The width at which text should be considered for word-wrapping. **/
		var width:Float;
		@:noCompletion var ?callback:() -> Void;
	} = null;
	/** `[metrics]` - A Text Metrics object. Use this to avoid expensive font size calculations in text heavy games. **/
	var metrics:TextMetrics = null;

	public function init():Void {
		#if phaser
		// if (fontSize is String) {
		// 	fontSize = Std.parseFloat(fontSize.replace("px", ""));
		// 	if (Math.isNaN(fontSize)) throw "fontSize cannot be parsed: " + fontSize;
		// }
		#end
	}

	public function setStroke(color:Color, thickness:Float):Void {
		this.stroke = color;
		this.strokeThickness = thickness;
	}

	// public function update():Void {}
	public var wordWrapWidth(get, never):Null<Float>;

	function get_wordWrapWidth():Null<Float> {
		return wordWrap?.width;
	}

	public function clone():TextStyle {
		return {
			fontFamily: this.fontFamily,
			fontSize: this.fontSize,
			backgroundColor: this.backgroundColor,
			color: this.color,
			verGradient: this.verGradient,
			stroke: this.stroke,
			strokeThickness: this.strokeThickness,
			shadow: this.shadow,
			align: this.align,
			maxLines: this.maxLines,
			fixedWidth: this.fixedWidth,
			fixedHeight: this.fixedHeight,
			rtl: this.rtl,
			testString: this.testString,
			wordWrap: this.wordWrap,
			metrics: this.metrics,
		};
	}

	public function isEqualTo(style:TextStyle):Bool {
		return this.fontFamily == style.fontFamily
			&& this.fontSize == style.fontSize
			&& this.backgroundColor == style.backgroundColor
			&& this.color == style.color
			&& this.verGradient?.length == style.verGradient?.length
			&& this.stroke == style.stroke
			&& this.strokeThickness == style.strokeThickness
			&& this.shadow?.color == style.shadow?.color
			&& this.align == style.align
			&& this.maxLines == style.maxLines
			&& this.fixedWidth == style.fixedWidth
			&& this.fixedHeight == style.fixedHeight
			&& this.rtl == style.rtl
			&& this.wordWrap?.width == style.wordWrap?.width;
		// && this.metrics == style.metrics;
	}
}

/** A Text Shadow configuration object as used by the Text Style. **/
typedef TextShadow = {
	/** `[offsetX=0]` - The horizontal offset of the shadow. **/
	var ?offsetX:Float;
	/** `[offsetY=0]` - The vertical offset of the shadow. **/
	var ?offsetY:Float;
	/** `[color=0x000]` - The color of the shadow, given as Int or a CSS string value. **/
	var ?color:Color;
	/** `[blur=0]` - The amount of blur applied to the shadow. Leave as zero for a hard shadow. **/
	var ?blur:Float;
	/** `[stroke=false]` - Apply the shadow to the stroke effect on the Text object? **/
	var ?stroke:Bool;
	/** `[fill=false]` - Apply the shadow to the fill effect on the Text object? **/
	var ?fill:Bool;
}

class Text implements IGameObject {
	public static var defaultFontFamily = "Courier";

	#if kha_html5
	public var phaser:CanvasText;
	#elseif phaser
	public var phaser:PText;
	#end

	public var isInteractive(default, set) = true;
	public var transformation(get, default) = FastMatrix3.identity();

	@:isVar public var width(get, set) = 0.0;
	@:isVar public var height(get, set) = 0.0;

	inline function get_width():Float {
		return this.width;
	}

	inline function set_width(width:Float):Float {
		return this.width = width;
	}

	inline function get_height():Float {
		return this.height;
	}

	inline function set_height(height:Float):Float {
		return this.height = height;
	}

	public var x(default, set) = 0.0;
	public var y(default, set) = 0.0;
	public var parent:IContainer = null;
	public var alpha(default, set) = 1.0;
	public var angle(default, set):Degrees = 0.0;
	public var visible(default, set) = true;
	public var scaleX(default, set) = 1.0;
	public var scaleY(default, set) = 1.0;
	public var originX(default, set) = 0.5;
	public var originY(default, set) = 0.5;
	public var tint:Color = 0xFFFFFFFF;
	public var state(default, never):ObjectState;
	public var type:GameObjectType = TextType;
	public var scale(get, set):Float;
	public final style:TextStyle;

	public var text(default, set):String;

	public var onTextUpdate:(text:String) -> String = null;

	function set_text(text) {
		setText('$text');
		return text;
	}

	@:keep function get_scale():Float {
		return (scaleX + scaleY) / 2;
	}

	@:keep function set_scale(scale:Float):Float {
		scaleX = scale;
		scaleY = scale;
		return scale;
	}

	public function new(text:String, ?style:TextStyle):Void {
		@:bypassAccessor this.text = text;
		style ??= {};
		this.style = style;
		style.fontFamily ??= Text.defaultFontFamily;
		#if kha_html5
		phaser = new CanvasText(text, style);
		width = Math.ceil(phaser.width);
		height = Math.ceil(phaser.height);
		applyVerGradient(style);
		#elseif phaser
		final s = style.clone();
		if (s.fontSize is Int) s.fontSize = cast '${s.fontSize}px';
		if (isColorType(s.backgroundColor))
			s.backgroundColor = cast s.backgroundColor.toRgbaCode();
		if (isColorType(s.color)) s.color = cast s.color.toRgbaCode();
		if (isColorType(s.stroke)) s.stroke = cast s.stroke.toRgbaCode();
		if (s.shadow != null) {
			if (isColorType(s.shadow.color)) s.shadow.color = cast s.shadow.color.toRgbaCode();
		}
		s.shadow ??= {};
		s.wordWrap ??= cast {};
		phaser = new PText(App._scene, 0, 0, text, cast s);
		width = Std.int(phaser.width);
		height = Std.int(phaser.height);
		applyVerGradient(s);
		#end
	}

	function applyVerGradient(style:TextStyle):Void {
		#if js
		final s = style;
		if (s.verGradient != null) {
			final gradient = phaser.context.createLinearGradient(0, 0, 0, height);
			for (i => obj in s.verGradient) {
				final isColor = isColorType(obj);
				if (isColor) {
					final color:Color = cast obj;
					var len = s.verGradient.length - 1;
					if (len < 1) len = 1;
					final pos = i / len;
					gradient.addColorStop(pos, color.toRgbCode());
				} else {
					final obj:{pos:Float, color:Color} = cast obj;
					gradient.addColorStop(obj.pos, obj.color.toRgbCode());
				}
			}
			phaser.setFill(gradient);
		}
		#end
	}

	static function isColorType(value:Any):Bool {
		final isColor = value != null && !Reflect.isObject(value);
		if (isColor) {
			final c:Color = value;
			colorAlphaCheck(c);
		}
		return isColor;
	}

	static function colorAlphaCheck(c:Color, ?pos:DebugPosInfos) {
		if (c.A == 0) {
			trace('Color `0x${c.hex(6)}` has zero alpha, you probably forgot to set alpha channel (use 0xFF${c.hex(6)}). Or set `0x01...` alpha instead', pos);
		}
	}

	public function setText(text:String):Void {
		#if js
		if (onTextUpdate != null) text = onTextUpdate(text);
		phaser.setText(text);
		@:bypassAccessor this.text = text;
		width = Std.int(phaser.width);
		height = Std.int(phaser.height);
		#end
	}

	function set_isInteractive(v) {
		isInteractive = v;
		// if (isInteractive) phaser.setInteractive();
		// else phaser.removeInteractive();
		return v;
	}

	function set_x(v) {
		#if kha
		return this.x = v;
		#elseif phaser
		return phaser.x = this.x = v;
		#end
	}

	function set_y(v) {
		#if kha
		return this.y = v;
		#elseif phaser
		return phaser.y = this.y = v;
		#end
	}

	function set_alpha(v) {
		#if kha
		return this.alpha = v;
		#elseif phaser
		return phaser.alpha = this.alpha = v;
		#end
	}

	function set_angle(v) {
		#if kha
		return this.angle = v;
		#elseif phaser
		this.angle = v;
		return phaser.angle = this.angle.toFloat();
		#end
	}

	function set_visible(v) {
		#if kha
		return this.visible = v;
		#elseif phaser
		return phaser.visible = this.visible = v;
		#end
	}

	function set_scaleX(v) {
		#if kha
		return this.scaleX = v;
		#elseif phaser
		return phaser.scaleX = this.scaleX = v;
		#end
	}

	function set_scaleY(v) {
		#if kha
		return this.scaleY = v;
		#elseif phaser
		return phaser.scaleY = this.scaleY = v;
		#end
	}

	public var displayWidth(get, set):Float;
	public var displayHeight(get, set):Float;

	inline function get_displayWidth():Float
		return width * scaleX;

	inline function get_displayHeight():Float
		return height * scaleY;

	inline function set_displayWidth(dw:Float):Float {
		scaleX = dw / width;
		return dw;
	}

	inline function set_displayHeight(dh:Float):Float {
		scaleY = dh / height;
		return dh;
	}

	function set_originX(v) {
		originX = v;
		#if phaser
		phaser.setOrigin(originX, originY);
		#end
		return v;
	}

	function set_originY(v) {
		originY = v;
		#if phaser
		phaser.setOrigin(originX, originY);
		#end
		return v;
	}

	#if kha
	function get_transformation() {
		return transformation;
	}
	#elseif phaser
	var tempMatrix = new phaser.gameobjects.components.TransformMatrix();

	function get_transformation() {
		final m = phaser.getWorldTransformMatrix(tempMatrix);
		transformation._00 = m.a;
		transformation._01 = m.b;
		transformation._10 = m.c;
		transformation._11 = m.d;
		transformation._20 = m.e;
		transformation._21 = m.f;
		return transformation;
	}
	#end

	public function setFill(color:Color):Text {
		colorAlphaCheck(color);
		style.color = color;
		#if kha_html5
		phaser.setFill(color);
		#elseif phaser
		phaser.setFill(color.toRgbaCode());
		#end
		return this;
	}

	public function setStroke(color:Color, size:Float):Text {
		colorAlphaCheck(color);
		style.stroke = color;
		#if kha_html5
		phaser.setStroke(color, size);
		#elseif phaser
		phaser.setStroke(color.toRgbaCode(), size);
		#end
		return this;
	}

	/**
		Set the shadow settings.
		@param offX The horizontal shadow offset.
		@param offY The vertical shadow offset.
		@param color The shadow color.
		@param blur The shadow blur radius.
		@param shadowStroke Whether to stroke the shadow.
		@param shadowFill Whether to fill the shadow.
		@return This Text object.
	**/
	public function setShadow(offX = 0.0, offY = 0.0, color:Color = Black, blur = 0.0, shadowStroke = false, shadowFill = true):Text {
		colorAlphaCheck(color);
		#if kha_html5
		phaser.setShadow(offX, offY, color, blur, shadowStroke, shadowFill);
		#elseif phaser
		phaser.setShadow(offX, offY, color.toRgbaCode(), blur, shadowStroke, shadowFill);
		#end
		return this;
	}

	/**
		Set the alignment of the text in this Text object.
		Alignment only works if the Text object has more than one line of text.
	**/
	public function setAlign(align:TextAlign):Void {
		#if kha_html5
		phaser.setAlign(align);
		#elseif phaser
		phaser.setAlign(cast align);
		#end
	}

	public function destroy():Void {
		#if (phaser || kha_html5)
		phaser.destroy();
		#end
	}

	#if kha
	public function render(g:kha.graphics2.Graphics) {
		#if kha_html5
		if (phaser.texture == null) return;
		if (!visible || alpha == 0) return;
		g.color = 0xFFFFFFFF;
		pushContainerSpace(g);

		final local = this.topLeftLocal();
		g.drawImage(phaser.texture, local.x, local.y);

		popContainerSpace(g);
		#end
	}

	@:noCompletion var prevAlpha:Float;

	// @:noCompletion var prevScaleQuality:ImageScaleQuality;

	function pushContainerSpace(g:Graphics):Void {
		prevAlpha = g.opacity;
		// prevScaleQuality = g.imageScaleQuality;
		g.opacity = g.opacity * alpha;
		// g.imageScaleQuality = scaleQuality;
		updateTransformation();
		g.pushTransformation(transformation);
	}

	function popContainerSpace(g:Graphics):Void {
		// g.imageScaleQuality = prevScaleQuality;
		g.opacity = prevAlpha;
		g.popTransformation();
	}

	public function updateTransformation():Void {
		transformation.setFrom(FastMatrix3.identity());
		transformation.setFrom(transformation.multmat(parent.transformation));
		BaseObjectUtils.apply2DTransformation(transformation, this);
	}
	#end
}
