package khm;

import haxe.extern.EitherType as Or;
import khm.Ease;
import khm.Event.EventKind;
import khm.IGameObject;
import khm.math.Degrees;
import khm.utils.Easing;
import khm.utils.macro.TypedefUtils;

typedef CustomTweenConfig = {
	var ?x:Float;
	var ?y:Float;
	var ?angle:Degrees;
	var ?scale:Float;
	var ?scaleX:Float;
	var ?scaleY:Float;
	var ?alpha:Float;
	var ?displayWidth:Float;
	var ?displayHeight:Float;
	var ?volume:Float;
}

private typedef DefaultTweenConfig = {
	var ?targets:Or<Any, Array<Any>>;
	@:noCompletion var ?from:Float;
	@:noCompletion var ?to:Float;
	var ?duration:Float;
	var ?delay:Float;
	var ?ease:Ease;
	var ?easeCallback:(t:Float) -> Float;
	var ?paused:Bool;
	var ?repeat:Int;
	var ?repeatDelay:Float;
	var ?yoyo:Bool;
	var ?yoyoDelay:Float;
	var ?completeDelay:Float;
	var ?onStart:() -> Void;
	var ?onStop:() -> Void;
	var ?onUpdate:(e:TweenEvent) -> Void;
	var ?onYoyo:() -> Void;
	var ?onRepeat:() -> Void;
	var ?onComplete:() -> Void;
}

typedef TweenConfig = CustomTweenConfig & DefaultTweenConfig;

private enum abstract TweenState(Int) {
	var Start;
	var Delay;
	var Update;
	var YoyoDelay;
	var Yoyo;
	var Repeat;
	var RepeatDelay;
	var CompleteDelay;
	var Complete;
}

@:allow(khm.Tween)
class TweenEvent {
	public var tween(default, null):Tween;
	public final targets:Array<Any>;
	public var ratio(default, null) = 0.0;
	public var timeRatio(default, null) = 0.0;

	public function new(tween:Tween, targets:Array<Any>) {
		this.tween = tween;
		this.targets = targets;
	}

	inline function setRatio(value:Float):Void {
		ratio = value;
	}

	inline function setTimeRatio(value:Float):Void {
		timeRatio = value;
	}

	public inline function getValue():Float {
		if (tween.isCounter) {
			return tween.counterValue;
		}
		// todo? prev counter hack, can be probably removed
		if (tween.fieldTweens.length == 1) {
			final obj = tween.fieldTweens[0];
			return ratio.lerp(obj.start, obj.end);
		}
		return ratio;
	}

	@:noCompletion
	public function destroy():Void {
		tween = null;
		targets.resize(0);
	}
}

@:allow(khm.TweenEvent)
class Tween {
	public static var tweenUpdateEvent = BeforeRender;

	@:keep public var type:GameObjectType = TweenType;
	public var state(default, null):TweenState = Start;

	public var targets:Array<Any>;
	public var duration:Float;
	public var delay:Float;
	public var ease:Ease;
	public var paused:Bool;
	public var repeat:Int;
	public var repeatDelay:Float;
	public var yoyo:Bool;
	public var yoyoDelay:Float;
	public var completeDelay:Float;
	public var onStart:() -> Void = null;
	public var onStop:() -> Void = null;
	public var onUpdate:(e:TweenEvent) -> Void = null;
	public var onYoyo:() -> Void = null;
	public var onRepeat:() -> Void = null;
	public var onComplete:() -> Void = null;

	var easeCallback:(t:Float) -> Float;
	var isCounter = false;
	var from:Float;
	var to:Float;
	var pos = 0.0;

	/**
		Current tween time that increments by `App.deltaTime`
		and used to update between tween states after hitting `delay`'s or `duration`.
	**/
	public var time = 0.0;

	var counterValue = 0.0;
	final fieldTweens:Array<{
		target:Dynamic,
		field:String,
		start:Float,
		end:Float
	}> = [];
	final callbackEvent:TweenEvent;
	var posInfos:DebugPosInfos = null;

	public function new(s:TweenConfig, ?pos:DebugPosInfos) {
		posInfos = pos;
		duration = s.duration ?? 1000.0;
		delay = s.delay ?? 0.0;
		paused = s.paused ?? false;
		ease = s.ease ?? SineInOut;
		easeCallback = s.easeCallback ?? getEasingCallback(ease);
		repeat = s.repeat ?? 0;
		repeatDelay = s.repeatDelay ?? 0.0;
		yoyo = s.yoyo ?? false;
		if (yoyo) duration *= 2;
		yoyoDelay = s.yoyoDelay ?? 0.0;
		completeDelay = s.completeDelay ?? 0.0;
		onStart = s.onStart;
		onStop = s.onStop;
		onUpdate = s.onUpdate;
		onYoyo = s.onYoyo;
		onRepeat = s.onRepeat;
		onComplete = s.onComplete;

		isCounter = s.from != null && s.to != null;
		if (isCounter) {
			s.targets ??= [];
			from = s.from;
			to = s.to;
		} else {
			if (s.targets == null) throw "targets is null";
		}
		if (!(s.targets is Array)) s.targets = [s.targets];
		else {
			final arr:Array<Any> = s.targets;
			s.targets = arr.copy();
		}
		targets = s.targets;

		static final configFields = TypedefUtils.getFields(DefaultTweenConfig);
		final fieldNames = Reflect.fields(s);
		for (field in configFields) {
			if (fieldNames.contains(field)) fieldNames.remove(field);
		}
		callbackEvent = new TweenEvent(this, targets);

		for (field in fieldNames) {
			if (isCounter) break;
			var end = Reflect.getProperty(s, field);
			if (end == null) continue;
			for (target in targets) {
				final start = readTargetStartValue(target, field);
				#if debug
				end = validateEndValue(s, field, start, end, pos);
				#end
				if (field == "a") {
					trace(start, end);
				}
				fieldTweens.push({
					target: target,
					field: field,
					start: start,
					end: end
				});
			}
		}

		Event.on((this : Any), tweenUpdateEvent, update, posInfos);
		if (duration == 0) update({
			target: null,
			kind: tweenUpdateEvent
		});
	}

	public static function counter(s:TweenConfig & {?from:Float, ?to:Float}, ?pos:DebugPosInfos):Tween {
		s.from ??= 0.0;
		s.to ??= 1.0;
		return new Tween(s, pos);
	}

	public static function removeForTarget(target:IGameObject):Void {
		final group = Event.events[tweenUpdateEvent];
		var i = group.length;
		while (i-- > 0) {
			final e = group[i];
			if (e.object == null || e.object.type != TweenType) continue;
			final tween:Tween = cast e.object;
			if (!tween.targets.contains(target)) continue;
			tween.stop();
		}
	}

	public static function getAll(target:IGameObject):Array<Tween> {
		final tweens:Array<Tween> = [];
		for (e in Event.events[tweenUpdateEvent]) {
			if (e.object == null || e.object.type != TweenType) continue;
			final tween:Tween = cast e.object;
			if (!tween.targets.contains(target)) continue;
			tweens.push(tween);
		}
		return tweens;
	}

	function readTargetStartValue(target:Any, field:String):Null<Float> {
		final start:Null<Float> = Reflect.getProperty(target, field);
		if (start == null) {
			final i = targets.indexOf(target);
			throw 'Field `$field` is null on targets[$i] (maybe field is removed by DCE?)';
		}
		return start;
	}

	function validateEndValue(fields:CustomTweenConfig, field:String, start:Float, end:Any, ?p:DebugPosInfos):Float {
		if (end is String) {
			final end:String = end;
			final r = ~/[-+\/*]=/g;
			if (r.match(end)) {
				final op = r.matched(0);
				final opValue = Std.parseFloat(end.replace(op, ""));
				final newEnd = switch op {
					case "+=": start + opValue;
					case "-=": start - opValue;
					case "*=": start * opValue;
					case "/=": start / opValue;
					case _: start;
				}
				// trace(op, opValue, newEnd); // todo
				Reflect.setProperty(fields, field, newEnd);
				final binop = op.charAt(0);
				var suggestion = '$field: obj.$field $binop $opValue';
				suggestion = suggestion.replace(" + -", " - ");
				trace('Warning: `$field: \'$op$field\'` (or \'$end\') is a String, use `$suggestion` instead', p);
				return newEnd;
			}
		} else if (Reflect.isObject(end)) {
			final end:{?to:Float} = end;
			final newEnd = end.to ?? {
				throw '`$field: $end` is Object and not supported as value';
			};
			trace('Warning: `$field: $end` is Object, use `$field: obj.$field + $newEnd` instead', p);
			Reflect.setProperty(fields, field, newEnd);
			return newEnd;
		}
		return end;
	}

	public function checkFieldTweenDuplicates():Void {
		#if debug
		for (target in targets) {
			final target:IGameObject = target;
			for (tween in getAll(target)) {
				if (tween.targets == targets) continue;
				if (tween.state != Update && tween.state != Yoyo) continue;
				for (prev in tween.fieldTweens) {
					if (!fieldTweens.inlineExists(item -> item.field == prev.field)) continue;
					// if (!fieldTweens.contains(prev.field)) continue;
					final prevPos = tween.posInfos;
					final prevPosData = '${prevPos.fileName}:${prevPos.lineNumber}';
					trace('Warning: target ${target.log()} already has field \'${prev.field}\' tween ($prevPosData)', posInfos);
				}
			}
		}

		for (target in targets) {
			final target:IGameObject = target;
			if (target.type != ContainerType) continue;
			final target:Container = cast target;
			if (target.state.debug.posContainerLock == null) continue;
			final lockNames = ["x", "y", "scale", "scaleX", "scaleY"];
			for (field in fieldTweens) {
				if (!lockNames.contains(field.field)) continue;
				var posText = '${posInfos.fileName}:${posInfos.lineNumber}';
				if (posText.length != 0) posText = '$posText: ';
				throw '${posText}Cannot tween field `${field.field}` of ${target.log()} with `posContainer`';
			}
		}
		#end
	}

	public function update(e:Event) {
		if (paused) return;
		updateState();
	}

	public function updateState() {
		switch state {
			case Start:
				if (onStart != null) onStart();
				state = Delay;
				updateState();

			case Delay:
				for (t in fieldTweens) {
					t.start = readTargetStartValue(t.target, t.field);
				}
				if (delay > 0 && time < delay) {
					time += App.deltaTime;
					return;
				} else {
					time -= delay;
					checkFieldTweenDuplicates();
					state = Update;
					updateState();
				}

			case Update:
				final max = yoyo ? duration / 2 : duration;
				updateStep(targets, max);
				if (onUpdate != null) {
					final easedPos = easeCallback(pos);
					#if debug
					if (Math.isNaN(easedPos)) throw easedPos;
					#end
					callbackEvent.setRatio(easedPos);
					callbackEvent.setTimeRatio(pos);
					onUpdate(callbackEvent);
				}
				if (time == max) {
					time = 0;
					state = YoyoDelay;
					// updateState(); // prevent callstack with addTimeEvent and small delay
					if (duration == 0) updateState();
				}

			case YoyoDelay:
				// todo? not sure about updating start or end values in other Delays
				// for (t in fieldTweens) {
				// 	t.start = readTargetStartValue(t.target, t.field);
				// }
				if (yoyoDelay > 0 && time < yoyoDelay) {
					time += App.deltaTime;
					return;
				} else {
					time -= yoyoDelay;
					if (yoyo) {
						if (onYoyo != null) onYoyo();
						checkFieldTweenDuplicates();
						state = Yoyo;
					} else {
						state = Repeat;
					}
					updateState();
				}

			case Yoyo: // similar to Update
				final max = yoyo ? duration / 2 : duration;
				updateStep(targets, max);
				if (onUpdate != null) {
					final easedPos = easeCallback(pos);
					#if debug
					if (Math.isNaN(easedPos)) throw easedPos;
					#end
					callbackEvent.setRatio(1 - easedPos);
					callbackEvent.setTimeRatio(1 - pos);
					onUpdate(callbackEvent);
				}
				if (time == max) {
					time = 0;
					state = Repeat;
					updateState();
				}

			case Repeat:
				if (repeat > 0 || repeat == -1) {
					if (repeat > 0) repeat--;
					if (onRepeat != null) onRepeat();
					time = 0;
					state = RepeatDelay;
					updateState();
				} else {
					time = 0;
					state = CompleteDelay;
					updateState();
				}

			case RepeatDelay:
				if (repeatDelay > 0 && time < repeatDelay) {
					time += App.deltaTime;
					return;
				} else {
					time -= repeatDelay;
					checkFieldTweenDuplicates();
					state = Update;
					updateState();
				}

			case CompleteDelay:
				if (completeDelay > 0 && time < completeDelay) {
					time += App.deltaTime;
					return;
				} else {
					time -= completeDelay;
					state = Complete;
					if (onComplete != null) onComplete();
					stop();
				}

			case Complete:
		}
	}

	function updateStep(targets:Array<Any>, max:Float):Void {
		time += App.deltaTime;
		if (time > max) time = max;
		if (max == 0) pos = 1;
		else pos = time / max;
		// if (pos > 0.999999) pos = 1;
		if (isCounter) {
			final easedPos = roundEaseCallback(pos);
			final start = state == Yoyo ? to : from;
			final end = state == Yoyo ? from : to;
			counterValue = easedPos.lerp(start, end);
			return;
		}
		for (target in targets) {
			step(target);
		}
	}

	function step(target:Dynamic):Void {
		for (t in fieldTweens) {
			final easedPos = roundEaseCallback(pos);
			final start = state == Yoyo ? t.end : t.start;
			final end = state == Yoyo ? t.start : t.end;
			Reflect.setProperty(t.target, t.field, easedPos.lerp(start, end));
		}
	}

	public function seekRatio(i:Float) {
		final max = yoyo ? duration / 2 : duration;
		time = i * max;
		if (paused) return;
		updateState();
	}

	inline function roundEaseCallback(pos:Float):Float {
		return roundFloat(easeCallback(pos));
	}

	inline function roundFloat(value:Float):Float {
		return Math.round(value * 10000000) / 10000000;
	}

	/** Stops tweens without `onComplete` callback, but with `onStop` instead. **/
	public function stop():Void {
		if (onStop != null) onStop();
		fieldTweens.resize(0);
		Event.off((this : Any), tweenUpdateEvent, update);
		callbackEvent.destroy();
	}

	/** Same as `stop()`, but calls `seekRatio(1)` and triggers `onComplete`. **/
	public function end():Void {
		if (state == Complete) return;
		// so seek will move to the end of Yoyo and not Update
		// maybe this should be handled with 0.5 case inside
		if (yoyo) state = Yoyo;
		seekRatio(1);
	}

	/** Returns static `Easing` callback based on your `Ease` enum value. **/
	public static function getEasingCallback(ease:Ease):(v:Float) -> Float {
		return switch ease {
			case Linear: Easing.linear;
			case SineIn: Easing.sineIn;
			case SineOut: Easing.sineOut;
			case SineInOut: Easing.sineInOut;
			case BackIn: Easing.backIn;
			case BackOut: Easing.backOut;
			case BackInOut: Easing.backInOut;
			case BounceIn: Easing.bounceIn;
			case BounceOut: Easing.bounceOut;
			case BounceInOut: Easing.bounceInOut;
			case CubicIn: Easing.cubicIn;
			case CubicOut: Easing.cubicOut;
			case CubicInOut: Easing.cubicInOut;
			case QuadIn: Easing.quadIn;
			case QuadOut: Easing.quadOut;
			case QuadInOut: Easing.quadInOut;
			case CircIn: Easing.circIn;
			case CircOut: Easing.circOut;
			case CircInOut: Easing.circInOut;
			case QuartIn: Easing.quartIn;
			case QuartOut: Easing.quartOut;
			case QuartInOut: Easing.quartInOut;
			case ExpoIn: Easing.expoIn;
			case ExpoOut: Easing.expoOut;
			case ExpoInOut: Easing.expoInOut;
			case QuintIn: Easing.quintIn;
			case QuintOut: Easing.quintOut;
			case QuintInOut: Easing.quintInOut;
		}
	}
}
