/**
 * Hertel-Mehlhorn convex polygonization from arbitrary triangulation.
 *
 * Based on:
 *
 * @see https://github.com/ivanfratric/polypartition				(C - by Ivan Fratric)
 * @see https://web.archive.org/web/20140102033642/http://www.philvaz.com/compgeom/								(by Phil Porvaznik)
 *
 * @author azrafe7
 */

package khm.geom;

using khm.geom.HertelMehlhorn.PolyTools;

private typedef HxPoint = khm.Types.Point;
private typedef Poly = Array<HxPoint>;
private typedef Tri = Poly;

@:expose
class HertelMehlhorn {
	static public var diagonals:Array<{from:HxPoint, to:HxPoint}>;

	/**
	 * Merges triangles (defining a triangulated polygon) into a set of convex polygons.
	 *
	 * @param	triangulation	An array of triangles defining the polygon.
	 * @return	An array of convex polygons being a decomposition of the original polygon.
	 */
	static public function polygonize(triangulation:Array<Tri>):Array<Poly> {
		var res:Array<Poly> = triangulation.concat([]);
		diagonals = [];

		var poly:Poly, qoly:Poly, newPoly:Poly;
		var d1:HxPoint, d2:HxPoint;
		var p:HxPoint, q:HxPoint;
		var isDiagonal:Bool;

		var polyIt, qolyIt, j;
		var outerIt = 0;
		var innerIt = 0;
		while (outerIt < res.length) {
			poly = res[outerIt];
			qoly = res[0];

			polyIt = 0;
			qolyIt = 0;
			while (polyIt < poly.length) {
				d1 = poly[polyIt];
				d2 = poly.at(polyIt + 1);

				isDiagonal = false;

				innerIt = outerIt + 1;
				while (innerIt < res.length) {
					qoly = res[innerIt];

					qolyIt = 0;
					while (qolyIt < qoly.length) {
						q = qoly[qolyIt];
						if (d2.x != q.x || d2.y != q.y) {
							qolyIt++;
							continue;
						}
						q = qoly.at(qolyIt + 1);
						if (d1.x != q.x || d1.y != q.y) {
							qolyIt++;
							continue;
						}
						isDiagonal = true;
						break;
					}
					if (isDiagonal) break;
					innerIt++;
				}

				if (!isDiagonal) {
					polyIt++;
					continue;
				}

				var d = ({from: d1, to: d2}); // d1 == poly[polyIt] and d2 == qoly[qolyIt]
				diagonals.push(d);

				p = poly.at(polyIt + 2);
				q = qoly.at(qolyIt - 1);

				// check if reflex
				if (PolyTools.isRight(d1, q, p)) {
					polyIt++;
					continue;
				}
				if (PolyTools.isRight(d2, p, q)) {
					polyIt++;
					continue;
				}

				// merge into newPoly
				newPoly = [];
				j = polyIt + 1;
				while (poly.wrappedIdx(j) != polyIt) {
					newPoly.push(poly.at(j));
					j++;
				}
				j = qolyIt + 1;
				while (qoly.wrappedIdx(j) != qolyIt) {
					newPoly.push(qoly.at(j));
					j++;
				}

				// add only if merged poly is convex (otherwise move to next triangle)
				if (PolyTools.isConvex(newPoly)) {
					res.splice(innerIt, 1);
					res[outerIt] = newPoly;
					poly = newPoly;

					polyIt = -1; // restart with newPoly
				}

				polyIt++;
			}

			outerIt++;
		}

		return res;
	}
}

class PolyTools {
	/**
	 * Assuming the polygon is simple (not self-intersecting), checks if it is convex.
	**/
	static public function isConvex(poly:Poly):Bool {
		var isPositive:Null<Bool> = null;

		for (i in 0...poly.length) {
			var lower:Int = (i == 0 ? poly.length - 1 : i - 1);
			var middle:Int = i;
			var upper:Int = (i == poly.length - 1 ? 0 : i + 1);
			var dx0:Float = poly[middle].x - poly[lower].x;
			var dy0:Float = poly[middle].y - poly[lower].y;
			var dx1:Float = poly[upper].x - poly[middle].x;
			var dy1:Float = poly[upper].y - poly[middle].y;
			var cross:Float = dx0 * dy1 - dx1 * dy0;

			// cross product should have same sign
			// for each vertex if poly is convex.
			var newIsPositive:Bool = (cross > 0 ? true : false);

			if (cross == 0) continue; // handle collinear case

			if (isPositive == null) isPositive = newIsPositive;
			else if (isPositive != newIsPositive) {
				return false;
			}
		}

		return true;
	}

	static inline public function at(poly:Poly, idx:Int):HxPoint {
		idx = wrappedIdx(poly, idx);
		return poly[idx];
	}

	/** Gets usable `idx` from `poly` (wrapping around if needed). */
	static inline public function wrappedIdx(poly:Poly, idx:Int):Int {
		var len:Int = poly.length;
		if (idx < 0) idx += len;
		return idx % len;
	}

	/** Gets the side (signed area) of `p` relative to the line extending `b`-`a` (> 0 -> left, < 0 -> right, == 0 -> collinear). */
	static inline public function side(p:HxPoint, a:HxPoint, b:HxPoint):Float {
		return (((a.x - p.x) * (b.y - p.y)) - ((b.x - p.x) * (a.y - p.y)));
	}

	/** Returns true if `p` is on the left of the line extending `b`-`a`. */
	static inline public function isLeft(p:HxPoint, a:HxPoint, b:HxPoint):Bool {
		return side(p, a, b) > 0;
	}

	/** Returns true if `p` is on the left or collinear to the line extending `b`-`a`. */
	static inline public function isLeftOrOn(p:HxPoint, a:HxPoint, b:HxPoint):Bool {
		return side(p, a, b) >= 0;
	}

	/** Returns true if `p` is on the right of the line extending `b`-`a`. */
	static inline public function isRight(p:HxPoint, a:HxPoint, b:HxPoint):Bool {
		return side(p, a, b) < 0;
	}

	/** Returns true if `p` is on the right or collinear to the line extending `b`-`a`. */
	static inline public function isRightOrOn(p:HxPoint, a:HxPoint, b:HxPoint):Bool {
		return side(p, a, b) <= 0;
	}

	/** Returns true if the specified triangle is degenerate (collinear points). */
	static inline public function isCollinear(p:HxPoint, a:HxPoint, b:HxPoint):Bool {
		return side(p, a, b) == 0;
	}
}
