package khm.internal;

import haxe.Json;
import haxe.Resource;
import haxe.io.Path;
import haxe.macro.Compiler;
import haxe.macro.Context;
import haxe.macro.Expr;
import khm.ResItem;

using StringTools;

#if macro
import sys.FileSystem;
import sys.io.File;
#end

class AssetsBuilder {
	@:persistent static final matchChars = ~/[^0-9A-z_]/g;
	@:persistent static final startsWithNumber = ~/^[0-9]/g;
	@:persistent static final replaceFirstRes = ~/^res_?/;

	static function filterAtlasesFromResourceList():Array<{id:String, path:String}> {
		final listNames = Resource.listNames();
		return listNames.filter(path -> {
			final name = Path.withoutDirectory(path);
			return name.startsWith("res") && name.endsWith(".png");
		}).map(path -> {
			final name = Path.withoutDirectory(path);
			return {
				id: name.replace(".png", ""),
				path: path.replace(".png", "")
			};
		});
	}

	macro public static function buildRes():Array<Field> {
		final fields = Context.getBuildFields();
		if (fields.length != 0) return fields;
		final resAtlasesPath = Context.definedValue("res_atlases_folder");
		final atlases = if (resAtlasesPath == null) {
			filterAtlasesFromResourceList();
		} else {
			final names = FileSystem.readDirectory(resAtlasesPath);
			names.filter(name -> {
				return name.startsWith("res") && name.endsWith(".png");
			}).map(name -> {
				return {
					id: name.replace(".png", ""),
					path: resAtlasesPath + "/" + name.replace(".png", "")
				};
			});
		}

		final pos = Context.currentPos();
		final assetsFolder = Context.definedValue("assets_folder") ?? Context.definedValue("res_folder");
		for (atlas in atlases) {
			var name = replaceFirstRes.replace(atlas.id, "");
			name = matchChars.replace(name, "_").toLowerCase();
			if (startsWithNumber.match(name)) name = '_$name';
			if (name == "") name = "res";

			final inClass = Context.getLocalClass().get();
			final valueExpr = makeType(inClass.pack, atlas, assetsFolder);
			final field:Field = {
				name: name,
				access: [APublic, AStatic, AInline],
				kind: FVar(null, valueExpr),
				pos: pos
			}
			fields.push(field);
		}
		if (!fields.exists(item -> item.name == "debug")) {
			final field:Field = {
				name: "debug",
				meta: [{
					name: ":noCompletion",
					pos: pos
				}],
				access: [APublic, AStatic, AInline],
				kind: FVar(macro :Dynamic, macro null),
				pos: pos
			}
			fields.push(field);
		}
		return fields;
	}

	#if macro
	static function makeType(pack:Array<String>, atlas:{id:String, path:String}, ?assetsFolder:String):Expr {
		final name = atlas.id;
		var pos = Context.currentPos();
		var className = matchChars.replace(name, "_").toUpperCase();

		var klass = macro class $className {};
		// klass.pack = pack.copy();
		klass.fields = buildAtlasField(atlas, assetsFolder);
		klass.fields.push({
			name: "_id",
			access: [APublic, AStatic, AFinal, AInline],
			kind: FVar(macro :String, macro $v{atlas.id}),
			pos: pos,
		});
		haxe.macro.Context.defineType(klass);

		final expr:Expr = {
			expr: EConst(CIdent(className)),
			pos: pos
		};
		return expr;
	}

	public static function buildAtlasField(atlas:{id:String, path:String}, ?assetsFolder:String):Array<Field> {
		final fields:Array<Field> = [];
		final pos = Context.currentPos();
		final json = Json.parse(File.getContent('${atlas.path}.json'));
		final frames:Iterable<Dynamic> = json.textures == null ? json.frames : json.textures[0].frames;
		var subfolder = replaceFirstRes.replace(atlas.id, "");
		if (subfolder != "") subfolder += "/";
		final values:Array<String> = [];
		final ids:Array<ResItem> = [];
		for (frame in frames) {
			var name = matchChars.replace(frame.filename, "_").toLowerCase();
			if (startsWithNumber.match(name)) name = '_$name';
			values.push(frame.filename);
			final id:ResItem = {
				frame: frame.filename,
				texture: atlas.id
			};
			ids.push(id);
			final field:Field = {
				name: name,
				access: [APublic, AStatic, AFinal],
				kind: FVar(macro :khm.ResItem, macro $v{id}),
				pos: pos,
			};
			if (assetsFolder != null) {
				final w = frame.sourceSize.w;
				final h = frame.sourceSize.h;
				final link = FileSystem.absolutePath('$assetsFolder/atlases/$subfolder${frame.filename}.png');
				final link = link.urlEncode().replace("%2F", "/");
				field.doc = '[![]($link)]($link)\n\n${w}x${h}';
			}
			fields.push(field);
		}
		fields.push({
			name: "_values",
			access: [APublic, AStatic, AFinal],
			kind: FVar(macro :Array<String>, macro $v{values}),
			pos: pos,
			// meta: [{
			// 	name: ":noCompletion",
			// 	pos: pos
			// }]
		});
		fields.push({
			name: "_ids",
			access: [APublic, AStatic, AFinal],
			kind: FVar(macro :Array<khm.ResItem>, macro $v{ids}),
			pos: pos,
		});
		return fields;
	}
	#end

	macro public static function buildAudio():Array<Field> {
		final fields = Context.getBuildFields();
		if (fields.length != 0) return fields;
		final path = Context.definedValue("res_audio_json") ?? Context.definedValue("audiores");
		if (path == null) return [];
		final pos = Context.currentPos();
		final json = Json.parse(File.getContent(path));
		final names:Array<String> = Reflect.fields(json.spritemap);
		for (file in names) {
			var name = matchChars.replace(file, "_").toLowerCase();
			if (startsWithNumber.match(name)) name = '_$name';
			fields.push({
				name: name,
				access: [APublic, AStatic, AFinal],
				kind: FVar(macro :String, macro $v{file}),
				pos: pos
			});
		}
		return fields;
	}
}
