package khm.text;

#if kha_html5
import haxe.extern.EitherType as Or;
import js.Browser.document;
import js.html.CanvasElement;
import js.html.CanvasGradient;
import js.html.CanvasRenderingContext2D;
import kha.Image;
import khm.Text.TextAlign;
import khm.Text.TextMetrics;
import khm.Text.TextStyle;

class CanvasText {
	/** Reuse canvas/context/texture from `CanvasText` pool with same `text` and `style`. **/
	public static var reuseEqualTextures = true;

	static var splitRegex = ~/(\r\n|\r|\n)/g;

	public final canvas:CanvasElement;
	public final context:CanvasRenderingContext2D;
	public var texture:Null<Image>;

	var fillGradientColor:CanvasGradient = null;

	final style:TextStyle;
	var _text:String;
	final padding = {
		left: 0,
		right: 0,
		top: 0,
		bottom: 0
	};

	public var width = 1.0;
	public var height = 1.0;

	/** Whether to automatically round line positions. **/
	var autoRound = true;
	var letterSpacing = 0;
	var resolution = 1;

	var lineSpacing = 0;
	var shadowEnabled:Bool;

	static var canvasPool:Array<CanvasElement> = [];

	public function new(text:String, ?style:TextStyle) {
		style ??= {};
		this.style = style;
		shadowEnabled = style.shadow != null;
		style.shadow ??= {};

		final canvasText = findEqualCanvasText(text, style);
		if (canvasText == null) {
			canvas = canvasPool.shift() ?? document.createCanvasElement();
			context = canvas.getContext("2d", {willReadFrequently: true});
			if (style.metrics == null) initMetrics(style);
		}
		setText(text);
		applyVerGradient(style);
	}

	public function destroy():Void {
		canvasPool.push(canvas);
	}

	function applyVerGradient(style:TextStyle):Void {
		final s = style;
		if (s.verGradient != null) {
			final gradient = context.createLinearGradient(0, 0, 0, height);
			for (i => obj in s.verGradient) {
				final isColor = @:privateAccess Text.isColorType(obj);
				if (isColor) {
					final color:Color = cast obj;
					var len = s.verGradient.length - 1;
					if (len < 1) len = 1;
					final pos = i / len;
					gradient.addColorStop(pos, color.toRgbCode());
				} else {
					final obj:{pos:Float, color:Color} = cast obj;
					gradient.addColorStop(obj.pos, obj.color.toRgbCode());
				}
			}
			fillGradientColor = gradient;
			// setFill(gradient);
		}
	}

	static final textPool:Map<String, Array<CanvasText>> = [];

	function updateTexture():Void {
		final canvasText = findEqualCanvasText(_text, style);
		if (canvasText == null) {
			textPool[_text] ??= [];
			if (!textPool[_text].contains(this)) {
				textPool[_text].push(this);
			}
			texture = Image.fromCanvas(canvas);
		} else {
			width = canvasText.width;
			height = canvasText.height;
			texture = canvasText.texture;
		}
	}

	function findEqualCanvasText(text:String, style:TextStyle):Null<CanvasText> {
		if (!reuseEqualTextures) return null;
		if (textPool[text] == null) return null;
		for (obj in textPool[text]) {
			if (obj == this) continue;
			if (obj.style.isEqualTo(style)) return obj;
		}
		return null;
	}

	function initMetrics(style:TextStyle):Void {
		syncFont();
		var metrics = context.measureText(style.testString);
		final obj = (metrics : Dynamic);
		if (obj.actualBoundingBoxAscent != null) {
			var ascent = obj.actualBoundingBoxAscent;
			var descent = obj.actualBoundingBoxDescent;
			style.metrics = {
				ascent: ascent,
				descent: descent,
				fontSize: ascent + descent
			};
		} else {
			// todo
			trace("warning: metrics is inaccurate");
			final m = context.measureText("M");
			final w = context.measureText("W");
			final max = Math.max(m.width, w.width);
			style.metrics = {
				ascent: max * (1 - 0.25),
				descent: max * 0.25,
				fontSize: max
			};
		}
	}

	public function setText(value:String):Void {
		if (value != this._text) {
			textPool[_text]?.remove(this);
			this._text = value;
			updateText();
		}
	}

	public function setStroke(color:Color, thickness:Float):Void {
		style.setStroke(color, thickness);
		updateText();
	}

	public function setShadow(offX = 0.0, offY = 0.0, color:Color = Black, blur = 0.0, shadowStroke = false, shadowFill = true):Void {
		style.shadow ??= {};
		style.shadow.offsetX = offX;
		style.shadow.offsetY = offY;
		style.shadow.color = color;
		style.shadow.blur = blur;
		style.shadow.stroke = shadowStroke;
		style.shadow.fill = shadowFill;
		updateText();
	}

	public function setFill(color:Or<Color, js.html.CanvasGradient>):Void {
		style.color = color;
		updateText();
	}

	public function setAlign(align:TextAlign):Void {
		trace("todo");
	}

	function syncStyle():Void {
		context.textBaseline = "alphabetic";

		if (fillGradientColor != null) {
			context.fillStyle = fillGradientColor;
		} else {
			context.fillStyle = style.color.toRgbaCode();
		}
		context.strokeStyle = style.stroke.toRgbaCode();

		context.lineWidth = style.strokeThickness;
		context.lineCap = "round";
		context.lineJoin = "round";
	}

	function syncFont():Void {
		final fontStyle = "";
		final _font = [fontStyle, style.fontSize + "px", style.fontFamily].join(" ").trim();
		context.font = _font;
	}

	function syncShadow(enabled:Bool):Void {
		if (enabled) { // shadowEnabled
			context.shadowOffsetX = style.shadow.offsetX;
			context.shadowOffsetY = style.shadow.offsetY;
			context.shadowColor = style.shadow.color.toRgbaCode();
			context.shadowBlur = style.shadow.blur;
		} else {
			context.shadowOffsetX = 0;
			context.shadowOffsetY = 0;
			// context.shadowColor = 0;
			context.shadowColor = null;
			context.shadowBlur = 0;
		}
	}

	function updateText():Void {
		final canvasText = findEqualCanvasText(_text, style);
		if (canvasText != null) {
			updateTexture();
			return;
		}

		var size = style.metrics;

		syncFont();

		var outputText = this._text;

		if (style.wordWrapWidth != null || style.wordWrap?.callback != null) {
			outputText = runWordWrap(this._text);
		}

		// Split text into lines
		var lines = splitRegex.split(outputText);

		var textSize = getTextSize(_text, size, lines);
		// trace(textSize);

		var padding = this.padding;

		var textWidth = 0.0;

		if (style.fixedWidth == 0) {
			this.width = textSize.width + padding.left + padding.right;

			textWidth = textSize.width;
		} else {
			this.width = style.fixedWidth;

			textWidth = this.width - padding.left - padding.right;

			if (textWidth < textSize.width) {
				textWidth = textSize.width;
			}
		}

		if (style.fixedHeight == 0) {
			this.height = textSize.height + padding.top + padding.bottom;
		} else {
			this.height = style.fixedHeight;
		}

		var w = this.width;
		var h = this.height;

		w *= resolution;
		h *= resolution;

		w = Std.int(Math.max(w, 1));
		h = Std.int(Math.max(h, 1));

		if (canvas.width != w || canvas.height != h) {
			canvas.width = Std.int(w);
			canvas.height = Std.int(h);

			// this.frame.setSize(w, h);

			// Because resizing the canvas resets the context
			syncFont();

			// if (style.rtl) context.direction = "rtl";
		} else {
			context.clearRect(0, 0, w, h);
		}

		context.save();

		context.scale(resolution, resolution);

		if (style.backgroundColor != null) {
			context.fillStyle = style.backgroundColor.toRgbaCode();
			context.fillRect(0, 0, w, h);
		}

		syncStyle();

		// Apply padding
		context.translate(padding.left, padding.top);

		var linePositionX;
		var linePositionY;

		// Draw text line by line
		for (i in 0...textSize.lines) {
			linePositionX = style.strokeThickness / 2;
			linePositionY = (style.strokeThickness / 2 + i * textSize.lineHeight) + size.ascent;

			if (i > 0) {
				linePositionY += (textSize.lineSpacing * i);
			}

			if (style.rtl) {
				linePositionX = w - linePositionX - padding.left - padding.right;
			} else if (style.align == Right) {
				linePositionX += textWidth - textSize.lineWidths[i];
			} else if (style.align == Center) {
				linePositionX += (textWidth - textSize.lineWidths[i]) / 2;
			} else if (style.align == Justify) {
				// To justify text line its width must be no less than 85% of defined width
				var minimumLengthToApplyJustification = 0.85;

				if (textSize.lineWidths[i] / textSize.width >= minimumLengthToApplyJustification) {
					var extraSpace:Float = textSize.width - textSize.lineWidths[i];
					var spaceSize = context.measureText(" ").width;
					var trimmedLine = lines[i].trim();
					var array = trimmedLine.split(" ");

					extraSpace += (lines[i].length - trimmedLine.length) * spaceSize;

					var extraSpaceCharacters = Math.floor(extraSpace / spaceSize);
					var idx = 0;

					while (extraSpaceCharacters > 0) {
						array[idx] += " ";
						var l = array.length - 1;
						if (l == 0) l = 1;
						idx = (idx + 1) % l;
						--extraSpaceCharacters;
					}

					lines[i] = array.join(" ");
				}
			}

			if (this.autoRound) {
				linePositionX = Math.round(linePositionX);
				linePositionY = Math.round(linePositionY);
			}

			if (style.strokeThickness != 0) {
				syncShadow(style.shadow.stroke != null);

				context.strokeText(lines[i], linePositionX, linePositionY);
			}

			if (style.color != null) {
				syncShadow(style.shadow.fill != null);

				// Looping fillText could be an expensive operation, we should ignore it if it is not needed
				if (this.letterSpacing != null) {
					var charPositionX = 0.0;

					var line = lines[i].split("");

					// Draw text letter by letter
					for (l in 0...line.length) {
						context.fillText(line[l], linePositionX + charPositionX, linePositionY);

						charPositionX += context.measureText(line[l]).width + this.letterSpacing;
					}
				} else {
					context.fillText(lines[i], linePositionX, linePositionY);
				}
			}
		}

		context.restore();

		// if (this.renderer && this.renderer.gl) {
		// 	this.frame.source.glTexture = this.renderer.canvasToTexture(canvas, this.frame.source.glTexture, true);
		// 	this.frame.glTexture = this.frame.source.glTexture;
		// }

		updateTexture();
	}

	function runWordWrap(text:String) {
		// if (style.wordWrapCallback) {
		// 	var wrappedLines = style.wordWrapCallback.call(style.wordWrapCallbackScope, text, this);
		// 	if (Array.isArray(wrappedLines)) {
		// 		wrappedLines = wrappedLines.join("\n");
		// 	}
		// 	return wrappedLines;
		// } else
		if (style.wordWrap?.width != null) {
			// if (style.wordWrapUseAdvanced) {
			// 	return this.advancedWordWrap(text, this.style.wordWrapWidth);
			// } else
			{
				return this.basicWordWrap(text, this.style.wordWrap?.width);
			}
		} else {
			return text;
		}
	}

	/**
	 * Advanced wrapping algorithm that will wrap words as the line grows longer than its horizontal
	 * bounds. Consecutive spaces will be collapsed and replaced with a single space. Lines will be
	 * trimmed of white space before processing. Throws an error if wordWrapWidth is less than a
	 * single character.
	 *
	 * @method Phaser.GameObjects.Text#advancedWordWrap
	 * @since 3.0.0
	 *
	 * @param {string} text - The text to perform word wrap detection against.
	 * @param {CanvasRenderingContext2D} context - The Canvas Rendering Context.
	 * @param {number} wordWrapWidth - The word wrap width.
	 *
	 * @return {string} The wrapped text.
	 */
	function advancedWordWrap(text:String, wordWrapWidth:Float) {
		var output = "";

		// Condense consecutive spaces and split into lines
		var lines = splitRegex.split(~/ +/gi.replace(text, " "));

		var linesCount = lines.length;

		for (i in 0...linesCount) {
			var line = lines[i];
			var out = "";

			// Trim whitespace
			line = ~/^ *|\s*$/gi.replace(line, "");

			// If entire line is less than wordWrapWidth append the entire line and exit early
			var lineWidth = context.measureText(line).width;

			if (lineWidth < wordWrapWidth) {
				output += line + "\n";
				continue;
			}

			// Otherwise, calculate new lines
			var currentLineWidth = wordWrapWidth;

			// Split into words
			var words = line.split(" ");

			for (j in 0...words.length) {
				var word = words[j];
				var wordWithSpace = word + " ";
				var wordWidth = context.measureText(wordWithSpace).width;

				if (wordWidth > currentLineWidth) {
					// Break word
					if (j == 0) {
						// Shave off letters from word until it"s small enough
						var newWord = wordWithSpace;

						while (newWord.length != 0) {
							// newWord = newWord.slice(0, -1);
							newWord = newWord.substr(0, newWord.length - 1);
							wordWidth = context.measureText(newWord).width;

							if (wordWidth <= currentLineWidth) {
								break;
							}
						}

						// If wordWrapWidth is too small for even a single letter, shame user
						// failure with a fatal error
						if (newWord.length == 0) {
							throw "wordWrapWidth < a single character";
						}

						// Replace current word in array with remainder
						var secondPart = word.substr(newWord.length);

						words[j] = secondPart;

						// Append first piece to output
						out += newWord;
					}

					// If existing word length is 0, don"t include it
					var offset = (words[j].length != 0) ? j : j + 1;

					// Collapse rest of sentence and remove any trailing white space
					var remainder = words.slice(offset).join(" ");
					remainder = ~/[ \n]*$/gi.replace(remainder, "");

					// Prepend remainder to next line
					// lines.splice(i + 1, 0, remainder);
					lines.insert(i + 1, remainder);

					linesCount = lines.length;

					break; // Processing on this line

					// Append word with space to output
				} else {
					out += wordWithSpace;
					currentLineWidth -= wordWidth;
				}
			}

			// Append processed line to output
			output += ~/[ \n]*$/gi.replace(out, "") + "\n";
		}

		// Trim the end of the string
		output = ~/[\s|\n]*$/gi.replace(output, "");

		return output;
	}

	/**
	 * Greedy wrapping algorithm that will wrap words as the line grows longer than its horizontal
	 * bounds. Spaces are not collapsed and whitespace is not trimmed.
	 *
	 * @method Phaser.GameObjects.Text#basicWordWrap
	 * @since 3.0.0
	 *
	 * @param {string} text - The text to perform word wrap detection against.
	 * @param {CanvasRenderingContext2D} context - The Canvas Rendering Context.
	 * @param {number} wordWrapWidth - The word wrap width.
	 *
	 * @return {string} The wrapped text.
	 */
	function basicWordWrap(text:String, wordWrapWidth:Float) {
		var result = "";
		var lines = splitRegex.split(text);
		var lastLineIndex = lines.length - 1;
		var whiteSpaceWidth = context.measureText(" ").width;

		for (i in 0...lines.length) {
			var spaceLeft = wordWrapWidth;
			var words = lines[i].split(" ");
			var lastWordIndex = words.length - 1;

			for (j in 0...words.length) {
				var word = words[j];
				var wordWidth = context.measureText(word).width;
				var wordWidthWithSpace = wordWidth;

				if (j < lastWordIndex) {
					wordWidthWithSpace += whiteSpaceWidth;
				}

				if (wordWidthWithSpace > spaceLeft) {
					// Skip printing the newline if it's the first word of the line that is greater
					// than the word wrap width.
					if (j > 0) {
						result += "\n";
						spaceLeft = wordWrapWidth;
					}
				}

				result += word;

				if (j < lastWordIndex) {
					result += " ";
					spaceLeft -= wordWidthWithSpace;
				} else {
					spaceLeft -= wordWidth;
				}
			}

			if (i < lastLineIndex) {
				result += "\n";
			}
		}

		return result;
	}

	/**
	 * Returns an object containing dimensions of the Text object.
	 *
	 * @function Phaser.GameObjects.GetTextSize
	 * @since 3.0.0
	 *
	 * @param {Phaser.GameObjects.Text} text - The Text object to calculate the size from.
	 * @param {Phaser.Types.GameObjects.Text.TextMetrics} size - The Text metrics to use when calculating the size.
	 * @param {string[]} lines - The lines of text to calculate the size from.
	 *
	 * @return {Phaser.Types.GameObjects.Text.GetTextSizeObject} An object containing dimensions of the Text object.
	 */
	function getTextSize(text:String, size:TextMetrics, lines:Array<String>) {
		// var canvas = text.canvas;
		// var context = text.context;
		// var style = text.style;

		var lineWidths = [];
		var maxLineWidth = 0.0;
		var drawnLines = lines.length;

		if (style.maxLines > 0 && style.maxLines < lines.length) {
			drawnLines = style.maxLines;
		}

		syncFont();

		// Text Width

		for (i in 0...drawnLines) {
			var lineWidth = style.strokeThickness;

			lineWidth += context.measureText(lines[i]).width;

			if (lines[i].length > 1) {
				lineWidth += letterSpacing * (lines[i].length - 1);
			}

			// Adjust for wrapped text
			if (style.wordWrap != null) {
				lineWidth -= context.measureText(" ").width;
			}

			lineWidths[i] = Math.ceil(lineWidth);
			maxLineWidth = Math.max(maxLineWidth, lineWidths[i]);
		}

		// Text Height

		var lineHeight = size.fontSize + style.strokeThickness;
		var height = lineHeight * drawnLines;
		var lineSpacing = lineSpacing;

		// Adjust for line spacing
		if (drawnLines > 1) {
			height += lineSpacing * (drawnLines - 1);
		}

		return {
			width: maxLineWidth,
			height: height,
			lines: drawnLines,
			lineWidths: lineWidths,
			lineSpacing: lineSpacing,
			lineHeight: lineHeight
		};
	}

	// override function render(g:kha.graphics2.Graphics) {
	// 	g.drawImage();
	// }
	// override function render(g:kha.graphics2.Graphics) {
	// 	if (!visible || alpha == 0) return;
	// 	g.color = 0xFFFFFFFF;
	// 	pushContainerSpace(g);
	// 	g.drawRect(0, 0, 100, 100);
	// 	g.drawImage(texture, 10, 10);
	// 	popContainerSpace(g);
	// }
}
#end
