package khm.tilemap;

import khm.tilemap.TilemapLayer;
#if kha
import kha.graphics2.Graphics;
#end

typedef TilemapDataJson = {
	?version:Int,
	?name:String,
	w:Int,
	h:Int,
	layers:Array<Array<Array<Int>>>,
	objects:Array<TilemapObject>,
}

typedef TilemapData = { // map format
	?version:Int,
	?name:String,
	w:Int,
	h:Int,
	layers:Array<TilemapLayer>,
	objects:Array<TilemapObject>,
}

typedef TilemapObject = {
	type:String,
	layerId:Int,
	x:Int,
	y:Int,
	data:Null<Dynamic>
}

class Tilemap extends Container {
	public var map(default, null):TilemapData;
	public var w(get, never):Int;
	public var h(get, never):Int;

	inline function get_w():Int {
		return map.w;
	}

	inline function get_h():Int {
		return map.h;
	}

	public var tileW(get, never):Int;

	inline function get_tileW():Int {
		return tileset.tileW;
	}

	public var tileH(get, never):Int;

	inline function get_tileH():Int {
		return tileset.tileH;
	}

	public final tileset:Tileset;

	final emptyTiles:Array<Tile> = [];

	public function new(tileset:Tileset, ?pos:DebugPosInfos) {
		super();
		this.setPosInfos(pos);

		originX = 0;
		originY = 0;
		#if kha
		scaleQuality = Low;
		#end
		this.tileset = tileset;
		for (layerId in 0...tileset.layers.length) {
			final sprite = new Sprite(tileset.nextFrame);
			// sprite.id = 0;
			// sprite.visible = false;
			final tile = new Tile(this, sprite, layerId, 0, -1, -1);
			tile;
			emptyTiles.push(tile);
		}
	}

	public function loadMap(map:TilemapData):Void {
		this.map = copyMap(map);
		setMapSize();
	}

	public function loadSubMap(map:TilemapData):Void {}

	public function loadJson(mapJson:TilemapDataJson):Void {
		map = fromJson(mapJson);
		children.getAll().iter(object -> object.off());
		removeAll();
		for (layer in map.layers) add(layer);
		setMapSize();
	}

	function setMapSize():Void {
		if (map.h != map.layers[0].h || map.w != map.layers[0].w) {
			final w = map.layers[0].w;
			final h = map.layers[0].h;
			throw 'Map size (${map.w}x${map.h}) properties are incorrect to layer 0 size (${w}x$h)';
		}
		setSize(map.w * tileW, map.h * tileH);
	}

	public function getTileX(ix:Int):Float {
		return ix * tileW + tileW / 2 - width * originX;
	}

	public function getTileY(iy:Int):Float {
		return iy * tileH + tileH / 2 - height * originY;
	}

	public function fromJson(map:TilemapDataJson):TilemapData {
		final layers:Array<TilemapLayer> = [
			for (l in 0...map.layers.length)
				TilemapLayer.fromJson(this, l, map.layers[l])
		];
		return {
			name: map.name,
			w: map.w,
			h: map.h,
			layers: layers,
			objects: map.objects,
		}
	}

	public function toJson(map:TilemapData):TilemapDataJson {
		final layers:Array<Array<Array<Int>>> = [
			for (l in 0...map.layers.length) [
				for (iy in 0...map.layers[l].h) [
					for (ix in 0...map.layers[l].w)
						map.layers[l].map[iy][ix].id
				]
			]
		];
		return {
			name: map.name,
			w: map.w,
			h: map.h,
			layers: layers,
			objects: map.objects,
		}
	}

	public function copyMap(map:TilemapData):TilemapData {
		final layers:Array<TilemapLayer> = [
			for (layer in map.layers) layer.copy()
		];
		@:nullSafety(Off)
		final objects:Array<TilemapObject> = [
			for (obj in map.objects) copyGameObject(obj)
		];
		return {
			name: map.name,
			w: map.w,
			h: map.h,
			layers: layers,
			objects: objects,
		}
	}

	function copyGameObject(obj:TilemapObject):TilemapObject {
		final data = obj.data == null ? null : Reflect.copy(obj.data);
		return {
			type: obj.type,
			layerId: obj.layerId,
			x: obj.x,
			y: obj.y,
			data: data
		}
	}

	public inline function isInside(x:Int, y:Int):Bool {
		return x > -1 && y > -1 && x < map.w && y < map.h;
	}

	public function getTile(layerId:Int, x:Int, y:Int):Tile {
		if (!isInside(x, y)) return emptyTiles[layerId];
		return map.layers[layerId].map[y][x];
	}

	public function getTileId(layerId:Int, x:Int, y:Int):Int {
		if (!isInside(x, y)) return 0;
		return map.layers[layerId].map[y][x].id;
	}

	public function setTile(layerId:Int, x:Int, y:Int, tile:Tile):Void {
		if (!isInside(x, y)) return;
		map.layers[layerId].map[y][x] = tile;
	}

	public function setTileId(layerId:Int, x:Int, y:Int, id:Int):Void {
		if (!isInside(x, y)) return;
		map.layers[layerId].map[y][x].id = id;
		final rect = tileset.layers[layerId][id].rect;
		// map.layers[layerId].map[y][x].setCrop(rect);
	}

	public inline function everyTile(layerId:Int, callback:(tile:Tile) -> Void):Void {
		for (iy in 0...h) {
			for (ix in 0...w) {
				final tile = map.layers[layerId].map[iy][ix];
				callback(tile);
			}
		}
	}

	public function getObject(layerId:Int, x:Int, y:Int, type:String):Null<TilemapObject> {
		for (obj in map.objects) {
			if (obj.x == x
				&& obj.y == y
				&& obj.layerId == layerId
				&& obj.type == type) return obj;
		}
		return null;
	}

	public function setObject(layerId:Int, x:Int, y:Int, type:String, data:Any):Void {
		if (data == null) {
			removeObject(layerId, x, y, type);
			return;
		}
		final obj = getObject(layerId, x, y, type);
		if (obj != null) {
			obj.data = data;
			return;
		}
		map.objects.push({
			type: type,
			layerId: layerId,
			x: x,
			y: y,
			data: data
		});
	}

	public function removeObject(layerId:Int, x:Int, y:Int, type:String):Void {
		final obj = getObject(layerId, x, y, type);
		if (obj != null) map.objects.remove(obj);
	}

	public function getProps(layerId:Int, id:Int):Null<Any> {
		return tileset.layers[layerId][id].props;
	}
}
