package khm.tools;

import khm.Event.EventKind;
import khm.Tween.TweenConfig;
import khm.Types.Point;
import khm.Types.Rect;
import khm.math.FastMatrix3;
import khm.ui.Widget;

class GameObjectTools {
	public static function on<T>(obj:IGameObject, event:EventKind<T>, callback:T, ?pos:DebugPosInfos):Void {
		Event.on(obj, event, callback, pos);
	}

	public static function once<T>(obj:IGameObject, event:EventKind<T>, callback:T, ?pos:DebugPosInfos):Void {
		Event.on(obj, event, callback, true, pos);
	}

	public static function off<T>(obj:IGameObject, ?event:EventKind<T>, ?callback:T):Void {
		Event.off(obj, event, callback);
	}

	public static function setOrigin(sprite:IGameObject, x:Float, ?y:Float):Void {
		sprite.originX = x;
		sprite.originY = y ?? x;
	}

	public static function addTween(obj:IGameObject, s:TweenConfig, ?pos:DebugPosInfos):Tween {
		s.targets ??= obj;
		return new Tween(s, pos);
	}

	public static function removeTweens(obj:IGameObject):Void {
		Tween.removeForTarget(obj);
	}

	public static function endTweens(obj:IGameObject):Void {
		final tweens = Tween.getAll(obj);
		for (tween in tweens) {
			tween.seekRatio(1);
			tween.stop();
		}
	}

	/**
		Iterates all `obj` parents and returns first parent for which `callback` returns `true`.

		If there is no parents, or all returns `false`, the result is `null`.
	**/
	public static inline function findParent(obj:IGameObject, callback:(obj:IContainer) -> Bool):Null<IContainer> {
		var parent:Null<IContainer> = obj.parent;
		while (parent != null) {
			if (callback(parent)) break;
			parent = parent.parent;
		}
		return parent;
	}

	/**
		Iterates all `obj` parents and returns first parent with `parent.type == SceneType`.

		If there is no scene in `obj` tree, the result is `null`.
	**/
	public static function findScene(obj:IGameObject):Null<Scene> {
		return cast findParent(obj, obj -> obj.type == SceneType);
	}

	public static function findSpriteByFrame(container:IContainer, id:ResItem):Null<Sprite> {
		final obj = container.children.inlineFind(obj -> {
			if (obj.type != SpriteType) return false;
			final sprite:Sprite = cast obj;
			return sprite.id.frame == id.frame && sprite.id.texture == id.texture;
		});
		return cast obj;
	}

	public static function getData(sprite:IGameObject, key:String):Any {
		return sprite.state.get(key);
	}

	public static function setData(sprite:IGameObject, key:String, value:Any, ?pos:DebugPosInfos):Void {
		sprite.state.set(key, value);
	}

	public static function setPosInfos(sprite:IGameObject, pos:DebugPosInfos):Void {
		if (pos == null) return;
		sprite.state.debug.posInfos = pos;
	}

	public static function posInfosLine(sprite:IGameObject):Null<String> {
		final pos = sprite.state.debug.posInfos ?? return null;
		return '${pos.fileName}:${pos.lineNumber}';
	}

	static function genLogContainerName(obj:Container):String {
		final typeName = obj.type == SceneType ? "scene" : "container";
		var name = obj.name;
		if (name == null) {
			final classPath = Type.getClassName(Type.getClass(obj));
			final className = classPath.split(".").pop();
			if (className == "Container") return typeName;
			name = className;
		}
		return '$typeName $name';
	}

	public static function log(child:IGameObject, depth = 1):String {
		return switch child.type {
			case SceneType, ContainerType, WidgetType:
				final c:Container = cast child;
				final name = genLogContainerName(cast child);
				if (depth == 0) {
					return '$name with ${c.children.length} children';
				}
				depth--;
				final children = if (c.children.length < 5) {
					c.children.map(obj -> log(obj, depth));
				} else {
					[
						log(c.children[0], depth),
						log(c.children[1], depth),
						"...",
						log(c.children[2], depth),
						log(c.children[3], depth),
					];
				}
				'$name with ${c.children.length} children: [ ${children.join(", ")} ]';
			case SpriteType:
				final sprite:Sprite = cast child;
				'sprite (${sprite.id.texture}.${sprite.id.frame})';
			case TweenType:
				final tween:Tween = cast child;
				final targets = if (tween.targets.length < 5) {
					tween.targets.map(obj -> log(obj, 0));
				} else {
					[
						log(tween.targets[0], 0),
						log(tween.targets[1], 0),
						"...",
						log(tween.targets[2], 0),
						log(tween.targets[3], 0),
					];
				}
				'tween with ${tween.targets.length} targets: [ ${targets.join(", ")} ]';
			case TextType:
				final text:Text = cast child;
				var s = text.text;
				if (s.length > 20) {
					s = s.substr(0, 10) + "..." + s.substr(-10);
				}
				'text (\'$s\')';
			case GraphicsType:
				final parent = child.parent ?? return "graphics (no parent)";
				'graphics (children[${parent.children.indexOf(child)}])';
			case ParticlesType:
				'particles';
		}
	}

	/**
		Returns top-left corner point of game object in `obj` parent space.

		For example, `obj` on x/y position `0, 0` with size `50x50` and originX/Y `0.5, 0.5` will return `{x: -25, y: -25}`.

		With same size and originX/Y `0, 0`, result will be `{x: 0, y: 0}`.

		If you want to get result relative to screen space, use `obj.getGlobalPosition()`.
	**/
	public static inline function topLeft(obj:IGameObject):Point {
		return {
			x: obj.x - obj.displayWidth * obj.originX,
			y: obj.y - obj.displayHeight * obj.originY,
		}
	}

	/**
		Returns top-left corner point, relative to object local scope, based on object position, scale and origin.

		For example, `obj` on any x/y position, with size `50x50` and originX/Y `0.5, 0.5` will return `{x: -25, y: -25}`.

		With same size and originX/Y `0, 0`, result will be `{x: 0, y: 0}`.
	**/
	public static inline function topLeftLocal(obj:IGameObject):Point {
		final local:Point = {
			x: -obj.width * obj.originX,
			y: -obj.height * obj.originY,
		}
		return local;
	}

	/** Returns rect of object position in global space */
	public static inline function getGlobalPosition(obj:IGameObject):Rect {
		#if debug
		if (obj.type != SceneType && obj.parent == null) throw obj;
		#end
		final t = obj.parent != null ? obj.parent.transformation : FastMatrix3.identity();
		final m:FastMatrix3 = t;
		final p = obj.topLeft();
		final x = m.scaleX * p.x + m.skewX * p.y + m.translateX;
		final y = m.scaleY * p.y + m.skewY * p.x + m.translateY;
		return {
			x: x,
			y: y,
			w: obj.displayWidth * m.scaleX,
			h: obj.displayHeight * m.scaleY
		}
	}

	public static function updateTransformation(obj:IGameObject):Void {
		switch obj.type {
			case SceneType:
				final obj:Scene = cast obj;
				obj.updateTransformation();
			case ContainerType:
				final obj:Container = cast obj;
				obj.updateTransformation();
			case SpriteType:
				final sprite:Sprite = cast obj;
				sprite.updateTransformation();
			case TweenType:
			case TextType:
				final sprite:Text = cast obj;
				sprite.updateTransformation();
			case GraphicsType:
				final sprite:Graphics = cast obj;
				sprite.updateTransformation();
			case ParticlesType:
			case WidgetType:
		}
	}

	public static inline function toGlobal(obj:IGameObject, globalX = 0.0, globalY = 0.0):Rect {
		updateTransformation(obj);
		final m:FastMatrix3 = obj.transformation;
		final x = m.getX(globalX, globalY);
		final y = m.getY(globalX, globalY);
		final scaleX = m.scaleX;
		final scaleY = m.scaleY;
		// var angle = m.rotationNormalized.toDeg();
		return {
			x: x,
			y: y,
			w: obj.width * scaleX,
			h: obj.height * scaleY
		};
	}

	public static inline function toLocal(fromObj:IGameObject, obj:IGameObject, globalX = 0.0, globalY = 0.0):Rect {
		updateTransformation(fromObj);
		// to global
		final m:FastMatrix3 = fromObj.transformation;
		final x = m.getX(globalX, globalY);
		final y = m.getY(globalX, globalY);
		var scaleX = m.scaleX;
		var scaleY = m.scaleY;
		// var angle = m.rotationNormalized.toDeg();

		// to local (obj)
		final m:FastMatrix3 = obj.transformation.inverse();
		final localX = m.getX(x, y);
		final localY = m.getY(x, y);
		scaleX *= m.scaleX;
		scaleY *= m.scaleY;
		// angle += m.rotationNormalized.toDeg();
		return {
			x: localX,
			y: localY,
			w: fromObj.width * scaleX,
			h: fromObj.height * scaleY
		}
	}

	public static inline function toLocalPos(fromObj:IGameObject, globalX = 0.0, globalY = 0.0, ?obj:IGameObject):Point {
		updateTransformation(fromObj);
		// to global
		var x = globalX;
		var y = globalY;
		var scaleX = 1.0;
		var scaleY = 1.0;
		if (obj != null) {
			final m:FastMatrix3 = obj.transformation;
			x = m.getX(globalX, globalY);
			y = m.getY(globalX, globalY);
			scaleX = m.scaleX;
			scaleY = m.scaleY;
		}
		// var angle = m.rotationNormalized.toDeg();

		// to local (obj)
		final m:FastMatrix3 = fromObj.transformation.inverse();
		final localX = m.getX(x, y);
		final localY = m.getY(x, y);
		scaleX *= m.scaleX;
		scaleY *= m.scaleY;
		// angle += m.rotationNormalized.toDeg();
		final sizeObj = obj ?? fromObj;
		return {
			x: localX,
			y: localY,
			// w: sizeObj.width * scaleX,
			// h: sizeObj.height * scaleY
		}
	}

	/**
		Adds `obj` to `parent` container with pos/size update to prevent visual changes.

		So you can move `obj` to a new container, `obj` will keep his previous
		global position, and then tween `obj` to position you want in a new container.
	**/
	public static function attach(parent:IContainer, obj:IGameObject):Void {
		updateTransformation(parent);

		// to global
		final m:FastMatrix3 = obj.transformation;
		if (obj.parent != null) obj.parent.remove(obj);
		obj.x = m.getX(0, 0);
		obj.y = m.getY(0, 0);
		obj.scaleX = m.scaleX;
		obj.scaleY = m.scaleY;
		obj.angle = m.rotationNormalized.toDeg();
		// to local (parent)
		parent.add(obj);
		final m:FastMatrix3 = parent.transformation.inverse();
		final x = obj.x;
		final y = obj.y;
		obj.x = m.getX(x, y);
		obj.y = m.getY(x, y);
		obj.scaleX *= m.scaleX;
		obj.scaleY *= m.scaleY;
		obj.angle += m.rotationNormalized.toDeg();
	}

	/** Returns IGameObject as Widget, or null, if `obj.type` is not `WidgetType` **/
	public static function toWidget(widget:IGameObject):Null<Widget> {
		if (widget.type == WidgetType) return cast widget;
		return null;
	}
}
