package khm.tools;

import khm.utils.FrameUtils;
import khm.utils.Hand;
#if phaser
import phaser.gameobjects.components.TransformMatrix;
#end

class PhaserTools {
	#if phaser
	public static function phaserObject(sprite:IGameObject):phaser.gameobjects.GameObject {
		return (sprite : Dynamic).phaser;
	}
	#end

	public static function setTexture(sprite:Sprite, texture:String):Void {
		sprite.setFrame({
			frame: sprite.id?.frame,
			texture: texture
		});
	}

	public static function play(sprite:Sprite, anim:Animation):Tween {
		final prevTween:Tween = sprite.getData("animTween");
		prevTween?.stop();
		anim.defaultTextureKey ??= anim.key;
		anim.sortFrames ??= true;
		if (anim.duration == null) {
			anim.frameRate ??= 24;
			anim.duration = anim.frames.length / anim.frameRate * 1000;
		}
		if (anim.sortFrames) {
			anim.frames.sort(FrameUtils.numericIdSort);
		}
		final animTween = sprite.addTween({
			ease: Linear,
			duration: anim.duration,
			delay: anim.delay,
			repeat: anim.repeat,
			repeatDelay: anim.repeatDelay,
			yoyo: anim.yoyo,
			onStart: () -> {
				if (anim.showOnStart) sprite.visible = true;
			},
			onComplete: () -> {
				if (anim.hideOnComplete) sprite.visible = false;
				if (anim.onComplete != null) anim.onComplete();
			},
			onUpdate: e -> {
				var i = Std.int(anim.frames.length * e.ratio);
				if (i > anim.frames.length - 1) i = anim.frames.length - 1;
				final frame = anim.frames[i];
				sprite.setFrame(frame);
			}
		});
		sprite.setData("animTween", animTween);
		return animTween;
	}

	public static function playLoop(sprite:Sprite, anim:Animation):Tween {
		final prevTween:Tween = sprite.getData("animTween");
		prevTween?.stop();
		anim.defaultTextureKey ??= anim.key;
		anim.sortFrames ??= true;
		if (anim.duration == null) {
			anim.frameRate ??= 24;
			anim.duration = anim.frames.length / anim.frameRate * 1000;
		}
		if (anim.sortFrames) {
			anim.frames.sort(FrameUtils.numericIdSort);
		}
		var i = 0;
		final frameDuration = anim.duration / anim.frames.length;
		final animTween = sprite.addTween({
			ease: Linear,
			duration: 0,
			delay: anim.delay,
			repeat: -1,
			repeatDelay: frameDuration,
			yoyo: anim.yoyo,
			onStart: () -> {
				if (anim.showOnStart) sprite.visible = true;
			},
			onStop: () -> { // onComplete
				if (anim.hideOnComplete) sprite.visible = false;
				if (anim.onComplete != null) anim.onComplete();
			},
			onRepeat: () -> {
				// var i = Std.int(anim.frames.length * e.ratio);
				i++;
				if (i > anim.frames.length - 1) i = 0;
				final frame = anim.frames[i];
				sprite.setFrame(frame);
			}
		});
		sprite.setData("animTween", animTween);
		return animTween;
	}

	public static function setDepth(obj:IGameObject, depth:Float, ?pos:DebugPosInfos):Void {
		#if debug
		if (depth == null) throw "depth is null";
		#end
		obj.state.depth = depth;
		obj.once(BeforeRender, setDepthSort);
	}

	static function setDepthSort(e:Event):Void {
		final obj = e.target;
		final container:Container = cast obj.parent ?? {
			trace(obj.log());
			throw "parent of this object is null at first scene render, depth sorting cannot be initialized";
		}
		// if only depth is updated just resort parent
		container.sort(obj -> obj.state.depth ?? 0);
		if (container.state.sortable) return;
		container.state.sortable = true;
		// also watch parent for content changes
		final oldCallback = container.onContentChanged;
		container.onContentChanged = (type, child) -> {
			final sortable:Bool = container.state.sortable;
			if (!sortable) return;
			if (type != Added) return;
			container.once(BeforeRender, e -> {
				oldCallback(type, child);

				container.sort(obj -> obj.state.depth ?? 0);
			});
		}
	}

	public static function moveUp(child:IGameObject):Void {
		final parent = child.parent;
		final i = parent.children.indexOf(child);
		if (i == -1) {
			trace("Parent: ", parent, "Child: ", child);
			throw 'Parent does not have child of type ${child.log()}';
		}
		parent.remove(child);
		parent.addAt(i + 1, child);
	}

	public static function bringToTop(child:IGameObject):Void {
		final parent = child.parent;
		if (!parent.children.contains(child)) {
			trace("Parent: ", parent, "Child: ", child);
			throw 'Parent does not have child of type ${child.log()}';
		}
		parent.remove(child);
		parent.add(child);
	}

	@:noCompletion
	public static function setInteractive(sprite:IGameObject):Void {
		sprite.isInteractive = true;
	}

	@:noCompletion
	public static function removeInteractive(sprite:IGameObject):Void {
		sprite.isInteractive = false;
	}

	public static function parentContainer(sprite:IGameObject):IContainer {
		if (sprite.parent?.type == SceneType) return null;
		return sprite.parent;
	}

	public static function addHand(parent:IContainer, frame:ResItem, ?pos:DebugPosInfos):Hand {
		final hand = new Hand(frame);
		hand.sprite.setPosInfos(pos);
		parent.add(hand.sprite);
		return hand;
	}

	#if phaser
	public static function getWorldTransformMatrix(sprite:IGameObject):TransformMatrix {
		final m = sprite.transformation;
		return new TransformMatrix(
			m._00, m._01,
			m._10, m._11,
			m._20, m._21
		);
	}
	#end
}
