package khm.ui.utils;

import khm.IContainer.Children;
import khm.Types.Point;
import khm.Types.Size;

using khm.ui.WidgetTools;

abstract class FlexBase extends Widget {
	/**
		Main axis of this widget.

		(A `Row` main axis is horizontal, and a `Column` main axis is vertical).
	**/
	public final mainAxis:Axis;
	public var scrollX = 0.0;
	public var scrollY = 0.0;
	public var isScrollable = false;

	public function new(mainAxis:Axis) {
		super();
		clipBehavior = Strict;
		this.mainAxis = mainAxis;
	}

	public function rebuildChildren():Void {
		var maxW = context.maxW;
		var maxH = context.maxH;
		if (isScrollable) {
			switch mainAxis {
				case Horizontal:
					maxW = Widget.maxSize;
				case Vertical:
					maxH = Widget.maxSize;
			}
		}
		for (i => child in children) {
			final child = child.toWidget() ?? continue;
			if (child.hasFlexFactor()) continue;
			child.context.maxW = maxW;
			child.context.maxH = maxH;
			// child.context.setFrom(context);
			final newChild = child.build();
			switch mainAxis {
				case Horizontal:
					maxW -= newChild.displayWidth;
				case Vertical:
					maxH -= newChild.displayHeight;
			}
			if (newChild == child) continue;
			child.dispose();
			remove(child);
			addAt(i, newChild);
		}
	}

	function updateFlexChildrenSizes():Void {
		var totalW = 0.0;
		var totalH = 0.0;
		for (obj in children) {
			if (obj.hasFlexFactor()) continue;
			totalW += obj.width;
			totalH += obj.height;
		}
		var totalFlex = children.getTotalFlex();
		for (i => child in children) {
			if (!child.hasFlexFactor()) continue;
			final child = child.toWidget();
			final ratio = child.flexFactor / totalFlex;
			final freeW = width - totalW;
			final freeH = height - totalH;
			if (mainAxis == Horizontal) {
				var maxW = freeW * ratio;
				if (maxW != Std.int(maxW)) maxW -= 0.0001;
				child.context.maxW = maxW;
				child.context.maxH = height;
			} else {
				var maxH = freeH * ratio;
				if (maxH != Std.int(maxH)) maxH -= 0.0001;
				child.context.maxW = width;
				child.context.maxH = maxH;
			}
			// child.context.setFrom(context);
			final newChild = child.build();
			if (newChild == child) continue;
			child.dispose();
			remove(child);
			addAt(i, newChild);
		}
	}

	inline function getMaxChildrenSize():Size {
		var maxW = 0.0;
		var maxH = 0.0;
		if (mainAxis == Horizontal) {
			for (object in children) {
				if (maxH < object.height) maxH = object.height;
				maxW += object.width;
			}
		} else {
			for (object in children) {
				if (maxW < object.width) maxW = object.width;
				maxH += object.height;
			}
		}
		return {
			w: maxW,
			h: maxH
		}
	}

	function getMainAxisOffset(i:Int, mainAxisAlign:MainAxisAlignment, totalSpace:Float):Float {
		final offset = switch mainAxisAlign {
			case Start: 0;
			case End:
				totalSpace;
			case Center:
				totalSpace / 2;
			case SpaceBetween:
				totalSpace / (children.length - 1) * i;
			case SpaceAround:
				final space = totalSpace / children.length;
				space / 2 + space * i;
			case SpaceEvenly:
				totalSpace / (children.length + 1) * (i + 1);
		}
		return switch mainAxis {
			case Horizontal:
				offset - scrollX;
			case Vertical:
				offset - scrollY;
		}
	}

	function getCrossAxisPos(obj:IGameObject, crossAxisAlign:CrossAxisAlignment, maxW:Float, maxH:Float):Float {
		final totalSpace = switch mainAxis {
			case Horizontal:
				maxH - obj.height;
			case Vertical:
				maxW - obj.width;
		}
		final off = switch mainAxis {
			case Horizontal: -scrollY;
			case Vertical: -scrollX;
		}
		return off + switch crossAxisAlign {
			case Start: 0;
			case End: totalSpace;
			case Center: totalSpace / 2;
			case Stretch:
				throw "Stretch not implemented";
		}
	}

	function clipChildren():Void {
		if (clipBehavior == None) return;
		if (clipBehavior == Crop) {
			context.setCrop(0, 0, context.maxW, context.maxH);
		}
		var isLooseClip = clipBehavior == Crop || clipBehavior == Loose;
		for (obj in children) {
			final p = obj.topLeft();
			var minX = p.x;
			var minY = p.y;
			var maxX = p.x + obj.displayWidth;
			var maxY = p.y + obj.displayHeight;
			if (!isLooseClip) {
				minX += obj.displayWidth;
				minY += obj.displayHeight;
				maxX -= obj.displayWidth;
				maxY -= obj.displayHeight;
			}
			if (maxX < 0 || maxY < 0) {
				obj.visible = false;
				if (debugClip) {
					obj.visible = true;
					obj.alpha = 0.5;
				}
			} else if (minX > context.maxW || minY > context.maxH) {
				obj.visible = false;
				if (debugClip) {
					obj.visible = true;
					obj.alpha = 0.5;
				}
			} else {
				obj.visible = true;
				if (debugClip) obj.alpha = 1;
			}
		}
	}
}
