package khm.utils;

import haxe.Json;
import haxe.crypto.Base64;
import haxe.io.Bytes;
#if kha
import kha.Assets;
import kha.Image;
#elseif phaser
import js.Browser.document;
import js.Browser.window;
import js.html.CanvasElement;
import js.html.Image;
#end

typedef AtlasFrameData = {
	final filename:String;
	final rotated:Bool;
	final trimmed:Bool;
	/** Non-trimmed original texture size. **/
	final sourceSize:{w:Int, h:Int};
	/**
		Trimmed rect cords of original texture.

		For example, if transparent texture has size `50x50`
		and has rect at `x: 10, y: 10` with rect size `w: 20, h: 20`,
		`spriteSourceSize` will be `{x: 10, y: 10, w: 30, h: 30}`.
	**/
	final spriteSourceSize:{
		x:Int,
		y:Int,
		w:Int,
		h:Int
	};
	/** Trimmed texture rect cords in atlas.

		Just like `spriteSourceSize`, but on custom atlas `x/y` position.
	**/
	final frame:{
		x:Int,
		y:Int,
		w:Int,
		h:Int
	};
}

typedef AtlasData = {
	?image:String,
	?format:String,
	?size:{w:Int, h:Int},
	?scale:Int,
	?textures:Array<AtlasData>,
	?frames:Array<AtlasFrameData>,
}

class Atlas {
	public static final textures:Array<String> = [];
	public static final atlases:Map<String, Array<AtlasFrameData>> = [];
	#if kha
	/** `Assets.images.get(id)` links **/
	public static final images:Map<String, Image> = [];
	#end

	#if kha
	public static function parseAtlases(ids:Array<String>):Void {
		for (id in ids) {
			parseAtlas(id);
		}
	}

	public static function getImage(id:ResItem):Image {
		var texture = Assets.images.get(id.texture);
		texture ??= Atlas.images[id.texture];
		return texture;
	}

	public static function parseAtlas(id:String):Void {
		final obj:AtlasData = Json.parse(Assets.blobs.get(id).toString());
		final name = ~/atlases_(res.*)_json/g.replace(id, "$1");
		if (!textures.contains(name)) textures.push(name);
		images[name] = Assets.images.get(id.replace("_json", ""));
		atlases[name] = obj.textures == null ? obj.frames : obj.textures[0].frames;
	}
	#end

	public static function loadBase64(name:String, imgBase64:Bytes, data:AtlasData, onComplete:() -> Void):Void {
		if (!textures.contains(name)) textures.push(name);
		Atlas.atlases[name] = data.textures == null ? data.frames : data.textures[0].frames;
		#if kha
		// final img = Image.fromImage(img, false);
		Image.fromEncodedBytes(imgBase64, "png", img -> {
			#if js
			Reflect.setField(Assets.images, name, img);
			#else
			images[name] = img;
			#end
			onComplete();
		}, s -> {
			trace(imgBase64);
			trace(name);
			throw s;
		});
		#elseif phaser
		final img = new Image();
		img.onload = () -> {
			// final data = Json.parse(getResourceString('${item.path}.json'));
			App._scene.load.textureManager.addAtlas(name, img, data);
			onComplete();
		}
		final pngBase64 = "data:image/png;base64,";
		img.src = pngBase64 + Base64.encode(imgBase64);
		#end
	}

	public static function loadImageBase64(name:String, imgBase64:Bytes, onComplete:() -> Void):Void {
		if (!textures.contains(name)) textures.push(name);
		#if kha
		Image.fromEncodedBytes(imgBase64, "png", img -> {
			#if js
			Reflect.setField(Assets.images, name, img);
			#else
			images[name] = img;
			#end
			onComplete();
		}, s -> {
			throw s;
		});
		#elseif phaser
		final img = new Image();
		img.onload = () -> {
			App._scene.load.textureManager.addImage(name, img);
			onComplete();
		}
		final pngBase64 = "data:image/png;base64,";
		img.src = pngBase64 + Base64.encode(imgBase64);
		#end
	}

	// public static function parseAtlasFromString(data:String) {
	// 	final obj:AtlasData = Json.parse(data);
	// 	atlases[name] = obj.textures == null ? obj.frames : obj.textures[0].frames;
	// }

	public static function hasAtlas(texture:String):Bool {
		return atlases[texture] != null;
	}

	public static function hasTexture(texture:String):Bool {
		return textures.contains(texture);
	}

	public static function getAtlasFrame(id:ResItem):AtlasFrameData {
		final atlasArray = atlases[id.texture];
		if (atlasArray == null) throw 'Atlas ${id.texture} not found, see `App.parseAtlases`';
		final frame = atlasArray.inlineFind(item -> item.filename == id.frame);
		if (frame == null) throw 'Frame ${id.frame} not found in atlas ${id.texture}';
		return frame;
	}
}
