package khm.utils;

import haxe.Json;
import haxe.Resource;
import haxe.crypto.Base64;
import haxe.io.Bytes;
import haxe.io.Path;
#if js
import js.Browser.document;
import js.Browser.window;
#end

private typedef ItemsData = Array<{
	id:String,
	path:String,
	data:Bytes,
	?json:String,
}>;

private typedef AudioData = {
	id:String,
	path:String,
	audio:Any
};

private typedef Base64LoaderConfig = {
	/**
		List of image names to load. By default loads all resources with `bg` prefix

		Example `--resource res/bg1.jpg@bg1` > `bg1`
	**/
	var ?images:Array<String>;

	/**
		List of atlases to load. By default loads all resources with `res` prefix

		Example: `--resource res/res.png` (with `res/res.json`) > `{id: "res", path: "res/res"}`
	**/
	var ?atlases:Array<{id:String, path:String}>;

	/**
		List of audios to load. By default loads `audio.mp3` and `audio*.mp3` (`*` is any characters)

		If there is `no_sound` define, then no sounds are loaded by default
	**/
	var ?audios:Array<{id:String, path:String}>;
	var ?filterImages:(name:String) -> Bool;
	var ?filterAtlases:(name:String) -> Bool;
	var onImagesLoaded:(config:Base64LoaderConfig) -> Void;
	var onAudioInited:() -> Void;
	var onComplete:() -> Void;
	var ?completeBeforeAudio:Bool;
}

class Base64Loader {
	public static function loadBase64Resources(scene:Scene, config:Base64LoaderConfig):Void {
		final listNames = Resource.listNames();
		config.filterImages ??= name -> name.startsWith("bg");
		config.filterAtlases ??= name -> true;
		if (config.images == null) {
			config.images = listNames.filter(config.filterImages);
		}
		if (config.atlases == null) {
			final atlases = listNames.filter(path -> {
				final name = Path.withoutDirectory(path);
				final isRes = name.startsWith("res") && name.endsWith(".png");
				if (isRes) return config.filterAtlases(name);
				return false;
			}).map(path -> path.replace(".png", ""));
			config.atlases = [
				for (path in atlases)
					{
						id: Path.withoutDirectory(path),
						path: path
					}
			];
		}
		if (config.audios == null) {
			#if no_sound
			final audios = [];
			#else
			final audios = listNames.filter(path -> {
				final name = Path.withoutDirectory(path);
				return name.startsWith("audio") && name.endsWith(".mp3");
			}).map(path -> path.replace(".mp3", ""));
			#end
			config.audios = [
				for (path in audios)
					{
						id: Path.withoutDirectory(path),
						path: path
					}
			];
		}
		config.completeBeforeAudio ??= false;
		if (config.audios.length > 0)
			@:privateAccess Lazyr.defaultAudio ??= config.audios[0].id;

		var imageData:ItemsData = [
			for (atlas in config.atlases)
				{
					id: atlas.id,
					path: atlas.path,
					data: getResourceBytes('${atlas.path}.png'),
					json: getResourceString('${atlas.path}.json'),
				}
		];
		imageData = imageData.concat([
			for (item in config.images)
				{
					id: item,
					path: item,
					data: getResourceBytes(item),
				}
		]);
		// loadFonts(obj.fonts);
		loadImages(imageData, () -> {
			config.onImagesLoaded(config);
			#if (phaser && !debug)
			App._scene.sound.asType0.pauseOnBlur = false;
			#end
			final audioData = [
				for (audio in config.audios)
					{
						id: audio.id,
						path: audio.path,
						data: getResourceBytes('${audio.path}.mp3')
					}
			];
			#if kha
			final isStrictAudio = true;
			#if !kha_html5
			final onComplete = config.onComplete;
			config.onComplete = () -> Lazyr.delayed(1, () -> onComplete());
			#end
			#elseif phaser
			final isStrictAudio = App._scene.sys.game.device.os.iOS || true;
			#end
			if (isStrictAudio) {
				#if kha
				#elseif phaser
				var audioEnabled = false;
				function onPointer(p):Void {
					if (audioEnabled) return;
					audioEnabled = true;
					loadAudios(audioData, item ->
						onAudioItemLoaded(scene, item), config.onAudioInited);
				}
				App._scene.input.on("pointerdown", onPointer);
				App._scene.input.on("pointerup", onPointer);
				#end
				config.onComplete();
			} else {
				if (config.completeBeforeAudio) config.onComplete();
				loadAudios(audioData, item -> onAudioItemLoaded(scene, item), () -> {
					config.onAudioInited();
					if (config.completeBeforeAudio) return;
					config.onComplete();
				});
			}
		});
	}

	static function getResourceString(name:String):String {
		return Resource.getString(name) ?? throw 'Resource $name is null';
	}

	static function getResourceBytes(name:String):Bytes {
		return Resource.getBytes(name) ?? throw 'Resource $name is null';
	}

	static function onAudioItemLoaded(scene:Scene, item:AudioData):Void {
		#if kha
		// todo
		#elseif phaser
		final _scene = @:privateAccess App._scene;
		_scene.cache.json.add(item.id, Json.parse(getResourceString('${item.path}.json')));
		_scene.cache.audio.add(item.id, item.audio);
		#end
	}

	#if js
	public static function loadFonts(items:ItemsData):Void {
		if (items == null) return;
		/* @font-face {
			font-family: 'DINCondensed-Bold';
			src: url('data:application/font-woff2;charset=utf-8;base64,QAAA==') format('woff2');
			font-weight: normal;
			font-style: normal;
		}*/
		for (item in items) {
			final fontBase64 = "data:application/font-woff2;charset=utf-8;base64,";
			var newStyle = document.createElement('style');
			newStyle.appendChild(document.createTextNode('
				@font-face {
					font-family: "${item.id}";
					src: url("${fontBase64 + Base64.encode(item.data)}") format("woff2");
					font-weight: normal;
					font-style: normal;
				}
			'));
			document.head.appendChild(newStyle);
			// <div style="font-family:'DINCondensed-Bold';visibility: hidden;position: relative;height: 0;">text</div>
			final fontDiv = document.createDivElement();
			fontDiv.style.fontFamily = item.id;
			fontDiv.style.visibility = "hidden";
			fontDiv.style.position = "relative";
			fontDiv.style.height = "0";
			fontDiv.innerHTML = "text";
			document.body.insertBefore(fontDiv, document.body.children[0]);
		}
	}
	#end

	public static function loadImages(
		items:ItemsData,
		onComplete:() -> Void
	):Void {
		if (items.length == 0) onComplete();
		var count = 0;
		for (item in items) {
			if (item.json == null) {
				Atlas.loadImageBase64(item.id, item.data, () -> {
					count++;
					if (count == items.length) onComplete();
				});
				continue;
			}
			Atlas.loadBase64(item.id, item.data, Json.parse(item.json), () -> {
				count++;
				if (count == items.length) onComplete();
			});
		}
	}

	public static function loadAudios(
		arr:ItemsData,
		onItemLoaded:(audio:AudioData) -> Void,
		?onComplete:() -> Void
	):Void {
		#if kha
		trace("loadAudios not impl for kha");
		onComplete();
		#elseif phaser
		if (arr.length == 0) {
			if (onComplete != null) onComplete();
			return;
		}
		final item = arr.shift();
		try {
			final audioCtx = js.Syntax.code("new (window.AudioContext || window.webkitAudioContext)()");
			final toBuffer:(s:String) -> Any = untyped Phaser.Utils.Base64.Base64ToArrayBuffer;
			final mp3Base64 = "data:audio/mpeg;base64,";
			final data = mp3Base64 + Base64.encode(item.data);
			audioCtx.decodeAudioData(toBuffer(data), buffer -> {
				onItemLoaded({id: item.id, path: item.path, audio: buffer});
				loadAudios(arr, onItemLoaded, onComplete);
			}, e -> {
				trace('Error with decoding audio ${item.id}: $e');
				loadAudios(arr, onItemLoaded, onComplete);
			});
		} catch (e:Any) {
			trace('Error before decoding audio ${item.id}: ${e}');
			loadAudios(arr, onItemLoaded, onComplete);
		}
		#end
	}
}
