package khm.verlet;

import haxe.EnumFlags;
import khm.math.Degrees;
import khm.math.Vector2;

typedef DotOptions = {x:Float, y:Float, ?fixed:Bool}

typedef ConstraintOptions = {
	indexA:Int,
	indexB:Int,
	?stiffness:Float,
	?edge:Bool
}

@:structInit
class BodyOptions {
	public var dots:Array<DotOptions>;
	public var constraints:Array<ConstraintOptions>;
	/** Generate clockwise edges in `constraints` array for polygons. **/
	public var autoEdges = false;
	public var type:BodyType = Polygon;
	public var x = 0.0;
	public var y = 0.0;
	public var mass = 1.0;
	public var radius:Null<Float> = null;
	/** Code line pos infos for debug purposes. **/
	public var posInfos:DebugPosInfos = null;
}

enum abstract BodyType(Int) {
	// var Point;
	// var Triangle;
	var Polygon;
	var Circle;
}

class Body {
	static var allowDuplicateConstraints = false;

	/** Code line pos infos for debug purposes. **/
	var posInfos:DebugPosInfos = null;

	public final world:World;
	public final type:BodyType;
	public final dots:Array<Dot> = [];
	public final constraints:Array<Constraint> = [];
	public var layers:EnumFlags<Dynamic> = new EnumFlags();
	public var ignoreLayers:EnumFlags<Dynamic> = new EnumFlags();
	/** External constraints, out of current body structure **/
	public final joints:Array<Constraint> = [];
	/** Edge lines of body, part from `constraints` array. Used to check collision with SAT. **/
	public final edges:Array<Constraint> = [];
	/** Centroid dot of this convex body **/
	public final center = new Vector2(0, 0);
	public final aabbCenter = new Vector2(0, 0);
	/** Half of AABB size for broad collision phase. **/
	public final halfSize = new Vector2(0, 0);
	public var min = 0.0;
	public var max = 0.0;
	/** `mass` always should be bigger that `0.0` **/
	public var mass:Float;
	/** Custom `state` object for your data and game object references **/
	// public var state:{?toJson:() -> String} = {};
	public var state:Any = {};

	// public var fixed = false;
	@:isVar public var fixed(get, set) = false;

	inline function get_fixed():Bool {
		return this.fixed;
	}

	inline function set_fixed(fixed:Bool):Bool {
		world.dynBodies.remove(this);
		if (!fixed) world.dynBodies.push(this);
		return this.fixed = fixed;
	}

	public var customResolve:(col:Collision) -> Void = null;

	public var onMove:(x:Float, y:Float, angle:Degrees) -> Degrees = null;

	public function new(world:World, opts:BodyOptions) {
		this.world = world;
		this.type = opts.type;
		this.posInfos = opts.posInfos;
		mass = opts.mass;

		final x = opts.x;
		final y = opts.y;
		for (config in opts.dots) {
			final dot = addDot(x + config.x, y + config.y, config.fixed);
		}
		if (dots.length > 0 && isAllDotsFixed()) fixed = true;

		if (opts.autoEdges) generateEdges();

		for (config in opts.constraints) {
			addConstraint(dots[config.indexA], dots[config.indexB], config.stiffness, config.edge);
		}
		updateBoundingBox();
	}

	public inline function isAllDotsFixed():Bool {
		return !dots.inlineExists(dot -> !dot.fixed);
	}

	function generateEdges(stiffness = 1.0):Void {
		if (!allowDuplicateConstraints && dots.length == 2) {
			addConstraint(dots[0], dots[1], stiffness, true);
			return;
		}
		for (i => dotA in dots) {
			final dotB = dots[i + 1] ?? dots[0];
			addConstraint(dotA, dotB, stiffness, true);
		}
	}

	public function addDot(x:Float, y:Float, fixed = false):Dot {
		final dot = new Dot(this, x, y, fixed);
		dots.push(dot);
		world.dots.push(dot);
		return dot;
	}

	public function removeDot(dot:Dot):Void {
		removeConstraintsByDot(dot);
		dots.remove(dot);
		world.dots.remove(dot);
		#if debug
		if (dots.contains(dot)) throw "constrains has copy of removed constraint";
		#end
	}

	public function removeConstraintsByDot(dot:Dot):Void {
		for (c in constraints.reversed()) {
			if (c.dotA != dot && c.dotB != dot) continue;
			removeConstraint(c);
		}
		for (joint in joints.reversed()) {
			if (joint.dotA != dot && joint.dotB != dot) continue;
			removeJoint(joint);
		}
	}

	public function removeConstraint(c:Constraint):Void {
		constraints.remove(c);
		edges.remove(c);
		// final bodyA = c.dotA.body;
		// final bodyB = c.dotB.body;
		// bodyA.constraints.remove(c);
		// bodyB.constraints.remove(c);
		// bodyA.edges.remove(c);
		// bodyB.edges.remove(c);

		world.constraints.remove(c);
		#if debug
		if (constraints.contains(c)) throw "constrains has copy of removed constraint";
		return;
		for (body in world.bodies) {
			if (body.edges.contains(c)) {
				trace(world.bodies.indexOf(body), body.edges.indexOf(c));
			}
		}
		#end
	}

	/** Removes joint constraint from both dots bodies **/
	public function removeJoint(c:Constraint):Void {
		final bodyA = c.dotA.body;
		final bodyB = c.dotB.body;
		bodyA.joints.remove(c);
		bodyB.joints.remove(c);
		world.constraints.remove(c);
	}

	public function destroy():Void {
		for (c in joints.reversed()) removeJoint(c);
		for (dot in dots.reversed()) removeDot(dot);
		world.bodies.remove(this);
		world.dynBodies.remove(this);
	}

	public function mergeDot(fromDot:Dot, toDot:Dot):Void {
		if (fromDot == toDot) throw "merge of same dot";
		for (c in fromDot.body.constraints.reversed()) {
			if (c.dotA != fromDot && c.dotB != fromDot) continue;
			final dotA = c.dotA == fromDot ? toDot : c.dotA;
			final dotB = c.dotB == fromDot ? toDot : c.dotB;
			final body = toDot.body;
			// for (c in body.constraints) {
			// if (c.dotA == fromDot && c.dotB == toDot) trace(1);
			// if (c.dotA == toDot && c.dotB == fromDot) trace(2);
			// }
			body.addConstraint(dotA, dotB, c.stiffness, c.isEdge);
		}
		fromDot.body.removeDot(fromDot);
		toDot.body.removeDot(fromDot);
		#if debug
		if (dots.contains(fromDot)) throw "constrains has copy of removed constraint";
		for (c in world.constraints.reversed()) {
			if (c.dotA != fromDot && c.dotB != fromDot) continue;
			// c.body.removeConstraint(c);
		}
		// final c = world.constraints.inlineFind(item -> {
		// 	return item.dotA == fromDot || item.dotB == fromDot;
		// });
		// if (c != null) {
		// 	trace(fromDot);
		// 	trace(c);
		// 	// throw "Constraint with merged dots still exist";
		// }
		#end

		// final sameFromConstraints = fromDot.body.constraints.filter(c -> {
		// 	toDot.body.constraints.inlineExists(c2 -> {
		// 		final isSame = c.dotA == c2.dotA && c.dotB == c2.dotB;
		// 		final isSame2 = c.dotA == c2.dotB && c.dotB == c2.dotA;
		// 		return isSame || isSame2;
		// 	});
		// });
		// trace(sameFromConstraints);
		// for (c in sameFromConstraints.reversed()) {
		// 	fromDot.body.removeConstraint(c);
		// 	toDot.body.removeConstraint(c);
		// }
	}

	public function addConstraint(dotA:Dot, dotB:Dot, stiffness = 1.0, isEdge = false):Constraint {
		final constraint = new Constraint(
			this,
			dotA,
			dotB,
			stiffness,
			isEdge
		);
		if (isEdge) edges.push(constraint);
		constraints.push(constraint);
		#if debug
		if (world.constraints.inlineExists(item -> {
			(item.dotA == dotA && item.dotB == dotB)
			|| (item.dotA == dotB && item.dotB == dotA);
		})) {
			if (!allowDuplicateConstraints)
				throw "Constraint between these two dots already exist";
		}
		#end
		world.constraints.push(constraint);
		return constraint;
	}

	public function updateBoundingBox():Void {
		if (dots.length == 0) return;
		var minX = World.MAX_FLOAT;
		var minY = World.MAX_FLOAT;
		var maxX = -World.MAX_FLOAT;
		var maxY = -World.MAX_FLOAT;
		var sumX = 0.0;
		var sumY = 0.0;
		for (dot in dots) {
			final p = dot.pos;
			#if debug
			if (p == null) throw this;
			#end
			if (p.x > maxX) maxX = p.x;
			if (p.x < minX) minX = p.x;
			if (p.y > maxY) maxY = p.y;
			if (p.y < minY) minY = p.y;
			sumX += p.x;
			sumY += p.y;
		}
		center.x = sumX / dots.length;
		center.y = sumY / dots.length;
		halfSize.x = (maxX - minX) * 0.5;
		halfSize.y = (maxY - minY) * 0.5;
		aabbCenter.x = (minX + maxX) * 0.5;
		aabbCenter.y = (minY + maxY) * 0.5;
	}

	public function isPointInsideAabb(x:Float, y:Float):Bool {
		final offX = Math.abs(x - aabbCenter.x) - halfSize.x;
		if (offX >= 0) return false;
		final offY = Math.abs(y - aabbCenter.y) - halfSize.y;
		if (offY >= 0) return false;
		return true;
	}

	public function isRectInsideAabb(x:Float, y:Float, w:Float, h:Float):Bool {
		final x = x + w / 2;
		final y = y + h / 2;
		final offX = Math.abs(x - aabbCenter.x) - (w / 2 + halfSize.x);
		if (offX >= 0) return false;
		final offY = Math.abs(y - aabbCenter.y) - (h / 2 + halfSize.y);
		if (offY >= 0) return false;
		return true;
	}

	public function projectAxis(axis:Vector2):Void {
		var d = dots[0].pos.dot(axis);
		min = d;
		max = d;
		for (i in 1...dots.length) {
			d = dots[i].pos.dot(axis);
			if (d > max) max = d;
			if (d < min) min = d;
		}
	}

	public inline function setGravity(v:Vector2):Void {
		for (dot in dots) dot.setGravity(v);
	}

	public function resetGravity():Void {
		for (dot in dots) dot.resetGravity();
	}

	public function isEdgeDot(dot:Dot):Bool {
		for (c in edges) {
			if (c.dotA == dot || c.dotB == dot) return true;
		}
		return false;
	}

	public function rotateBy(delta:Degrees) {
		for (dot in dots) {
			final angle = MathLib.angle.betweenPoints(center, dot.pos) + delta;
			final dist = center.sub(dot.pos).length;
			final oldSpeedX = dot.speedX;
			final oldSpeedY = dot.speedY;
			dot.pos.x = center.x + angle.cos() * dist;
			dot.pos.y = center.y + angle.sin() * dist;
			dot.speedX = 0;
			dot.speedY = 0;
		}
	}

	#if kha
	public function renderDots(g:kha.graphics2.Graphics):Void {
		g.color = 0xFFB8A28A;
		g.fillRect(center.x - 2, center.y - 2, 4, 4);

		final edges = world.renderEdgesOnly ? edges : constraints;
		#if debug
		if (edges.length == 0) {
			if (dots[0] == null) throw "body with zero dots should be removed from world";
			else drawDot(g, dots[0]);
		}
		#end
		for (i => edge in edges) {
			// final p = edge.dotA.pos;
			// final p2 = edge.dotB.pos;
			// final rest = p.x == edge.dotA.prevPos.x && p.y == edge.dotA.prevPos.y;
			// final color:Color = rest ? 0xFFD8BABA : 0xFFFF8888;
			// g.color = color;
			// g.drawLine(p.x, p.y, p2.x, p2.y, 2);

			// drawDotSpeed(g, edge.dotA);
			drawDot(g, edge.dotA);

			if (i < edges.length - 1) continue;
			// drawDotSpeed(g, edge.dotB);
			drawDot(g, edge.dotB);
		}
	}

	public function renderConstraints(g:kha.graphics2.Graphics):Void {
		final edges = world.renderEdgesOnly ? edges : constraints;
		for (i => edge in edges) {
			final p = edge.dotA.pos;
			final p2 = edge.dotB.pos;
			final rest = p.x == edge.dotA.prevPos.x && p.y == edge.dotA.prevPos.y;
			final color:Color = rest ? 0xFFD8BABA : 0xFFFF8888;
			g.color = color;
			g.drawLine(p.x, p.y, p2.x, p2.y, 2);

			drawDotSpeed(g, edge.dotA);
			// drawDot(g, edge.dotA);

			if (i < edges.length - 1) continue;
			drawDotSpeed(g, edge.dotB);
			// drawDot(g, edge.dotB);
		}
	}

	inline function drawDot(g:kha.graphics2.Graphics, dot:Dot):Void {
		final color:Color = 0xFFFF8888;
		g.color = color;
		final p = dot.pos;
		g.fillRect(p.x - 2, p.y - 2, 4, 4);
	}

	function drawDotSpeed(g:kha.graphics2.Graphics, dot:Dot):Void {
		if (!world.renderDotSpeed) return;
		final p = dot.pos;
		final s = {
			x: dot.speedX,
			y: dot.speedY,
		};
		final angle = MathLib.angle.between(s.x, s.y, 0, 0);
		final dist = MathLib.dist.between(s.x, s.y, 0, 0);
		final rad = angle.toRad();
		var color:Color = 0xFF88FF92;
		g.color = color;
		g.drawLine(
			p.x,
			p.y,
			p.x + s.x + rad.cos() * dist * 5,
			p.y + s.y + rad.sin() * dist * 5,
			2
		);
	}
	#end
}
