package khm.verlet;

import game.Game;
import khm.math.Vector2;

@:structInit
class CollisionResponse {
	/**
		Collision response vector normal.
		Also useful for friction in resolve.

		`axis.mult(depth)` is `response` vector.
	**/
	public var axis:Vector2;
	/** How deep is penetration **/
	public var depth:Float;
	public var edge:Null<Constraint>;
	public var closestDot:Null<Dot>;
	public var edgeRatio:Float;
	/** All collision edges (only for poly-circle collision) **/
	public var edges:Array<Constraint>;
	/** Only for poly-circle collision **/
	public var circleOnEdgeDot:Vector2;
	/** Used for friction calculation **/
	public var edgeMassRatio:Float;
	public var pointMassRatio:Float;
	public var lambda:Float;
}

class Collision {
	// axis cache for sat check step
	final testAxis = new Vector2(0, 0);

	// for visual debug
	public final lastEdges:Array<Constraint> = [];
	public final lastClosestDots:Array<Dot> = [];
	public final lastPoints:Array<Vector2> = [];
	public final lastAxis:Array<{
		dot:Dot,
		edge:Constraint,
		axis:Vector2,
		depth:Float
	}> = [];

	// cache for resolve
	public final result:CollisionResponse = {
		axis: new Vector2(0, 0),
		depth: 0.0,
		edge: null,
		closestDot: null,
		edgeRatio: 0,
		edges: [],
		circleOnEdgeDot: new Vector2(0, 0),
		pointMassRatio: 0,
		edgeMassRatio: 0,
		lambda: 0
	}

	/**
		Propagate collision resolution after `body.customResolve`
		to execute default response behavior on `body`
	**/
	public var propagate = false;

	// var pointMassRatio:Float;
	var world:World;

	public function new(world:World) {
		this.world = world;
	}

	public var satCount = 0;
	public var debugBody:Body = null;

	public function isAabb(body:Body, body2:Body):Bool {
		final offX = Math.abs(body2.aabbCenter.x - body.aabbCenter.x) - (body2.halfSize.x +
			body.halfSize.x);
		if (offX >= 0) return false;
		final offY = Math.abs(body2.aabbCenter.y - body.aabbCenter.y) - (body2.halfSize.y +
			body.halfSize.y);
		if (offY >= 0) return false;
		return true;
	}

	public function sat(body:Body, body2:Body):Null<CollisionResponse> {
		if (!isAabb(body, body2)) return null;
		satCount++;

		var minDistance = World.MAX_FLOAT;
		final edgeCount = body.edges.length;
		final edgeCount2 = body2.edges.length;
		result.edge = null;
		var parallelEdge:Constraint = null;

		for (i in 0...edgeCount + edgeCount2) {
			final edge:Constraint = i < edgeCount ? body.edges[i] : body2.edges[i - edgeCount];
			// place both position on test axis
			normal(testAxis, edge.dotA.pos, edge.dotB.pos);
			// update min/max for bodies
			body.projectAxis(testAxis);
			body2.projectAxis(testAxis);
			// calculate minimal distance between bodies
			// var dist:Float = body.min < body2.min ? body2.min - body.max : body.min - body2.max;
			final isNotOverlaps = body.min > body2.max || body2.min > body.max;
			if (isNotOverlaps) return null;
			var dist = Math.min(body.max, body2.max) - Math.max(body.min, body2.min);

			final c1 = body.min > body2.min && body.max < body2.max;
			final c2 = body2.min > body.min && body2.max < body.max;
			if (c1 || c2) {
				// if containment exists then get the overlap plus the distance
				// to between the two end points that are the closest
				final max = (body.max - body2.max).abs();
				final min = (body.min - body2.min).abs();
				if (max > min) {
					// if the min differences is less than the max then we need
					// to flip the penetration axis
					testAxis.x = -testAxis.x;
					testAxis.y = -testAxis.y;
					dist += min;
				} else {
					dist += max;
				}
			}
			if (dist < 0) return null;

			if (dist <= minDistance) {
				final isSameDist = dist == minDistance;
				minDistance = dist;
				result.axis.setFrom(testAxis);
				result.edge = edge;
				// keep parallel edge with same dist
				if (!isSameDist) parallelEdge = edge;
			}
		}
		result.depth = minDistance;

		if (result.edge.body != body2) {
			final dotBody = body2;
			body2 = body;
			body = dotBody;
		}

		final dotBody = body;
		final edgeBody = body2;
		final center = dotBody.center.sub(edgeBody.center);
		final n = center.dot(result.axis);
		if (n < 0) {
			result.axis.x = -result.axis.x;
			result.axis.y = -result.axis.y;
		} else {
			if (parallelEdge != null) result.edge = parallelEdge;
		}

		lastEdges.push(result.edge);

		if (dotBody.edges.length > 1) {
			// compare dotA dist for every body edge
			result.closestDot = dotBody.edges.findMin(edge -> {
				final dot = edge.dotA;
				final line = dot.pos.sub(edgeBody.center);
				final dist = result.axis.dot(line);
				return dist;
			}, World.MAX_FLOAT).dotA;
		} else if (dotBody.edges.length == 1) {
			// check both edge dots for lines (they don't wrap around)
			final edge = dotBody.edges[0];
			final dotADist = result.axis.dot(edge.dotA.pos.sub(edgeBody.center));
			final dotBDist = result.axis.dot(edge.dotB.pos.sub(edgeBody.center));
			result.closestDot = dotADist < dotBDist ? edge.dotA : edge.dotB;
		} else {
			// no edges, this is single dot body
			result.closestDot = dotBody.dots[0];
		}

		lastAxis.push({
			dot: result.closestDot,
			edge: result.edge,
			axis: result.axis.clone(),
			depth: result.depth
		});

		lastClosestDots.push(result.closestDot);

		return result;
	}

	function normal(axis:Vector2, a:Vector2, b:Vector2):Vector2 {
		// perpendicular
		final nx = a.y - b.y;
		final ny = b.x - a.x;
		// normalize
		final len = 1.0 / Math.sqrt(nx * nx + ny * ny);
		axis.x = nx * len;
		axis.y = ny * len;
		// final n = new Vector2(a.y - b.y, b.x - a.x);
		// axis.setFrom(n.normalized());
		return axis;
	}

	inline function checkCustomResolve(result:CollisionResponse):Bool {
		var isExit = false;
		if (result.edge.body.customResolve != null) {
			propagate = false;
			result.edge.body.customResolve(this);
			if (!propagate) isExit = true;
		}
		if (isExit) return isExit;
		if (result.closestDot.body.customResolve != null) {
			propagate = false;
			result.closestDot.body.customResolve(this);
			if (!propagate) isExit = true;
		}
		return isExit;
	}

	public function resolve(?result:CollisionResponse):Void {
		final result = result ?? this.result;
		final posA = result.edge.dotA.pos;
		final posB = result.edge.dotB.pos;
		final dotPos = result.closestDot.pos;

		final resp = result.axis.mult(result.depth);

		final isHorizontalCol = Math.abs(posA.x - posB.x) > Math.abs(posA.y - posB.y);
		// calculate where on the edge the collision point lies (0.0 - 1.0)
		// if-check prevents divide by zero
		final edgeRatio = if (isHorizontalCol) {
			(dotPos.x - resp.x - posA.x) / (posB.x - posA.x);
		} else {
			(dotPos.y - resp.y - posA.y) / (posB.y - posA.y);
		}

		final edgeBody = result.edge.body;
		final pointBody = result.closestDot.body;
		final invTotalMass = 1.0 / (edgeBody.mass + pointBody.mass);
		final pointMassRatio = edgeBody.fixed ? 1.0 : edgeBody.mass * invTotalMass;
		final edgeMassRatio = pointBody.fixed ? 1.0 : pointBody.mass * invTotalMass;

		// calculate a scaling factor that ensures that
		// the point lies on the edge after the collision response
		final t = edgeRatio;
		final lambda = 0.1 / (t * t + (1 - t) * (1 - t));

		result.lambda = lambda;
		result.edgeRatio = edgeRatio;
		result.pointMassRatio = pointMassRatio;
		result.edgeMassRatio = edgeMassRatio;

		if (checkCustomResolve(result)) return;

		inline function minLimit(v):Float {
			if (Math.abs(v) < 0.05) v = 0;
			return v;
		}
		inline function maxLimit(v, max):Float {
			final a = Math.abs(v);
			if (a > max) v = max * (v / a);
			return v;
		}

		final lambdaEdge = lambda * edgeMassRatio;
		if (!result.edge.dotA.fixed) {
			posA.x -= resp.x * (1 - t) * lambdaEdge;
			posA.y -= resp.y * (1 - t) * lambdaEdge;
		}
		if (!result.edge.dotB.fixed) {
			posB.x -= resp.x * t * lambdaEdge;
			posB.y -= resp.y * t * lambdaEdge;
		}
		if (!result.closestDot.fixed) {
			dotPos.x += minLimit(resp.x * pointMassRatio);
			dotPos.y += minLimit(resp.y * pointMassRatio);
			// if (closestDot.body == debugBody) trace(resp);
		}

		applyFriction();
	}

	function applyFriction():Void {
		final r = result;
		final posA = r.edge.dotA.pos;
		final posB = r.edge.dotB.pos;
		final prevPosA = r.edge.dotA.prevPos;
		final prevPosB = r.edge.dotB.prevPos;
		final dotPos = r.closestDot.pos;
		final dotPrevPos = r.closestDot.prevPos;

		final relVel = new Vector2(
			dotPos.x - dotPrevPos.x - (posA.x + posB.x - prevPosA.x - prevPosB.x) * 0.5,
			dotPos.y - dotPrevPos.y - (posA.y + posB.y - prevPosA.y - prevPosB.y) * 0.5
		);

		final tangent = new Vector2(-r.axis.y, r.axis.x);

		// apply friction
		final relTv:Float = relVel.dot(tangent);
		final relTanVel = tangent.mult(relTv);
		final rt:Vector2 = relTanVel;
		final t = r.edgeRatio;
		final friction = world.friction / world.iterations;
		if (!r.closestDot.fixed) {
			dotPrevPos.x += rt.x * friction * result.pointMassRatio;
			dotPrevPos.y += rt.y * friction * result.pointMassRatio;
		}
		final lambdaEdge = r.lambda * r.edgeMassRatio;
		if (!r.edge.dotA.fixed) {
			prevPosA.x -= rt.x * (1 - t) * friction * lambdaEdge;
			prevPosA.y -= rt.y * (1 - t) * friction * lambdaEdge;
		}
		if (!r.edge.dotB.fixed) {
			prevPosB.x -= rt.x * t * friction * lambdaEdge;
			prevPosB.y -= rt.y * t * friction * lambdaEdge;
		}

		if (world.resolveDebugger) {
			world.resolveDebugger = false;
			#if js js.Lib.debug(); #end
		}
	}

	public function pointInPolygon(px:Float, py:Float, body:Body):Bool {
		var isInside = false;
		for (edge in body.edges) {
			final posA = edge.dotA.pos;
			final posB = edge.dotB.pos;
			if (((posA.y > py && posB.y < py) || (posA.y < py && posB.y > py))
				&& (px < (posB.x - posA.x) * (py - posA.y) / (posB.y - posA.y) + posA.x)) {
				isInside = !isInside;
			}
		}
		return isInside;
	}

	public static inline function closestPointOnLineSegment(p:Vector2, start:Vector2, end:Vector2):Vector2 {
		final line = end.sub(start);
		final dot = line.dot(p.sub(start));
		final lengthSquared = line.x * line.x + line.y * line.y;
		var param = -1.0;
		// in case of 0 length line
		if (lengthSquared != 0) param = dot / lengthSquared;

		final result = new Vector2(start.x + param * line.x, start.y + param * line.y);
		if (param < 0) result.setFrom(start);
		else if (param > 1) result.setFrom(end);
		return result;
	}

	public function solveCircleOnPolygon(circle:Circle, polygonBody:Body):Bool {
		if (!isAabb(circle, polygonBody)) return false;
		satCount++;
		final circlePos = circle.dot.pos;
		final radius = circle.radius;
		var minDist = World.MAX_FLOAT;
		final closest = new Vector2(0, 0);
		final isInside = pointInPolygon(circle.center.x, circle.center.y, polygonBody);
		// Find closest point within radius of circle on rigid body
		result.edge = null;
		result.edges.resize(0);
		for (edge in polygonBody.edges) {
			final point = closestPointOnLineSegment(circlePos, edge.dotA.pos, edge.dotB.pos);
			final delta = circlePos.sub(point);
			final distSquared = delta.x * delta.x + delta.y * delta.y;
			if (distSquared < radius * radius || isInside) {
				result.edges.push(edge);
				if (distSquared < minDist) {
					minDist = distSquared;
					result.edge = edge;
					closest.setFrom(point);
				}
			}
		}
		if (result.edge == null) return false;
		// Calculate collision normal and penetration depth
		final delta = closest.sub(circlePos);
		final dist = delta.length;

		result.depth = dist - radius;
		result.axis.setFrom(delta.div(dist));

		if (isInside) {
			result.axis.x = -result.axis.x;
			result.axis.y = -result.axis.y;
			result.depth = -result.depth - radius * 2;
		}

		if (world.renderCircles) lastEdges.push(result.edge);

		// Calculate mass ratios, friction restitution coefficients
		final invTotalMass = 1.0 / (polygonBody.mass + circle.mass);
		final circleBodyRatio = polygonBody.fixed ? 1.0 : polygonBody.mass * invTotalMass;
		final rigidBodyRatio = circle.fixed ? 1.0 : circle.mass * invTotalMass;

		final polyFriction = world.friction;
		final circleFriction = world.friction;
		final frictionCoef = (polyFriction + circleFriction) * 0.5;
		// final coefficientOfRestitution = (polygonBody.restitution + circle.restitution) * 0.5;
		final restitution = 0.5;
		final coefficientOfRestitution = (restitution + restitution) * 0.5;

		final posA = result.edge.dotA.pos;
		final posB = result.edge.dotB.pos;
		final prevPosA = result.edge.dotA.prevPos;
		final prevPosB = result.edge.dotB.prevPos;
		// Calculate where on the edge the collision point lies (0.0 - 1.0)
		// If-check prevents divide by zero.
		final t = if (Math.abs(posA.x - posB.x) > Math.abs(posA.y - posB.y)) {
			(closest.x - posA.x) / (posB.x - posA.x);
		} else {
			(closest.y - posA.y) / (posB.y - posA.y);
		}

		result.circleOnEdgeDot.setFrom(closest);
		if (world.renderCircles) {
			lastClosestDots.push(cast {pos: result.circleOnEdgeDot.clone()});
		}

		result.closestDot = circle.dot;
		result.edgeRatio = t;

		if (checkCustomResolve(result)) return true;

		if (world.renderCircles) lastAxis.push({
			dot: result.closestDot,
			edge: result.edge,
			axis: new Vector2(result.axis.x, result.axis.y),
			depth: result.depth
		});

		final edgeDir = new Vector2(result.axis.y, -result.axis.x);
		if (!polygonBody.fixed) {
			// Calculate a scaling factor that ensures that the point lies on the edge after the collision response
			final lambda = 1.0 / (t * t + (1 - t) * (1 - t)) * rigidBodyRatio;
			final velA = posA.sub(prevPosA);
			final velB = posB.sub(prevPosB);
			// Correct edge point positions
			// Add friction to point A
			if (!result.edge.dotA.fixed) {
				posA.x -= result.axis.x * result.depth * lambda * (1 - t);
				posA.y -= result.axis.y * result.depth * lambda * (1 - t);
				final frictionImpulseA = calculateFrictionImpulse(velA, edgeDir, result.depth, frictionCoef);
				prevPosA.x += frictionImpulseA.x * (1 - t);
				prevPosA.y += frictionImpulseA.y * (1 - t);
			}
			// Add friction to point B
			if (!result.edge.dotB.fixed) {
				posB.x -= result.axis.x * result.depth * lambda * t;
				posB.y -= result.axis.y * result.depth * lambda * t;
				final frictionImpulseB = calculateFrictionImpulse(velB, edgeDir, result.depth, frictionCoef);
				prevPosB.x += frictionImpulseB.x * t;
				prevPosB.y += frictionImpulseB.y * t;
			}
		}
		// Calculate circle velocity before position is corrected
		final circleVel = circlePos.sub(circle.dot.prevPos);
		// If dot product between edge normal and travel direction is more than 0, then the circle is moving towards the edge
		final dot = circleVel.dot(result.axis);
		if (!circle.fixed && dot > 0) {
			// Correct circle position
			circlePos.x += result.axis.x * result.depth * circleBodyRatio;
			circlePos.y += result.axis.y * result.depth * circleBodyRatio;
			// Split velocity vector up into its normal and perpendicular components
			final velNormal = result.axis.mult(dot);
			final velPerp = circleVel.sub(velNormal);
			// Rotational velocity transformed from angle to world space
			final circumference = 2 * Math.PI * radius;
			final prevRotationVelocity = (circle.rotation - circle.prevRotation) * world.viscosity;
			// final prevRotationVelocity = circle.prevRotationVelocity;
			var rotVel = prevRotationVelocity / (2 * Math.PI) * circumference;
			// Signed length of perpendicular velocity
			final xEdgeDir = result.axis.y;
			final yEdgeDir = -result.axis.x;
			final velEdgeDot = (velPerp.x * xEdgeDir + velPerp.y * yEdgeDir).sign();
			final perpendicularVel = velPerp.length * velEdgeDot;

			// todo
			// final friction = world.friction / world.iterations;
			// rotVel = rotVel * 0.9;
			// var perpendicularVel = perpendicularVel * 0.9;
			// if (rotVel.abs() < 0.1) rotVel = 0;
			// if (perpendicularVel.abs() < 0.1) perpendicularVel = 0;

			// Difference between rotational and perpendicular velocity
			final velDiff = (rotVel - perpendicularVel) * frictionCoef;
			// Decrease rotational velocity
			rotVel -= velDiff;
			// if (rotVel.abs() < 0.1) rotVel = 0;

			// Increase perpendicular velocity of circle
			final xVelChange = xEdgeDir * velDiff;
			final yVelChange = yEdgeDir * velDiff;
			velPerp.x += xVelChange;
			velPerp.y += yVelChange;
			// Nudge velocity of rigid body in opposite direction
			if (!polygonBody.fixed) {
				if (!result.edge.dotA.fixed) {
					prevPosA.x += xVelChange * (1 - t);
					prevPosA.y += yVelChange * (1 - t);
				}
				if (!result.edge.dotB.fixed) {
					prevPosB.x += xVelChange * t;
					prevPosB.y += yVelChange * t;
				}
			}
			// Recombine perpendicular and normal velocities reflected about the edge normal
			circle.dot.speedX = velPerp.x - velNormal.x * coefficientOfRestitution;
			circle.dot.speedY = velPerp.y - velNormal.y * coefficientOfRestitution;
			// Update rotational velocity
			circle.prevRotation = circle.rotation - ((rotVel / circumference) * 2 * Math.PI);

			final speed = new Vector2(circle.dot.speedX, circle.dot.speedY);
			if (speed.length < 2) circle.prevRotation = circle.rotation;
		}
		return true;
	}

	inline function calculateFrictionImpulse(
		velocity:Vector2,
		edgeDir:Vector2,
		depth:Float,
		frictionCoef:Float
	):Vector2 {
		final impulse = new Vector2(0, 0);
		// Only add friction impulse if point is in motion
		if (velocity.x != 0 || velocity.y != 0) {
			// Direction of motion normalized
			final len = 1 / velocity.length;
			final velDir = velocity.mult(len);

			// Dot product between velocity and edge direction multiplied by penetration depth and friction coefficient
			final friction = velDir.dot(edgeDir) * depth * frictionCoef;

			// Friction impulse along the negative edge direction
			impulse.x = -edgeDir.x * friction;
			impulse.y = -edgeDir.y * friction;
		}
		return impulse;
	}

	public function solveCircleOnCircle(circleA:Circle, circleB:Circle):Bool {
		final posA = circleA.dot.pos;
		final posB = circleB.dot.pos;
		final radiusA = circleA.radius;
		final radiusB = circleB.radius;

		final delta = posA.sub(posB);
		final distSquared = delta.x * delta.x + delta.y * delta.y;
		final minDist = radiusA + radiusB;

		if (distSquared > minDist * minDist) return false;

		final dist = Math.sqrt(distSquared);
		final depth = minDist - dist;
		final normal = delta.div(dist);

		result.axis.setFrom(normal);
		result.depth = depth;
		result.edge = null;

		// if (checkCustomResolve(result)) return true;
		result.closestDot = circleB.dot;
		var isExit = false;
		// if (result.edge.body.customResolve != null) {
		if (circleA.customResolve != null) {
			propagate = false;
			circleA.customResolve(this);
			if (!propagate) isExit = true;
		}
		if (isExit) return isExit;

		result.closestDot = circleA.dot;
		if (circleB.customResolve != null) {
			propagate = false;
			circleB.customResolve(this);
			if (!propagate) isExit = true;
		}
		if (isExit) return isExit;

		if (world.renderCircles) {
			lastAxis.push({
				dot: circleA.dot,
				edge: result.edge,
				axis: new Vector2(result.axis.x, result.axis.y),
				depth: result.depth
			});
			lastAxis.push({
				dot: circleB.dot,
				edge: result.edge,
				axis: new Vector2(result.axis.x, result.axis.y).mult(-1),
				depth: result.depth
			});
		}

		final invTotalMass = 1.0 / (circleA.mass + circleB.mass);
		final aRatio = circleB.fixed ? 1.0 : circleB.mass * invTotalMass;
		final bRatio = circleA.fixed ? 1.0 : circleA.mass * invTotalMass;

		final circleFriction = world.friction;
		final circleFriction2 = world.friction;
		final frictionCoef = (circleFriction + circleFriction2) * 0.5;
		// final coefficientOfRestitution = (polygonBody.restitution + circle.restitution) * 0.5;
		final restitution = 0.5;
		final coefficientOfRestitution = (restitution + restitution) * 0.5;

		final deltaVel = new Vector2(
			circleA.dot.speedX - circleB.dot.speedX,
			circleA.dot.speedY - circleB.dot.speedY
		);
		final velocityAlongNormal = deltaVel.dot(normal);
		final impulse = velocityAlongNormal * (1.0 + coefficientOfRestitution);

		if (!circleA.fixed) {
			circleUpdate(circleA, normal, depth, impulse, frictionCoef, aRatio);
		}
		if (!circleB.fixed) {
			circleUpdate(circleB, normal.mult(-1), depth, impulse, frictionCoef, bRatio);
		}
		// final speed = new Vector2(circleA.dot.speedX, circleA.dot.speedY);
		// if (speed.length < 2) circleA.prevRotation = circleA.rotation;
		// final speed = new Vector2(circleB.dot.speedX, circleB.dot.speedY);
		// if (speed.length < 2) circleB.prevRotation = circleB.rotation;
		return true;
	}

	function circleUpdate(
		circle:Circle, normal:Vector2, depth:Float,
		impulse:Float, frictionCoef:Float, ratio:Float
	):Void {
		final vel = new Vector2(circle.dot.speedX, circle.dot.speedY);
		// Calculate velocity along tangential axis
		final vnDot = vel.dot(normal);
		final tangentVel = vel.sub(normal.mult(vnDot));

		// Signed length of tangential velocity
		final tangent = new Vector2(-normal.y, normal.x);
		final velSign = tangentVel.dot(tangent).sign();
		// final velSign = tangent.y >= 0 ? 1 : -1;
		final tangentVel = tangentVel.length * velSign;

		// Rotational velocity transformed from angle to world space
		final circumference = 2 * Math.PI * circle.radius;

		final prevRotationVelocity = (circle.rotation - circle.prevRotation) * world.viscosity;
		// final prevRotationVelocity = circle.prevRotationVelocity;

		var rotVel = prevRotationVelocity / (2 * Math.PI) * circumference;

		// Difference between rotational and tangential velocity
		final velDiff = (rotVel - tangentVel) * frictionCoef;

		// Increase velocity along tangential axis and decrease rotational velocity
		vel.x += tangent.x * velDiff;
		vel.y += tangent.y * velDiff;
		rotVel -= velDiff;

		// Update velocity vector with bounce impulse
		vel.x -= normal.x * impulse * ratio;
		vel.y -= normal.y * impulse * ratio;

		// Game.debugLine = "" + vel.x.toFixed() + "   " + vel.y.toFixed();
		// Game.debugLine2 = "" + velDiff.toFixed();
		// Game.debugLine3 = "" + rotVel.toFixed();
		// Update rotational velocity
		circle.prevRotation = circle.rotation - ((rotVel / circumference) * 2 * Math.PI);

		// Correct circle position
		final dot = circle.dot;
		dot.pos.x += normal.x * depth * ratio;
		dot.pos.y += normal.y * depth * ratio;

		final speed = new Vector2(circle.dot.speedX, circle.dot.speedY);
		if (speed.length < 2) circle.prevRotation = circle.rotation;

		if (world.resolveDebugger) {
			world.resolveDebugger = false;
			#if js js.Lib.debug(); #end
		}

		// Update velocity only if
		if (impulse < 0) {
			// circle.dot.speedX = vel.x;
			// circle.dot.speedY = vel.y;
		}
	}
}
