package khm.verlet;

import khm.math.Vector2;

/**
	The resulting triangle of a Delaunay triangulation operation.
	@see `Delaunay.triangulate`
**/
class DelaunayTriangle {
	/** First vertex of the triangle. **/
	public final p1:Vector2;
	/** Second vertex of the triangle. **/
	public final p2:Vector2;
	/** Third vertex of the triangle. **/
	public final p3:Vector2;

	/** Create a new Delaunay result triangle. **/
	public function new(p1, p2, p3) {
		this.p1 = p1;
		this.p2 = p2;
		this.p3 = p3;
	}
}

private class DelaunayEdge {
	public final p1:Vector2;
	public final p2:Vector2;

	public function new(p1, p2) {
		this.p1 = p1;
		this.p2 = p2;
	}

	public inline function equals(e:DelaunayEdge) {
		return (p1 == e.p1 && p2 == e.p2) || (p1 == e.p2 && p2 == e.p1);
	}
}

/**
	From `heaps/h2d/col/Delaunay.hx` by `ncannasse`.

	A Delaunay triangulation utility. See `Delaunay.triangulate`.
**/
class Delaunay {
	/**
		Performs a Delaunay triangulation on a given set of Points and returns a list of calculated triangles.
		See here for more information: https://en.wikipedia.org/wiki/Delaunay_triangulation
	**/
	public static function triangulate(points:Array<Vector2>):Array<DelaunayTriangle> {
		final nv = points.length;
		if (nv < 3) return null;

		final trimax = 4 * nv;

		// Find the maximum and minimum vertex bounds.
		// This is to allow calculation of the bounding supertriangle
		var xmin = points[0].x;
		var ymin = points[0].y;
		var xmax = xmin;
		var ymax = ymin;

		for (pt in points) {
			if (pt.x < xmin) xmin = pt.x;
			if (pt.x > xmax) xmax = pt.x;
			if (pt.y < ymin) ymin = pt.y;
			if (pt.y > ymax) ymax = pt.y;
		}

		final dx = xmax - xmin;
		final dy = ymax - ymin;
		final dmax = (dx > dy) ? dx : dy;

		final xmid = (xmax + xmin) * 0.5;
		final ymid = (ymax + ymin) * 0.5;

		// Set up the supertriangle
		// This is a triangle which encompasses all the sample points.
		// The supertriangle coordinates are added to the end of the
		// vertex list. The supertriangle is the first triangle in
		// the triangle list.

		final p0 = new Vector2(xmid - 2 * dmax, ymid - dmax);
		final p1 = new Vector2(xmid, ymid + 2 * dmax);
		final p2 = new Vector2(xmid + 2 * dmax, ymid - dmax);
		points.push(p0);
		points.push(p1);
		points.push(p2);

		final triangles = [];
		triangles.push(new DelaunayTriangle(points[nv], points[nv + 1], points[nv + 2])); // SuperTriangle placed at index 0

		// Include each point one at a time into the existing mesh
		for (i in 0...nv) {
			final delaunayEdges = [];
			// delaunayEdges.resize(0);

			// Set up the DelaunayEdge buffer.
			// If the point (vertex(i).x,vertex(i).y) lies inside the circumcircle then the
			// three delaunayEdges of that triangle are added to the DelaunayEdge buffer and the triangle is removed from list.
			var j = -1;
			while (++j < triangles.length) {
				if (inCircle(points[i], triangles[j].p1, triangles[j].p2, triangles[j].p3)) {
					delaunayEdges.push(new DelaunayEdge(triangles[j].p1, triangles[j].p2));
					delaunayEdges.push(new DelaunayEdge(triangles[j].p2, triangles[j].p3));
					delaunayEdges.push(new DelaunayEdge(triangles[j].p3, triangles[j].p1));
					triangles.splice(j, 1);
					j--;
				}
			}

			// In case we the last duplicate point we removed was the last in the array
			if (i >= nv) continue;

			// Remove duplicate delaunayEdges
			// Note: if all triangles are specified anticlockwise then all
			// interior delaunayEdges are opposite pointing in direction.
			var j = delaunayEdges.length - 2;
			while (j >= 0) {
				var k = delaunayEdges.length;
				while (--k >= j + 1) {
					if (delaunayEdges[j].equals(delaunayEdges[k])) {
						delaunayEdges.splice(k, 1);
						delaunayEdges.splice(j, 1);
						k--;
					}
				}
				j--;
			}

			// Form new triangles for the current point
			// Skipping over any tagged delaunayEdges.
			// All delaunayEdges are arranged in clockwise order.
			j = -1;
			while (++j < delaunayEdges.length) {
				if (triangles.length >= trimax) return null;
				triangles.push(new DelaunayTriangle(delaunayEdges[j].p1, delaunayEdges[j].p2, points[i]));
			}
		}

		// Remove triangles with supertriangle vertices
		// These are triangles which have a vertex number greater than nv

		inline function isOut(p) {
			return p == p0 || p == p1 || p == p2;
		}
		// var i = triangles.length;
		// while (i-- > 0) {
		// 	if (isOut(triangles[i].p1) || isOut(triangles[i].p2) || isOut(triangles[i].p3)) {
		// 		triangles.splice(i, 1);
		// 	}
		// }
		var i = triangles.length - 1;
		while (i >= 0) {
			if (isOut(triangles[i].p1) || isOut(triangles[i].p2) || isOut(triangles[i].p3)) {
				triangles.splice(i, 1);
			}
			i--;
		}

		// Remove SuperTriangle vertices
		points.pop();
		points.pop();
		points.pop();
		return triangles;
	}

	static inline var EPSILON = 1e-10;

	static inline function fabs(a:Float):Float {
		return a < 0 ? -a : a;
	}

	static inline function inCircle(p:Vector2, p1:Vector2, p2:Vector2, p3:Vector2):Bool {
		if (fabs(p1.y - p2.y) < EPSILON && fabs(p2.y - p3.y) < EPSILON) return false;
		else {
			var m1 = 0.0;
			var m2 = 0.0;

			var mx1 = 0.0;
			var mx2 = 0.0;

			var my1 = 0.0;
			var my2 = 0.0;

			var xc = 0.0;
			var yc = 0.0;

			if (fabs(p2.y - p1.y) < EPSILON) {
				m2 = -(p3.x - p2.x) / (p3.y - p2.y);
				mx2 = (p2.x + p3.x) * 0.5;
				my2 = (p2.y + p3.y) * 0.5;
				// Calculate CircumCircle center (xc,yc)
				xc = (p2.x + p1.x) * 0.5;
				yc = m2 * (xc - mx2) + my2;
			} else if (fabs(p3.y - p2.y) < EPSILON) {
				m1 = -(p2.x - p1.x) / (p2.y - p1.y);
				mx1 = (p1.x + p2.x) * 0.5;
				my1 = (p1.y + p2.y) * 0.5;
				// Calculate CircumCircle center (xc,yc)
				xc = (p3.x + p2.x) * 0.5;
				yc = m1 * (xc - mx1) + my1;
			} else {
				m1 = -(p2.x - p1.x) / (p2.y - p1.y);
				m2 = -(p3.x - p2.x) / (p3.y - p2.y);
				mx1 = (p1.x + p2.x) * 0.5;
				mx2 = (p2.x + p3.x) * 0.5;
				my1 = (p1.y + p2.y) * 0.5;
				my2 = (p2.y + p3.y) * 0.5;
				// Calculate CircumCircle center (xc,yc)
				xc = (m1 * mx1 - m2 * mx2 + my2 - my1) / (m1 - m2);
				yc = m1 * (xc - mx1) + my1;
			}
			var dx = p2.x - xc;
			var dy = p2.y - yc;
			final rsqr = dx * dx + dy * dy;
			dx = p.x - xc;
			dy = p.y - yc;
			final drsqr = dx * dx + dy * dy;
			return drsqr <= rsqr;
		}
	}
}
