package khm.verlet;

import khm.math.Vector2;

class Dot {
	public final body:Body;
	public final pos:Vector2;
	public final prevPos:Vector2;
	public var fixed:Bool;
	public var viscosity:Float;

	/** Value of `pos.x - prevPos.x` **/
	public var speedX(get, set):Float;
	/** Value of `pos.y - prevPos.y` **/
	public var speedY(get, set):Float;

	/** By default links to `world.gravity` vector object. Can be updated with `setGravity`. **/
	public var gravity(default, null):Vector2;

	public function new(body:Body, x:Float, y:Float, fixed = false) {
		this.body = body;
		pos = new Vector2(x, y);
		prevPos = new Vector2(x, y);
		this.fixed = fixed;
		viscosity = body.world.viscosity;
		gravity = body.world.gravity;
	}

	public function integrate():Void {
		if (fixed) return;
		final world = body.world;
		final pos = this.pos;
		final prevPos = this.prevPos;
		final oldX = pos.x;
		final oldY = pos.y;
		final vis = viscosity;
		// final speed = new Vector2(speedX, speedY);
		pos.x += (pos.x - prevPos.x) * vis + gravity.x;
		pos.y += (pos.y - prevPos.y) * vis + gravity.y;
		prevPos.x = oldX;
		prevPos.y = oldY;

		// circle is just one point
		if (body.type == Circle) {
			final circle:Circle = cast body;
			final rotation = circle.rotation;
			final friction = world.friction / world.iterations;
			final velocity = (rotation - circle.prevRotation) * vis;
			circle.prevRotation = rotation;
			circle.rotation += velocity;
			// trace(speed.length);
			// if (speed.length < 2) circle.prevRotation = circle.rotation;
			// circle.prevRotationVelocity = velocity;
		}

		if (world.bounds == null) return;
		final minX = world.bounds.x;
		final minY = world.bounds.y;
		final maxX = world.bounds.x + world.bounds.w;
		final maxY = world.bounds.y + world.bounds.h;

		if (pos.x < minX) pos.x = minX;
		else if (pos.x > maxX) pos.x = maxX;

		if (pos.y < minY) pos.y = minY;
		else if (pos.y > maxY) {
			final frictionBounds = world.frictionBounds / world.iterations;
			pos.x -= (pos.y - maxY) * (pos.x - prevPos.x) * frictionBounds;
			pos.y = maxY;
		}
	}

	public function movePoint(v:Vector2):Void {
		pos.setFrom(v);
		if (fixed) {
			prevPos.setFrom(pos);
			for (constraint in body.constraints) {
				if (constraint.dotA != this && constraint.dotB != this) continue;
				constraint.updateInitialDist();
			}
		}
		body.updateBoundingBox();
	}

	public inline function setGravity(v:Vector2):Void {
		gravity = new Vector2(v.x, v.y);
	}

	/** Resets gravity to `world.gravity` reference. **/
	public function resetGravity():Void {
		gravity = body.world.gravity;
	}

	public function hasJoints():Bool {
		for (c in body.joints) {
			if (c.dotA == this || c.dotB == this) return true;
		}
		return false;
	}

	public function constraintsCount():Int {
		var n = 0;
		for (c in body.constraints) {
			if (c.dotA == this || c.dotB == this) n++;
		}
		return n;
	}

	public function getConstraints():Array<Constraint> {
		final arr = [];
		for (c in body.constraints) {
			if (c.dotA == this || c.dotB == this) arr.push(c);
		}
		return arr;
	}

	inline function get_speedX():Float {
		return pos.x - prevPos.x;
	}

	inline function set_speedX(value:Float):Float {
		prevPos.x = pos.x - value;
		return value;
	}

	inline function get_speedY():Float {
		return pos.y - prevPos.y;
	}

	inline function set_speedY(value:Float):Float {
		prevPos.y = pos.y - value;
		return value;
	}
}
