package khm.verlet;

import game.Game;
import haxe.Rest;
import khm.Types.Rect;
import khm.math.Degrees;
import khm.math.Vector2;
import khm.verlet.Body.BodyOptions;
import khm.verlet.Triangle.TriangleOptions;

@:structInit
class WorldOptions {
	public var iterations = 5;
	public var gravity:Vector2 = {x: 0, y: 0.1};
	public var friction = 0.2;
	public var frictionBounds = 0.2;
	/** Air viscosity in free space **/
	public var viscosity = 1.0;
	public var bounds:Null<Rect> = null;
}

/**
 * Verlet-integration physic engine
 * Based on Gerard Ferrandez and Ondrej Mocny sources
 */
class World {
	public static inline var MAX_FLOAT = 999999.0;

	public final constraints:Array<Constraint> = [];
	public final dots:Array<Dot> = [];
	public final bodies:Array<Body> = [];
	public final dynBodies:Array<Body> = [];
	public final collision:Collision;
	public var iterations:Int;
	public var gravity:Vector2;
	public var friction:Float;
	public var frictionBounds:Float;
	@:inheritDoc(WorldOptions.viscosity)
	public var viscosity:Float;
	public var bounds:Null<Rect> = null;
	public var paused = false;
	public var pauseNextTick = false;
	public var resolveDebugger = false;
	public var activeIteration = 0;
	public var renderEdgesOnly = true;
	public var renderCircles = true;
	public var renderDotSpeed = true;
	public final afterCollisionCallbacks:Array<() -> Void> = [];

	public function new(?opts:WorldOptions):Void {
		if (opts == null) opts = {};
		gravity = opts.gravity;
		iterations = opts.iterations;
		friction = opts.friction;
		frictionBounds = opts.frictionBounds;
		viscosity = opts.viscosity;
		bounds = opts.bounds;
		collision = new Collision(this);
	}

	public function step(dt:Float):Void {
		if (paused) return;
		integrate();
		solve();
		if (pauseNextTick) {
			pauseNextTick = false;
			paused = true;
		}
	}

	public function integrate():Void {
		for (dot in dots) dot.integrate();
	}

	public function solve():Void {
		// reset info for debug render
		collision.lastEdges.resize(0);
		collision.lastClosestDots.resize(0);
		collision.lastPoints.resize(0);
		collision.lastAxis.resize(0);
		collision.satCount = 0;
		var iter = 0;
		var col = 0;
		final bodiesLength = bodies.length;
		for (n in 0...iterations) {
			activeIteration = n;
			// keep constraint distances
			for (constraint in constraints) constraint.solve();

			// Recalculate bounding boxes (for broad phase in sat)
			for (body in bodies) {
				if (body.fixed) continue;
				body.updateBoundingBox();
			}

			// collision detection (narrow phase)

			for (i in 0...bodiesLength - 1) {
				final body = bodies[i];
				for (j in i + 1...bodiesLength) {
					final body2 = bodies[j];
					// var i = bodiesLength - 1;
					// while (i-- > 0) {
					// 	final body = bodies[i];
					// 	var j = bodiesLength;
					// 	while (j-- > i + 1) {
					final body2 = bodies[j];
					iter++;
					// if (body == null || body2 == null) continue;
					if (body.fixed && body2.fixed) continue;
					if (!isLayersMatch(body, body2)) continue;
					switch [body.type, body2.type] {
						case [Polygon, Polygon]:
							final response = collision.sat(body, body2);
							if (response != null) collision.resolve(response);
						case [Polygon, Circle]:
							final isCol = collision.solveCircleOnPolygon(cast body2, body);
						case [Circle, Polygon]:
							final isCol = collision.solveCircleOnPolygon(cast body, body2);
						case [Circle, Circle]:
							final isCol = collision.solveCircleOnCircle(cast body, cast body2);
					}
					col++;
				}
			}

			for (body in bodies) {
				if (body.onMove == null) continue;
				final angle:Degrees = if (body.dots.length > 1) {
					MathLib.angle.betweenPoints(body.dots[0].pos, body.dots[1].pos);
				} else if (body.type == Circle) {
					(cast body : Circle).rotation.toDeg();
				} else 0.0;
				body.onMove(body.center.x, body.center.y, angle);
			}
		}

		for (cb in afterCollisionCallbacks) cb();
		afterCollisionCallbacks.resize(0);
		Game.debugLine3 = "" + iter + " | " + col + " | " + collision.satCount;
	}

	function isLayersMatch(a:Body, b:Body):Bool {
		// search for any of layers in ignore list for both bodies
		if (b.ignoreLayers.toInt() & a.layers.toInt() != 0
			|| a.ignoreLayers.toInt() & b.layers.toInt() != 0) {
			return false;
		}
		return true;
	}

	#if kha
	public function debugRender(g:kha.graphics2.Graphics):Void {
		for (body in bodies) body.renderDots(g);
		for (body in bodies) body.renderConstraints(g);

		for (c in constraints) {
			if (c.body != null) continue;
			// final prevP = edges[i - 1] ?? edges.last();
			final p = c.dotA.pos;
			final p2 = c.dotB.pos;
			final rest = c.dotA.pos.x == c.dotA.prevPos.x && c.dotA.pos.y == c.dotA.prevPos.y;
			g.color = rest ? 0xFFBAD8BD : 0xFFB6FF88;
			g.drawLine(p.x, p.y, p2.x, p2.y, 2);
		}
		for (edge in collision.lastEdges) {
			g.color = 0xFFFF00D0;
			g.drawLine(
				edge.dotA.pos.x, edge.dotA.pos.y,
				edge.dotB.pos.x, edge.dotB.pos.y,
				2
			);
		}
		// for (obj in collision.lastAxis) {
		// 	final edge = obj.edge;
		// 	g.color = 0xFFFFDD00;
		// 	g.drawLine(
		// 		edge.dotA.pos.x, edge.dotA.pos.y,
		// 		edge.dotB.pos.x, edge.dotB.pos.y,
		// 		2
		// 	);
		// }
		for (obj in collision.lastAxis) {
			g.color = 0xFFFF0000;
			final dot = obj.dot;
			g.drawLine(
				dot.pos.x, dot.pos.y,
				dot.pos.x + obj.axis.x * obj.depth * 4,
				dot.pos.y + obj.axis.y * obj.depth * 4,
				2
			);
			// g.drawLine(
			// 	edge.dotA.pos.x, edge.dotA.pos.y,
			// 	edge.dotB.pos.x + obj.axis.x * obj.depth * 4,
			// 	edge.dotB.pos.y + obj.axis.y * obj.depth * 4,
			// 	2
			// );
		}
		for (dot in collision.lastClosestDots) {
			g.color = 0xFF00E1FF;
			g.fillRect(
				dot.pos.x - 2, dot.pos.y - 2,
				4, 4
			);
		}
		for (pos in collision.lastPoints) {
			g.color = 0xFF00FF6E;
			g.fillRect(
				pos.x - 2, pos.y - 2,
				4, 4
			);
		}
	}
	#end

	public function addBody(opts:BodyOptions, ?pos:DebugPosInfos):Body {
		opts.posInfos = pos;
		final body = new Body(this, opts);
		pushBody(body);
		return body;
	}

	public function addRect(x:Float, y:Float, w:Float, h:Float, ?set:{?mass:Float, ?fixed:Bool}, ?pos:DebugPosInfos):Rectangle {
		set ??= {};
		set.fixed ??= false;
		final opts:BodyOptions = {
			mass: set.mass,
			x: x,
			y: y,
			dots: [
				{x: 0, y: 0, fixed: set.fixed},
				{x: w, y: 0, fixed: set.fixed},
				{x: w, y: h, fixed: set.fixed},
				{x: 0, y: h, fixed: set.fixed}
			],
			autoEdges: true,
			constraints: [],
			posInfos: pos,
		};
		if (!set.fixed) {
			opts.constraints.push({indexA: 0, indexB: 2}); // [\]
			opts.constraints.push({indexA: 3, indexB: 1}); // [/]
		}
		final body = new Rectangle(this, opts);
		pushBody(body);
		return body;
	}

	public function addCircle(x:Float, y:Float, radius:Float, ?set:{?mass:Float, ?fixed:Bool}, ?pos:DebugPosInfos):Circle {
		set ??= {};
		set.fixed ??= false;
		final body = new Circle(this, radius, {
			mass: set.mass,
			x: x,
			y: y,
			dots: [{x: 0, y: 0, fixed: set.fixed}],
			constraints: [],
			posInfos: pos,
		});
		pushBody(body);
		return body;
	}

	public function createBlob(x:Float, y:Float, r:Float, mass:Float):Blob {
		final body = new Blob(this, x, y, r, mass);
		pushBody(body);
		return body;
	}

	public function createTriangle(x:Float, y:Float, x2:Float, y2:Float, x3:Float, y3:Float, ?config:TriangleOptions):Body {
		final body = new Triangle(this, x, y, x2, y2, x3, y3, config);
		pushBody(body);
		return body;
	}

	function pushBody(body:Body):Void {
		bodies.push(body);
		if (!body.fixed) dynBodies.push(body);
	}

	public function addJoint(fromDot:Dot, toDot:Dot, stiffness = 1.0):Constraint {
		final constraint = new Constraint(
			null,
			fromDot,
			toDot,
			stiffness
		);
		fromDot.body.joints.push(constraint);
		toDot.body.joints.push(constraint);
		constraints.push(constraint);
		return constraint;
	}

	public function clear():Void {
		bodies.resize(0);
		dots.resize(0);
		constraints.resize(0);
	}

	public function trace(a:Any, ?b:Any, ?c:Any, ?pos:haxe.PosInfos) {
		if (activeIteration != 0) return;
		if (b != null && c != null) trace(a, b, c, pos);
		else if (b != null) trace(a, b, pos);
		else trace(a, pos);
	}
}
