package khm.verlet.editor;

import khm.geom.HertelMehlhorn;
import khm.geom.poly2tri.Point as PolyPoint;
import khm.geom.poly2tri.VisiblePolygon;
import khm.math.Vector2;
import khm.utils.DebugRectOverlay;
import khm.verlet.Body;
import khm.verlet.Delaunay;
import khm.verlet.Dot;
import khm.verlet.DotUtils;
import khm.verlet.World;

using khm.tools.PhaserTools;

@:expose
class VerletEditor extends Scene {
	public static var instance:Null<VerletEditor>;
	public static var fontSize = 20;

	public var scene:Scene;
	public var world:World;
	public var closestDot:Dot;
	public var dragMode = false;
	public final selectedDots:Array<Dot> = [];

	var activeToolId:VerletEditorTool = SelectionTool;
	final tools = [];

	public var activeTool(get, never):Scene;

	function get_activeTool():Scene {
		return tools[activeToolId.toInt()];
	}

	public function new(scene:Scene, ?verlet:World) {
		super();
		instance = this;
		this.scene = scene;
		this.world = verlet;
		DebugRectOverlay.ignoreObject(this);
		final g = new Graphics();
		g.font = "Courier";
		add(g);

		on(ScenesChange, e -> {
			if (e.scene == this) return;
			if (App.scenes.last() == this) return;
			if (isAttached) bringToTop();
		});

		tools = [
			new Selection(this),
			new Line(this),
			new RectObj(this)
		];
		for (tool in tools) tool.visible = false;
		activeTool.visible = true;
		activeTool.onShow();

		on(BeforeRender, e -> {
			g.begin();
			// g.scale = scene.camera.scale;
			// g.x = -scene.camera.x;
			// g.y = -scene.camera.y;
			camera.originX = scene.camera.originX;
			camera.originY = scene.camera.originY;
			camera.x = scene.camera.x - scene.x;
			camera.y = scene.camera.y - scene.y;
			camera.scale = scene.camera.scale;
			g.drawString(
				'Tool: ${activeTool?.name ?? ''}',
				camera.x - scene.width / 2 / camera.scale,
				camera.y - scene.height / 2 / camera.scale
			);
			g.color = 0xCC00B7FF;
			for (dot in selectedDots) {
				g.drawCircle(dot.pos.x, dot.pos.y, 6, 4);
			}
			if (closestDot != null) {
				g.color = 0xAAFF0000;
				// todo draw in local scene world cords
				// pushContainerSpace(scene.transformation);
				// scene.updateTransformation();
				// g.pushTransformation(scene.transformation);

				g.drawCircle(closestDot.pos.x, closestDot.pos.y, 8, 4);

				// g.popTransformation();
				// popContainerSpace()
			}

			// final p = App.pointers[0]?.toLocal(scene);
			// if (p != null && !p.isDown) {
			// 	// updateClosestDot(p);
			// 	for (body in world.bodies) {
			// 		final isInside = body.isPointInsideAabb(p.x, p.y);
			// 		g.drawRect(
			// 			body.aabbCenter.x - body.halfSize.x,
			// 			body.aabbCenter.y - body.halfSize.y,
			// 			body.halfSize.x * 2, body.halfSize.y * 2
			// 		);
			// 		if (isInside) g.drawCircle(p.x, p.y, 8, 8);
			// 	}
			// }

			g.end();
		});
	}

	// #if kha
	// override function render(g:kha.graphics2.Graphics) {
	// 	if (!visible || alpha == 0) return;
	// 	g.color = 0xFFFFFFFF;
	// 	pushContainerSpace(g);
	// 	for (obj in children)
	// 		obj.render(g);
	// 	popContainerSpace(g);
	// }
	// #end

	public function setWorld(world:World):Void {
		this.world = world;

		final keep = Lazyr.keepDebugValue;
		world.renderEdgesOnly = keep("renderEdgesOnly", () -> world.renderEdgesOnly);
		world.renderCircles = keep("renderCircles", () -> world.renderCircles);
		world.renderDotSpeed = keep("renderDotSpeed", () -> world.renderDotSpeed);
		on(KeyDown, e -> {
			e.propagate = false;
			final isCmd = App.keys.get(Control) || App.keys.get(Meta);
			final isAlt = App.keys.get(Alt);
			switch e.key {
				case One: setTool(SelectionTool);
				case Two: setTool(LineTool);
				case Three: setTool(RectObjTool);
				case Backspace:
					removeSelectedDots();
					setTool(SelectionTool);
				case S if (isCmd): onSave();
				case Q, P: world.paused = !world.paused;
				case CloseBracket:
					world.pauseNextTick = true;
					world.paused = false;
				case E: world.renderEdgesOnly = !world.renderEdgesOnly;
				case C if (isAlt): world.renderCircles = !world.renderCircles;
				case S: world.renderDotSpeed = !world.renderDotSpeed;
				case D: world.resolveDebugger = true;
				case I if (!isCmd):
					final dot = selectedDots[0] ?? closestDot;
					if (dot != null) {
						trace(dot);
						trace(dot.getConstraints());
					}
				case T:
					final dot = selectedDots[0] ?? closestDot;
					if (dot != null) triangulateClosedBody(dot.body);
				case _:
					e.propagate = true;
			}
			if (!e.propagate) App.preventDefault();
		});
	}

	public function setTool(id:VerletEditorTool):Void {
		if (activeToolId == id) return;
		final prevTool = activeTool;
		prevTool.visible = false;
		prevTool.onHide();

		activeToolId = id;
		activeTool.visible = true;
		activeTool.onShow();
		trace(Type.getClassName(Type.getClass(activeTool)).split(".").last());
	}

	public function updateClosestDot(p:Pointer, minDistance = 40.0, ?ignoreDots:Array<Dot>):Null<Dot> {
		if (world == null) return null;
		closestDot = null;
		// var minDistance = 99999.0;

		final off = minDistance * 2;
		for (body in world.bodies) {
			if (!body.isRectInsideAabb(p.x - off / 2, p.y - off / 2, off, off)) continue;
			for (dot in body.dots) {
				if (ignoreDots != null) {
					if (ignoreDots.contains(dot)) continue;
				}
				final dist = dot.pos.sub({x: p.x, y: p.y}).length;
				// select closest dot
				if (dist < minDistance) {
					closestDot = dot;
					minDistance = dist;
				}
			}
		}
		return closestDot;
	}

	public function getClosestDots(mainDot:Dot, minDistance = 3.0):Array<Dot> {
		final dots = [];
		final p = mainDot.pos;
		final off = minDistance * 2;
		for (body in world.bodies) {
			if (!body.isRectInsideAabb(p.x - off / 2, p.y - off / 2, off, off)) continue;
			for (dot in body.dots) {
				if (dot == mainDot) continue;
				final dist = dot.pos.sub({x: p.x, y: p.y}).length;
				// select closest dot
				if (dist < minDistance) dots.push(dot);
			}
		}
		return dots;
	}

	public function toggleDotSelection(dot:Dot):Bool {
		if (selectedDots.contains(dot)) {
			selectedDots.remove(dot);
		} else {
			selectedDots.push(dot);
		}
		return selectedDots.contains(dot);
	}

	public function removeSelectedDots():Void {
		for (dot in selectedDots) {
			dot.body.removeDot(dot);
			if (dot.body.dots.length == 0) dot.body.destroy();
		}
		selectedDots.resize(0);
	}

	public function triangulateClosedBody(body:Body):Void {
		final isConcave = DotUtils.isConcave(body.dots);
		final isClosed = isBodyDotsClosed(body);
		// if (!isConcave || !isClosed) return;
		final points = getBodyEdgesPos(body);
		trace(points);
		final triangles = Delaunay.triangulate(points);
		final vp = new VisiblePolygon();
		final ppoints = points.map(p -> {
			new PolyPoint(p.x, p.y);
		});
		vp.addPolyline(ppoints);
		vp.performTriangulationOnce();

		final triangles = @:privateAccess vp.sweepContext.triangles;
		final triangles = triangles.map(t -> {
			t.points.map(p -> {x: p.x, y: p.y});
		});
		final polys = HertelMehlhorn.polygonize(triangles);

		final dotOptions = polys.map(poly -> {
			poly.map(p -> {
				final d:DotOptions = {x: p.x, y: p.y, fixed: true};
				d;
			});
		});
		body.destroy();
		final newBodies = [];
		for (dots in dotOptions) {
			final body = world.addBody({
				dots: dots,
				// dots: [
				// 	// {x: t.p3.x, y: t.p3.y, fixed: true},
				// 	// {x: t.p2.x, y: t.p2.y, fixed: true},
				// 	// {x: t.p1.x, y: t.p1.y, fixed: true},
				// 	{x: t.p1.x, y: t.p1.y, fixed: true},
				// 	{x: t.p2.x, y: t.p2.y, fixed: true},
				// 	{x: t.p3.x, y: t.p3.y, fixed: true},
				// ],
				constraints: [],
				autoEdges: true,
			});
			newBodies.push(body);
		}

		// return;

		// final triangles = triangles.map(points -> {
		// 	// final points = triangle.points;
		// 	return {
		// 		p1: {x: points[0].x, y: points[0].y},
		// 		p2: {x: points[1].x, y: points[1].y},
		// 		p3: {x: points[2].x, y: points[2].y},
		// 	};
		// });
		// body.destroy();
		// final newBodies = [];
		// for (t in triangles) {
		// 	final body = world.addBody({
		// 		dots: [
		// 			// {x: t.p3.x, y: t.p3.y, fixed: true},
		// 			// {x: t.p2.x, y: t.p2.y, fixed: true},
		// 			// {x: t.p1.x, y: t.p1.y, fixed: true},
		// 			{x: t.p1.x, y: t.p1.y, fixed: true},
		// 			{x: t.p2.x, y: t.p2.y, fixed: true},
		// 			{x: t.p3.x, y: t.p3.y, fixed: true},
		// 		],
		// 		constraints: [],
		// 		autoEdges: true,
		// 	});
		// 	newBodies.push(body);
		// }

		return;

		final dups:Map<String, Array<Dot>> = [];
		for (body in newBodies) {
			for (dot in body.dots) {
				final id = '${dot.pos.x},${dot.pos.y}';
				dups[id] ??= [];
				dups[id].push(dot);
			}
		}
		untyped console.log(dups);
		for (arr in dups) {
			for (last in arr.reversed()) {
				final first = arr[0];
				if (last == first) continue;
				first.body.mergeDot(last, first);
				// last.body.removeDot(last);
			}
		}
	}

	function isBodyDotsClosed(body:Body):Bool {
		for (c in body.edges) {
			if (c.dotA.constraintsCount() < 2) return false;
		}
		return true;
	}

	function getBodyEdgesPos(body:Body):Array<Vector2> {
		final points:Array<Vector2> = [];
		for (c in body.edges) {
			points.push(c.dotA.pos);
		}
		return points;
	}

	dynamic public function onSave():Void {}
}
