package;

import haxe.Json;
import haxe.crypto.Md5;
import haxe.io.Path;
import js.Node.console;
import js.Node;
import js.lib.Promise;
import js.node.Buffer;
import js.node.ChildProcess;
import js.node.Fs.Fs;
import sys.FileSystem;
import sys.io.File;

using Lambda;
using StringTools;

enum abstract PackerType(String) {
	var MAX_RECTS_BIN = 'MaxRectsBin';
	var MAX_RECTS_PACKER = 'MaxRectsPacker';
	var OPTIMAL_PACKER = 'OptimalPacker';
}

enum abstract ScaleMethod(String) {
	var BILINEAR = 'BILINEAR';
	var NEAREST_NEIGHBOR = 'NEAREST_NEIGHBOR';
	var HERMITE = 'HERMITE';
	var BEZIER = 'BEZIER';
}

enum abstract PackerMethod(String) {
	var BEST_SHORT_SIDE_FIT = "BestShortSideFit";
	var BEST_LONG_SIDE_FIT = "BestLongSideFit";
	var BEST_AREA_FIT = "BestAreaFit";
	var BOTTOM_LEFT_RULE = "BottomLeftRule";
	var CONTACT_POINT_RULE = "ContactPointRule";
	var SMART = "Smart";
	var SQUARE = "Square";
	var SMART_SQUARE = "SmartSquare";
	var SMART_AREA = "SmartArea";
	var SQUARE_AREA = "SquareArea";
	var SMART_SQUARE_AREA = "SmartSquareArea";
}

enum abstract PackerExporterType(String) {
	var JSON_HASH = 'JsonHash';
	var JSON_ARRAY = 'JsonArray';
	var CSS = 'Css';
	var OLD_CSS = 'OldCss';
	var PIXI = 'Pixi';
	var PHASER_HASH = 'PhaserHash';
	var PHASER_ARRAY = 'PhaserArray';
	var PHASER3 = 'Phaser3';
	var XML = 'XML';
	var STARLING = 'Starling';
	var COCOS2D = 'Cocos2d';
	var SPINE = 'Spine';
	var UNREAL = 'Unreal';
	var UIKIT = 'UIKit';
	var UNITY3D = 'Unity3D';
}

enum abstract BitmapFilterType(String) {
	var GRAYSCALE = 'grayscale';
	var MASK = 'mask';
	var NONE = 'none';
}

typedef TexturePackerOptions = {
	/**
	 * Name of output files.
	 *
	 * @default pack-result
	 */
	?textureName:String,

	/**
	 * Max single texture width in pixels
	 *
	 * @default 2048
	 */
	?width:Int,
	/**
	 * Max single texture height in pixels
	 *
	 * @default 2048
	 */
	?height:Int,
	/**
	 * Fixed texture size
	 *
	 * @default false
	 */
	?fixedSize:Bool,
	/**
	 * Force power of two textures sizes
	 *
	 * @default false
	 */
	?powerOfTwo:Bool,
	/**
	 * Spaces in pixels around images
	 *
	 * @default 0
	 */
	?padding:Int,
	/**
	 * Extrude border pixels size around images
	 *
	 * @default 0
	 */
	?extrude:Int,
	/**
	 * Allow image rotation
	 * @default true
	 */
	?allowRotation:Bool,
	/**
	 * Allow detect identical images
	 *
	 * @default true
	 */
	?detectIdentical:Bool,
	/**
	 * Allow trim images
	 *
	 * @default true
	 */
	?allowTrim:Bool,
	/**
	 * Trim mode (trim/crop)
	 *
	 * @default {@link TrimMode.TRIM}
	 * @see {@link TrimMode}
	 * @see {@link allowTrim}
	 */
	?trimMode:String,
	/**
	 * Threshold alpha value
	 *
	 * @default 0
	 */
	?alphaThreshold:Int,
	/**
	 * Remove file extensions from frame names
	 *
	 * @default false
	 */
	?removeFileExtension:Bool,
	/**
	 * Prepend folder name to frame names
	 *
	 * @default true
	 */
	?prependFolderName:Bool,
	/**
	 * Output file format (png/jpg)
	 *
	 * @default {@link TextureFormat.PNG}
	 * @see {@link TextureFormat}
	 */
	?textureFormat:String,
	/**
	 * Export texture as base64 string to atlas meta tag
	 *
	 * @default false
	 */
	?base64Export:Bool,
	/**
	 * Scale size and positions in atlas
	 *
	 * @default 1
	 */
	?scale:Int,
	/**
	 * Texture scaling method
	 *
	 * @default ScaleMethod.BILINEAR
	 */
	?scaleMethod:ScaleMethod,
	/**
	 * "Tinify" texture using TinyPNG
	 *
	 * @default false
	 */
	?tinify:Bool,
	/**
	 * TinyPNG key
	 *
	 * @default empty string
	 */
	?tinifyKey:String,
	/**
	 * Type of packer
	 * @see PackerType
	 * @default {@link PackerType.MAX_RECTS_BIN}
	 */
	?packer:PackerType,
	/**
	 * Pack method
	 *
	 * @default {@link MaxRectsBinMethod.BEST_SHORT_SIDE_FIT}
	 * @see MaxRectsBinMethod
	 * @see MaxRectsPackerMethod
	 */
	?packerMethod:PackerMethod,
	/**
	 * Name of predefined exporter (), or custom exporter (see below)
	 *
	 * @default JsonHash
	 */
	?exporter:PackerExporterType,
	/**
	 * Bitmap filter type
	 *
	 * @see BitmapFilterType
	 * @default {@link BitmapFilterType.NONE}
	 */
	?filter:BitmapFilterType,
	/**
	 * External application info.
	 * Required fields: url and version
	 *
	 * @default null
	 */
	?appInfo:Dynamic,
}

// @:jsRequire("free-tex-packer-core")
extern class TexturePacker {
	@:selfCall
	function texturePacker(
		images:Array<{path:String, contents:Buffer}>,
		?opts:TexturePackerOptions,
		?onComplete:(
			files:Array<{name:String, buffer:Buffer}>, error:Dynamic
		) -> Void
	):String;
}

class GenAtlas {
	static function main() {
		final projectPath = haxe.macro.Compiler.getDefine("projectPath");
		if (projectPath == null)
			throw '-D projectPath=... not found, use `haxelib run khm build atlases` from project folder';
		// projectPath ??= ".";
		var texturePackerModule:TexturePacker = try {
			final buffer:Buffer = ChildProcess.execSync("npm root --location=global");
			final path = buffer.toString().trim();
			Node.require(path + "/free-tex-packer-core");
		} catch (err) {
			console.log("Install free-tex-packer-core globally first: check out README.md");
			Node.process.exit(1);
			return;
		}

		final args = Sys.args();
		final onlyAtlas = valueOfArg(args, "--only");
		final atlasesPath = '$projectPath/assets/atlases';
		final outFolder = "res/atlases";
		final outBgFolder = "res";
		final outPath = '$projectPath/$outFolder';
		final outBgPath = '$projectPath/$outBgFolder';
		ensureDir(outPath);

		var folders = getAtlasFolders(atlasesPath);
		if (onlyAtlas != null) {
			if (onlyAtlas == "res") folders = [""];
			else folders = folders.filter(s -> s == onlyAtlas);
		}

		final lazyr = readLazyrConfig('$projectPath/lazyr.json');
		final texOptions:TexturePackerOptions = lazyr.texOptions;

		final promises:Array<Promise<Any>> = [];
		for (folder in folders) {
			var assetsFolder = '$atlasesPath/$folder';
			var outFilePath = '$outPath/res_$folder';
			if (folder == "") {
				assetsFolder = '$atlasesPath';
				outFilePath = '$outPath/res';
			}
			if (FileSystem.exists('$outFilePath.json')) {
				if (!checkDirHash(assetsFolder)) continue;
			}
			promises.push(
				makeAtlas(assetsFolder, outFilePath, texturePackerModule, texOptions)
			);
		}
		Promise.all(promises).then(_ -> {
			updateHxml('$projectPath/build-resources.hxml', outPath, outBgPath, atlasesPath);
			console.log("\n");
		});
	}

	static function getAtlasFolders(atlasesPath:String):Array<String> {
		final folders = FileSystem.readDirectory(atlasesPath).filter(name -> {
			if (name.startsWith(".") || name.startsWith("_")) return false;
			if (FileSystem.isDirectory('$atlasesPath/$name')) return true;
			return false;
		});
		folders.push("");
		return folders;
	}

	static function readLazyrConfig(path:String):{
		texOptions:TexturePackerOptions
	} {
		final config = {
			texOptions: {
				tinifyKey: "",
				padding: 1,
				trimMode: "trim",
				powerOfTwo: true,
			}
		}
		if (!FileSystem.exists(path)) {
			console.log('Warning: $path not found, checkout README');
			return config;
		}
		final lazyr:Dynamic = Json.parse(File.getContent(path));
		if (lazyr == null) return config;
		final opts:TexturePackerOptions = lazyr.texpacker;
		if (opts == null) return config;
		if (opts.tinifyKey != null) config.texOptions.tinifyKey = opts.tinifyKey;
		if (opts.padding != null) config.texOptions.padding = opts.padding;
		if (opts.trimMode != null) config.texOptions.trimMode = opts.trimMode;
		if (opts.powerOfTwo != null) config.texOptions.powerOfTwo = opts.powerOfTwo;
		return config;
	}

	static function valueOfArg(args:Array<String>, key:String):Null<String> {
		final i = args.indexOf(key);
		if (i == -1) return null;
		return args[i + 1];
	}

	static function isImageExtension(ext:String):Bool {
		return switch ext {
			case "png", "jpg", "jpeg", "webp": true;
			case _: false;
		}
	}

	static function makeAtlas(
		folderPath:String,
		outFilePath:String,
		texturePackerModule:TexturePacker,
		texOptions:TexturePackerOptions
	):Promise<Any> {
		final texturePacker = texturePackerModule.texturePacker;

		var items = FileSystem.readDirectory(folderPath);
		var trimMode = texOptions.trimMode;
		if (items.contains(".trimMode-crop")) trimMode = "crop";
		if (items.contains(".trimMode-trim")) trimMode = "trim";
		var tinify = texOptions.tinifyKey == "" ? false : true;
		if (items.contains(".tinify-false")) tinify = false;

		items = items.filter(name -> {
			if (FileSystem.isDirectory('$folderPath/$name')) return false;
			if (name.startsWith(".")) return false;
			final ext = Path.extension(name);
			if (!isImageExtension(ext)) return false;
			return true;
		});
		if (items.length == 0) return Promise.resolve();

		final images:Array<{path:String, contents:Buffer}> = [];
		for (name in items) {
			images.push({
				path: '$folderPath/$name',
				contents: Fs.readFileSync('$folderPath/$name')
			});
		}
		final promise = new Promise((resolve, reject) -> {
			texturePacker(images, {
				textureName: Path.withoutDirectory(outFilePath),
				removeFileExtension: true,
				prependFolderName: false,
				tinify: tinify,
				tinifyKey: texOptions.tinifyKey,
				exporter: PackerExporterType.PHASER3,
				// exporter: PackerExporterType.PIXI,
				padding: texOptions.padding,
				allowTrim: true,
				allowRotation: false,
				powerOfTwo: texOptions.powerOfTwo,
				trimMode: trimMode,
				packer: PackerType.OPTIMAL_PACKER,
			}, (files, error) -> {
				if (error != null) {
					removeDirHash(folderPath);
					var msg = 'Error: packaging error for $outFilePath.png';
					if (error.description != null) msg += ':${error.description}';
					console.log(msg);
					resolve(null);
					return;
				}

				final endsWithNum = ~/-[0-9]\./g;
				final outDir = Path.directory(outFilePath);
				var hasSkippedAtlases = false;
				removeFolder('$outDir/_skipped');
				for (item in files) {
					if (endsWithNum.match(item.name)) {
						if (item.name.endsWith(".png")) {
							console.log('Skipped: ${item.name}: too big atlas is splitted');
							ensureDir('$outDir/_skipped');
							Fs.writeFileSync('$outDir/_skipped/${item.name}', item.buffer);
							hasSkippedAtlases = true;
						}
						continue;
					}
					Fs.writeFileSync('$outDir/${item.name}', item.buffer);
					final size = Fs.statSync('$outDir/${item.name}').size;
					final kb = Std.int(size / 1024 * 10) / 10;
					if (item.name.endsWith(".png")) {
						final size = PngDimensions.calculate(item.buffer);
						console.log(item.name, '$kb KiB', '${size.width}x${size.height}');
					} else {
						// console.log(item.name, '$kb KiB');
					}
				}
				if (hasSkippedAtlases) {
					removeDirHash(folderPath);
					console.log("Error: see splitted atlas textures in res/_skipped");
				}
				resolve(null);
			});
		});
		return promise;
	}

	public static function ensureDir(path:String):Void {
		if (!FileSystem.exists(path)) FileSystem.createDirectory(path);
	}

	public static function removeFolder(folderPath:String):Void {
		if (!FileSystem.exists(folderPath)) return;
		for (name in FileSystem.readDirectory(folderPath)) {
			final path = '$folderPath/$name';
			if (FileSystem.isDirectory(path)) continue;
			FileSystem.deleteFile(path);
		}
		FileSystem.deleteDirectory(folderPath);
	}

	/** Returs `true` if folders content changed and updates cache file. **/
	static function checkDirHash(folder:String):Bool {
		if (Sys.args().contains("--force")) return true;
		var hashStats = "";
		final items = FileSystem.readDirectory(folder);
		for (name in items) {
			if (name.startsWith(".")) continue;
			final path = '$folder/$name';
			if (FileSystem.isDirectory(path)) continue;
			final stat = FileSystem.stat(path);
			final fileInfo = '${path}:${stat.size}:${stat.mtime.getTime()}\n';
			hashStats += fileInfo;
		}
		final hash = Md5.encode(hashStats);
		var isFolderChanged = true;
		final cachePath = '$folder/._atlas_hash';
		if (FileSystem.exists(cachePath)) {
			isFolderChanged = File.getContent(cachePath) != hash;
		}
		if (isFolderChanged) File.saveContent(cachePath, hash);
		else console.log('Skip atlas generation for $folder');
		return isFolderChanged;
	}

	static function removeDirHash(folder:String):Void {
		final cachePath = '$folder/._atlas_hash';
		if (!FileSystem.exists(cachePath)) return;
		FileSystem.deleteFile(cachePath);
	}

	static function updateHxml(hxmlPath:String, outPath:String, outBgPath:String, atlasesPath:String) {
		final hxmlName = Path.withoutDirectory(hxmlPath);
		if (!FileSystem.exists(hxmlPath)) {
			console.log('/$hxmlName not found to update');
			return;
		}
		final bgs = FileSystem.readDirectory(outBgPath).filter(name -> {
			if (FileSystem.isDirectory('$outBgPath/$name')) return false;
			return true;
		}).filter(name -> {
			if (name.startsWith("bg")) return true;
			return false;
		});
		final atlases = FileSystem.readDirectory(outPath).filter(name -> {
			if (FileSystem.isDirectory('$outBgPath/$name')) return false;
			return true;
		}).filter(name -> {
			if (name.startsWith("res")) return true;
			return false;
		});
		final hxml = File.getContent(hxmlPath);
		final lines = hxml.split("\n");
		final imagesBlock = "# IMAGES";
		final atlasesBlock = "# ATLASES";
		final endBlock = "# END";
		if (!lines.contains(imagesBlock) || !lines.contains(atlasesBlock)
			|| !lines.contains(endBlock)) {
			console.log("# IMAGES / # ATLASES / # END blocks not found in build-resources.hxml");
			return;
		}
		final imagesBlockId = lines.indexOf(imagesBlock);
		final atlasesBlockId = lines.indexOf(atlasesBlock);
		final endBlockId = lines.indexOf(endBlock);
		if (imagesBlockId >= atlasesBlockId || atlasesBlockId >= endBlockId) {
			console.log('$hxmlName generator blocks should be in order:');
			console.log("# IMAGES ... # ATLASES ... # END");
			return;
		}
		var i = imagesBlockId + 1;
		lines.splice(i, endBlockId - imagesBlockId - 1);
		for (bg in bgs) {
			final name = Path.withoutDirectory(Path.withoutExtension(bg));
			lines.insert(i, '--resource res/$bg@$name');
			i++;
		}
		lines.insert(i, "# ATLASES");
		i++;
		for (atlas in atlases) {
			lines.insert(i, '--resource res/atlases/$atlas');
			i++;
		}
		File.saveContent(hxmlPath, lines.join("\n"));
		final atlases = atlases.filter(s -> s.endsWith(".png")).map(s ->
			Path.withoutExtension(s));
		final folders = getAtlasFolders(atlasesPath).map(s -> {
			return s.length == 0 ? "res" : 'res_$s';
		});
		for (name in folders) {
			atlases.remove(name);
		}
		if (atlases.length > 0) console.log('\nWarning: Custom atlases in $outPath/:');
		for (name in atlases) {
			console.log(name);
		}
	}
}

class PngDimensions {
	static final pngSignature = 'PNG\r\n\x1a\n';
	static final pngImageHeaderChunkName = 'IHDR';
	// Used to detect "fried" png's: http://www.jongware.com/pngdefry.html
	static final pngFriedChunkName = 'CgBI';

	public static function validate(buffer:Buffer) {
		if (pngSignature == buffer.toString('ascii', 1, 8)) {
			var chunkName = buffer.toString('ascii', 12, 16);
			if (chunkName == pngFriedChunkName) {
				chunkName = buffer.toString('ascii', 28, 32);
			}
			if (chunkName != pngImageHeaderChunkName) {
				throw 'Invalid PNG';
			}
			return true;
		}
		return false;
	}

	public static function calculate(buffer:Buffer) {
		if (buffer.toString('ascii', 12, 16) == pngFriedChunkName) {
			return {
				height: buffer.readUInt32BE(36),
				width: buffer.readUInt32BE(32)
			}
		}
		return {
			height: buffer.readUInt32BE(20),
			width: buffer.readUInt32BE(16)
		}
	}
}
