package;

import haxe.Json;
import haxe.crypto.Base64;
import haxe.io.Bytes;
import haxe.io.Path;
import js.Node.process;
import js.node.Buffer;
import js.npm.ws.Server as WSServer;
import js.npm.ws.WebSocket;
import sys.FileSystem;
import sys.io.File;

enum abstract WsEventType(String) {
	var Connected;
	var WriteFile;
	var WriteTextFile;
}

typedef WsEvent = {
	type:WsEventType,
	?connected:{},
	?writeFile:{
		path:String,
		base64:String,
	},
	?writeTextFile:{
		path:String,
		text:String,
	}
}

@:keep
@:expose
class Observer {
	public var logEvents = false;

	final projectDir:String;
	final buildDir:String;
	final scriptName:String;
	final wss:WSServer;
	final clients:Array<WebSocket> = [];

	public function new(projectDir:String, buildDir:String, scriptName:String, port:Int) {
		this.projectDir = projectDir;
		this.buildDir = buildDir;
		this.scriptName = scriptName;
		wss = new WSServer({port: port});
		function exit() {
			trace("exit");
			process.exit();
		}
		process.on("exit", exit);
		process.on("SIGINT", exit); // ctrl+c
		process.on("uncaughtException", (log) -> {
			trace(log);
		});
		process.on("unhandledRejection", (reason, promise) -> {
			trace("Unhandled Rejection at:", reason);
		});
		wss.on("connection", onConnect);
	}

	function onConnect(ws:WebSocket) {
		if (logEvents) trace("Client connected");
		clients.push(ws);

		final reply:WsEvent = {
			type: Connected,
			connected: {}
		};
		ws.send(Json.stringify(reply), null);

		ws.on("close", function(err) {
			if (logEvents) trace("Client disconnected");
			clients.remove(ws);
		});
		ws.on("message", onMessage);
	}

	function onMessage(e):Void {
		final data:WsEvent = Json.parse(e);
		if (logEvents) {
			final t:String = cast data.type;
			final t = t.charAt(0).toLowerCase() + t.substr(1);
			trace('Event: ${data.type}', Reflect.field(data, t));
		}
		switch data.type {
			case Connected:
			case WriteFile:
				final data = data.writeFile;
				var path = data.path;
				if (!Path.isAbsolute(path)) path = Path.join([projectDir, path]);
				File.saveBytes(path, Base64.decode(data.base64));
			case WriteTextFile:
				final data = data.writeTextFile;
				var path = data.path;
				if (!Path.isAbsolute(path)) path = Path.join([projectDir, path]);
				File.saveContent(path, data.text);
		}
	}

	function broadcast(data:WsEvent):Void {
		for (client in clients) client.send(Json.stringify(data), null);
	}
}
