package game;

import khm.Sprite;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.Collision;
import khm.utils.macro.EnumAbstractUtils;

enum abstract CollisionSide(Int) {
	var None;
	var Top;
	var Bottom;
	var Left;
	var Right;

	public function toString():String {
		return EnumAbstractUtils.toString(abstract);
	}
}

typedef CollisionSideByDepth = {
	side:CollisionSide,
	depthX:Float,
	depthY:Float,
}

class Col {
	/** Returns side of `tile` that collided with `rect`. **/
	public static function getTileCollisionSide(rect:Rect, speed:Point, isX:Bool, tile:Tile):CollisionSide {
		final tilePos = tile.obj.topLeft();
		final tileRect:Rect = {
			x: tilePos.x,
			y: tilePos.y,
			w: tile.obj.width,
			h: tile.obj.height
		}
		return getRectCollisionSide(rect, speed, isX, tileRect);
	}

	static final response:CollisionSideByDepth = {
		side: Top,
		depthX: 0,
		depthY: 0,
	};

	public static function getCollisionSideByDepth(rect:Rect, tileRect:Rect):Null<CollisionSideByDepth> {
		if (!Collision.aabb2(rect, tileRect)) return null;
		final center = rect.center();
		final center2 = tileRect.center();
		final deltaX = center.x - center2.x;
		final deltaY = center.y - center2.y;
		final totalHalfW = (rect.w + tileRect.w) / 2;
		final totalHalfH = (rect.h + tileRect.h) / 2;
		final overlapX = totalHalfW - deltaX.abs();
		final overlapY = totalHalfH - deltaY.abs();

		if (overlapX.abs() < overlapY.abs()) {
			response.side = deltaX > 0 ? Right : Left;
			response.depthX = overlapX * deltaX.sign();
			response.depthY = 0;
		} else {
			response.side = deltaY > 0 ? Bottom : Top;
			response.depthX = 0;
			response.depthY = overlapY * deltaY.sign();
		}
		return response;
	}

	public static function getRectCollisionSide(rect:Rect, speed:Point, isX:Bool, tileRect:Rect):CollisionSide {
		if (!Collision.aabb2(rect, tileRect)) return None;

		if (isX) {
			if (speed.x > 0) {
				return Left;
			} else if (speed.x < 0) {
				return Right;
			}
		} else {
			if (speed.y > 0) {
				return Top;
			} else if (speed.y < 0) {
				return Bottom;
			}
		}
		return None;
	}

	static final tilesCache:Array<Tile> = [];

	public static function getOvelappingTiles(rect:Rect, tilemap:Tilemap, layer:Int):Array<Tile> {
		tilesCache.resize(0);
		final startX = Math.floor(rect.x / tilemap.tileW);
		final startY = Math.floor(rect.y / tilemap.tileH);
		final endX = Math.ceil((rect.x + rect.w) / tilemap.tileW);
		final endY = Math.ceil((rect.y + rect.h) / tilemap.tileH);
		for (iy in startY...endY) {
			for (ix in startX...endX) {
				final tile = tilemap.getTile(layer, ix, iy);
				tilesCache.push(tile);
			}
		}
		return tilesCache;
	}
}
