package game;

import khm.Container;
import khm.Ease;
import khm.IGameObject;
import khm.Lazyr;
import khm.MathLib;
import khm.Particles;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.Tween;
import khm.Types.IPoint;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tilemap;
import khm.utils.Collision;

@:structInit
class PlatformConfig {
	public var movePoints:Array<IPoint>;
	public var speed:Float;
	public var sizeW = 1;
	public var sizeH = 1;
	public var ease:Ease = SineInOut;
	public var pathMode:PlatformPathMode = YoyoDir;
	public var isUsedLift = false;
	public var hasPathLine = true;
}

enum PlatformPathMode {
	OneDir;
	YoyoDir;
}

class Platform extends Container {
	public final iid:String;

	final game:Game;
	final tilemap:Tilemap;

	public final movePoints:Array<IPoint>;

	final ease:Ease;
	final pathMode:PlatformPathMode;
	var curPointI = 0;
	var pointRatio = 0.0;
	var isYoyo = false;

	public final speed:Point = {x: 0, y: 0};
	public var speedRatio:Float;
	public var isActive = false;

	var sizeW:Int;
	var sizeH:Int;

	public function new(iid:String, game:Game, tilemap:Tilemap, config:PlatformConfig) {
		super();
		this.iid = iid;
		this.game = game;
		this.tilemap = tilemap;
		movePoints = config.movePoints;
		speedRatio = config.speed / 100;
		sizeW = config.sizeW;
		sizeH = config.sizeH;
		ease = config.ease;
		pathMode = config.pathMode;
		final p = movePoints[0];
		if (movePoints.length < 2 && !config.isUsedLift) {
			trace('no move points for platform on ${p.x},${p.y}');
		}

		final layer = tilemap.map.layers[1];
		for (i => nextP in movePoints) {
			if (!config.hasPathLine) break;
			if (i == 0) continue;
			final p = movePoints[i - 1];
			final lineRectC = layer.addContainer();
			final platform_line = lineRectC.addNineSlice(res.platform_line, 3, 1);
			platform_line.height = 3;
			final dx = p.x - nextP.x;
			final dy = p.y - nextP.y;
			final w = MathLib.dist.betweenPoints(cast p, cast nextP);
			final ang = MathLib.angle.betweenPoints(cast p, cast nextP);
			platform_line.width = w * tilemap.tileW;
			platform_line.angle = ang;
			platform_line.alpha = 0.5;
			lineRectC.x = (0.5.lerp(p.x, nextP.x) + sizeW / 2) * tilemap.tileW;
			lineRectC.y = (0.5.lerp(p.y, nextP.y) + sizeH / 2) * tilemap.tileH;

			lineRectC.width = platform_line.width;
			lineRectC.height = platform_line.width;
			layer.addAt(0, lineRectC);
		}

		this.setOrigin(0);
		setSize(16, 16);
		final layer = tilemap.tileset.layers[1];
		for (iy in 0...sizeH) {
			for (ix in 0...sizeW) {
				final sprite = Sprite.fromTile(res.tileset, layer[4].rect);
				sprite.setOrigin(0);
				sprite.x = ix * tilemap.tileW;
				sprite.y = iy * tilemap.tileH;
				add(sprite);
			}
		}
		fit();
		x = p.x * tilemap.tileW;
		y = p.y * tilemap.tileH;
	}

	public function getTotalDuration():Float {
		return switch pathMode {
			case OneDir:
				1000 / 60 / speedRatio * (movePoints.length - 1);
			case YoyoDir:
				1000 / 60 / speedRatio * 2 * (movePoints.length - 1);
		}
	}

	public function update(objects:Array<{final speed:Point; final obj:IGameObject;}>) {
		if (!isActive) return;
		final curPoint = movePoints[curPointI];
		final nextI = isYoyo ? curPointI - 1 : curPointI + 1;
		final nextPoint = movePoints[nextI];
		final easing = Tween.getEasingCallback(ease);
		pointRatio += speedRatio * game.timeRatio;
		pointRatio = pointRatio.clamp(0, 1);
		final tileW = tilemap.tileW;
		final tileH = tilemap.tileH;
		final sx = (curPoint.x) * tileW;
		final sy = (curPoint.y) * tileH;
		final ex = (nextPoint.x) * tileW;
		final ey = (nextPoint.y) * tileH;
		final newX = easing(pointRatio).lerp(sx, ex);
		final newY = easing(pointRatio).lerp(sy, ey);
		speed.x = newX - x;
		speed.y = newY - y;

		x += speed.x;
		y += speed.y;

		if (pointRatio == 1) {
			pointRatio = 0;
			final isLast = isYoyo ? nextI == 0 : nextI == movePoints.length - 1;
			if (isYoyo) {
				curPointI--;
			} else {
				curPointI++;
			}
			switch pathMode {
				case OneDir:
					if (isLast) {
						speed.x = 0;
						speed.y = 0;
						isActive = false;
						isYoyo = !isYoyo;
						onComplete();
					}
				case YoyoDir:
					if (isLast && !isYoyo) {
						isYoyo = true;
					} else if (isLast && isYoyo) {
						isYoyo = false;
						speed.x = 0;
						speed.y = 0;
						isActive = false;
						onComplete();
					}
			}
		}
		onUpdate();
	}

	dynamic public function onUpdate():Void {}

	dynamic public function onComplete():Void {}
}
