package game;

import game.Anim.FrameConfig;
import game.Col.CollisionSide;
import khm.Container;
import khm.Res;
import khm.Sprite;
import khm.Types.Point;
import khm.Types.Rect;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.Collision;

enum abstract PlayerAnimType(String) {
	var Stand;
	var Run;
	var Brake;
	var Jump;
	var Soar;
	var Fall;
}

class Player {
	final game:Game;
	final tilemap:Tilemap;

	public final obj:Container;

	final playerSprite:Sprite;

	public final speed:Point = {x: 0, y: 0};

	public final accelerationSpeed = 0.2;
	public final maxRunSpeed = 1.5;
	public final maxFallSpeed = 5;
	public var groundSlowdown = 0.15;
	public var airSlowdown = 0.1;
	public var jumpSpeed = 2.8;
	// public var jumpSpeed = 2.2;
	public var gravity = 0.1;

	final coyoteTicks = 8;
	var coyoteCounter = 0;

	public var onGround = false;
	public var isMoving = false;

	final insideTiles:Array<Tile> = [];
	var isLeftDir = false;

	final frameW = 20;
	final frameH = 20;

	public final anim:Anim;

	public var isPushedByPlatform = false;
	public var isCanFly = false;

	public var isLocked(default, null) = false;

	public function new(game:Game, tilemap:Tilemap) {
		this.game = game;
		this.tilemap = tilemap;
		obj = new Container();
		obj.isPixelPerfect = true;
		playerSprite = obj.addSprite(Res.res.player);
		playerSprite.setCrop({
			x: 0,
			y: 0,
			w: frameW,
			h: frameH
		});
		obj.shrink();
		obj.width = 8;
		obj.height = 15;
		playerSprite.y -= (20 - 15) / 2;

		coyoteCounter = coyoteTicks;

		final frames:Array<FrameConfig> = [
			{set: "stand", delay: 9},
			{},
			{},
			{},
			{},
			{set: "run", delay: 5},
			{},
			{},
			{set: "brake"},
			{set: "jump"},
			{set: "soar"},
			{set: "fall"},
			{},
		];
		anim = new Anim(frames);
	}

	public function reset():Void {
		speed.x = 0;
		speed.y = 0;
		onGround = false;
		isMoving = false;
		isLocked = false;
	}

	public function lock() {
		isLocked = true;
	}

	public function unlock() {
		isLocked = false;
	}

	public function update():Void {
		var gravity = gravity;
		var groundSlowdown = groundSlowdown;
		var airSlowdown = airSlowdown;
		final timeRatio = game.timeRatio;
		if (game.inSelectionMode) {
			gravity *= timeRatio;
			groundSlowdown *= timeRatio;
			airSlowdown *= timeRatio;
		}
		if (isLocked) speed.x = 0;
		speed.y += gravity;
		if (onGround && coyoteCounter > 0) {
			coyoteCounter--;
			coyoteCounter = coyoteCounter.limitMin(0);
		}
		if (coyoteCounter == 0) onGround = false;

		final mult = game.timeRatio;
		var steps = Std.int(Math.max(speed.x, speed.y)).abs().limitMin(1);
		// steps = 1;
		for (i in 0...steps * 2) {
			final isX = i % 2 == 0;
			if (isX) obj.x += speed.x / steps * mult;
			else obj.y += speed.y / steps * mult;
			final tl = obj.topLeft();
			final rect:Rect = {
				x: tl.x,
				y: tl.y,
				w: obj.width,
				h: obj.height,
			};

			// check platforms
			for (platform in game.platforms) {
				final platformTopLeft = platform.topLeft();
				final tileRect:Rect = {
					x: platformTopLeft.x,
					y: platformTopLeft.y,
					w: platform.width,
					h: platform.height
				}
				// for diagonal movements down force player to be on platform
				final r2 = rect.copy();
				r2.y += platform.speed.y;
				if (!Collision.aabb2(r2, tileRect)) continue;
				rect.x += platform.speed.x;
				rect.y += platform.speed.y;
				final data = Col.getCollisionSideByDepth(rect, tileRect) ?? continue;
				// trace(data.side, data.depthY, platform.speed.y.abs());
				// if (data.depthX.abs() > (platform.speed.x.abs() + 1)
				// 	|| data.depthY.abs() > (platform.speed.y.abs() + 1)) {
				// 	game.restart();
				// }
				final side = data.side;

				rectCollisionResponse(tileRect, side, rect);
				if (side == Top) {
					speed.y = 0.0001;
					onGround = true;
				}
				obj.x = rect.x + obj.width / 2;
				obj.y = rect.y + obj.height / 2;
			}

			final tl = obj.topLeft();
			final rect:Rect = {
				x: tl.x,
				y: tl.y,
				w: obj.width,
				h: obj.height,
			};

			final tiles = Col.getOvelappingTiles(rect, tilemap, 1);
			// remove tiles from skip collision list if we are not touching them now
			for (tile in insideTiles.reversed()) {
				if (!tiles.contains(tile)) insideTiles.remove(tile);
			}

			for (tile in tiles) {
				final state = tile.getState();
				final type = tile.getType();
				if (!isTileHasCollision(tile)) continue;

				final side = Col.getTileCollisionSide(rect, speed, isX, tile);
				if (side == None) continue;
				if (type == MovingBlock) {
					if (skipMovingBlockCollision(state, side, tile, rect)) continue;
				}

				final oldX = rect.x;
				final oldY = rect.y;
				// trace(side, speed.y);
				tileCollisionResponse(tile, side, rect);
				if ((rect.x - oldX).abs() > 3 || (rect.y - oldY).abs() > 3) {
					game.restart();
					return;
				}

				obj.x = rect.x + obj.width / 2;
				obj.y = rect.y + obj.height / 2;
			}
		}

		if (isMoving) {
			// no slowdowns
		} else {
			if (onGround) {
				if (speed.x.abs() > 0) speed.x -= groundSlowdown * speed.x.sign();
				if (speed.x.abs() < groundSlowdown) speed.x = 0;
			} else {
				if (speed.x.abs() > 0) speed.x -= airSlowdown * speed.x.sign();
				if (speed.x.abs() < airSlowdown) speed.x = 0;
			}
		}

		if (speed.x.abs() > 5) speed.x = 5 * speed.x.sign();
		if (speed.y.abs() > 5) speed.y = 5 * speed.y.sign();

		if (speed.x.abs() <= 0.01) speed.x = 0;
		if (speed.y.abs() <= 0.01) speed.y = 0;

		setAnim();
	}

	function skipMovingBlockCollision(state:TileState, side:CollisionSide, tile:Tile, rect:Rect):Bool {
		final tilePos = tile.obj.topLeft();
		final tileRect:Rect = {
			x: tilePos.x,
			y: tilePos.y,
			w: tile.obj.width,
			h: tile.obj.height
		}
		// allow to jump up when block pulls
		// final depthSide = Col.getCollisionSideByDepth(rect, tileRect)?.side ?? Bottom;
		// trace(depthSide, insideTiles);
		// if (state.moveState == Activation && depthSide != Top) {
		// 	// disable collision for tiles we touched when they are in activation
		// 	if (!insideTiles.contains(tile)) insideTiles.push(tile);
		// 	return true;
		// }

		// allow to jump up when block pulls
		if (state.movingBlock?.moveState == Activation) {
			if (side != Top) {
				if (!insideTiles.contains(tile)) insideTiles.push(tile);
				return true;
			} else {
				final iy:Int = tile.iy;
				final y = rect.center().y;
				final tileTopY = iy * tilemap.tileH;
				final isPlayerOnTop = y < tileTopY + 0.1;
				if (!isPlayerOnTop) {
					if (!insideTiles.contains(tile)) insideTiles.push(tile);
					return true;
				}
			}
		}

		if (insideTiles.contains(tile)) return true;

		// allow to stay a bit on top when block pushes inside
		if (state.movingBlock?.moveState == Deativation) {
			if (side != Top) return true;
			else {
				final iy:Int = tile.iy;
				final y = rect.center().y;
				final tileTopY = iy * tilemap.tileH;
				final isPlayerOnTop = y < tileTopY + 0.1;
				if (!isPlayerOnTop) return true;
			}
		}
		return false;
	}

	function tileCollisionResponse(tile:Tile, side:CollisionSide, rect:Rect):Void {
		final ix = tile.ix;
		final iy = tile.iy;
		switch side {
			case None:
				return;
			case Top:
				final newY = iy * tilemap.tileH - rect.h;
				#if debug
				// if ((rect.y - newY).abs() > 2) trace("tileCollisionResponse err");
				#end
				rect.y = newY;
				speed.y = 0;
				onGround = true;
				coyoteCounter = coyoteTicks;
				if (tile.getType() == JumpPadActive) {
					activateJumpPad(rect);
				}
				if (tile.getType() == ElevatorBlock) {
					game.activateElevator(tile.getState().elevator);
				}
			case Bottom:
				rect.y = iy * tilemap.tileH + tilemap.tileH;
				speed.y = 0;
			case Left:
				rect.x = ix * tilemap.tileW - rect.w;
				speed.x = 0;
			case Right:
				rect.x = ix * tilemap.tileW + tilemap.tileW;
				speed.x = 0;
		}
	}

	public function rectCollisionResponse(tileRect:Rect, side:CollisionSide, rect:Rect):Void {
		switch side {
			case None:
				return;
			case Top:
				final newY = tileRect.y - rect.h;
				#if debug
				if ((rect.y - newY).abs() > 2) trace("rectCollisionResponse err", rect.y - newY);
				// if ((rect.y - newY).abs() > 2) throw tileRect;
				#end
				rect.y = newY;
				if (speed.y > 0) speed.y = 0;
				onGround = true;
				coyoteCounter = coyoteTicks;
			case Bottom:
				rect.y = tileRect.y + tileRect.h;
				if (speed.y < 0) speed.y = 0;
			case Left:
				rect.x = tileRect.x - rect.w;
				if (speed.x > 0) speed.x = 0;
			case Right:
				rect.x = tileRect.x + tileRect.w;
				if (speed.x < 0) speed.x = 0;
		}
	}

	function isTileHasCollision(tile:Tile):Bool {
		if (tile.id == 0) return false;
		final state = tile.getState();
		final props:{collide:Bool} = tilemap.getProps(1, tile.id);
		// final collide = state.collide ?? props?.collide ?? false;
		final collide = props?.collide ?? false;
		if (!collide) return false;
		switch tile.getType() {
			case MovingBlockInactive:
			case MovingBlock:
				if (state.movingBlock.moveState == Inactive) return false;
			case JumpPadInactive:
			case JumpPadActive:
			case PlatformCallerInactive:
			case PlatformCallerActive:
			case ElevatorBlock:
		}
		return true;
	}

	public function jump():Void {
		if (isLocked) return;
		final tl = obj.topLeft();
		final rect = {
			x: tl.x,
			y: tl.y - 2,
			w: obj.width,
			h: obj.height,
		};
		final tiles = Col.getOvelappingTiles(rect, tilemap, 1);
		for (tile in tiles) {
			if (insideTiles.contains(tile)) continue;
			if (isTileHasCollision(tile)) {
				// todo check
				final state = tile.getState();
				final moveState = tile.getTileMoveState();
				if (moveState == null) return;
				if (moveState == Active) return;
			}
		}
		coyoteCounter = 0;
		onGround = false;
		speed.y = 0;
		speed.y -= jumpSpeed;
	}

	public function limitRunSpeed() {
		if (speed.x.abs() > maxRunSpeed) speed.x = maxRunSpeed * speed.x.sign();
	}

	public function limitFallSpeed() {
		if (speed.y.abs() > maxFallSpeed) speed.y = maxFallSpeed * speed.y.sign();
	}

	function setAnim():Void {
		var left = false;
		var right = false;
		var up = false;
		final keys = khm.App.keys;
		if (!game.inSelectionMode) {
			if (keys.get(Left) || keys.get(A)) left = true;
			if (keys.get(Right) || keys.get(D)) right = true;
			if (keys.get(Up) || keys.get(W)) up = true;
			if (left != right) isLeftDir = left;
		}

		if (onGround) {
			if (left == right) {
				if (speed.x != 0) {
					anim.setType("brake");
					if (speed.x < -maxRunSpeed / 2) isLeftDir = true;
					else if (speed.x > maxRunSpeed / 2) isLeftDir = false;
				} else anim.setType("stand");
			} else {
				if (speed.x.abs() > 0) anim.setType("run");
				else anim.setType("stand");
			}
			// if (isPushedByPlatform) anim.setType("brake");
		} else {
			if (speed.y < -0.5) anim.setType("jump");
			else if (speed.y > 0.5) anim.setType("fall");
			else anim.setType("soar");
		}
		final mult = game.inSelectionMode ? 5 : 1;
		anim.playType(mult);
		playerSprite.flipX = isLeftDir;

		playerSprite.setCrop({
			x: anim.frameI * frameW,
			y: 0,
			w: frameW,
			h: frameH
		});
	}

	function activateJumpPad(rect:Rect) {
		onGround = false;
		rect.y--;
		speed.y = -4.3;
		final particles = tilemap.addParticles({
			frames: [Res.res.pixel_particle],
			lifespan: 500,
			speed: {min: 10, max: 25},
			angle: {min: -180 + 10, max: -180 + 15},
			gravityY: 20,
			alpha: {start: 0.5, end: 0},
		});
		particles.explode(10, rect.x + rect.w / 2, rect.y + rect.h);
		final particles = tilemap.addParticles({
			frames: [Res.res.pixel_particle],
			lifespan: 500,
			speed: {min: 10, max: 25},
			angle: {min: 0 - 10, max: 0 - 15},
			gravityY: 20,
			alpha: {start: 0.5, end: 0},
		});
		particles.explode(10, rect.x + rect.w / 2, rect.y + rect.h);
	}
}
