package game;

import game.Anim.FrameConfig;
import game.TileState.TileMoveState;
import khm.App;
import khm.Color;
import khm.Container;
import khm.Res.res;
import khm.Res;
import khm.Sprite;
import khm.tilemap.Tile;
import khm.tilemap.Tilemap;
import khm.utils.DebugPanel;

class Selector {
	final game:Game;
	final tilemap:Tilemap;

	public final timeline:Array<{tile:Tile, delay:Float}> = [];

	var lastTime = 0.0;
	var outlineAnim:Anim;
	final selectionOutlineFrames:Array<FrameConfig>;

	public var isPlaying = false;
	public var timelineLimit = 3;

	public function new(game:Game, tilemap:Tilemap) {
		this.game = game;
		this.tilemap = tilemap;

		selectionOutlineFrames = [
			{
				set: "show",
				delay: 5,
				loop: false,
				atEnd: "topLoop"
			},
			{},
			{},
			{set: "topLoop", loop: true, delay: 3},
			{},
			{},
			{},
			{},
			{},
			{},
			{},
		];
	}

	public function addTile(tile:Tile):Void {
		final isEmpty = timeline.length == 0;
		final stamp = App.stamp();
		final delay = isEmpty ? 0 : stamp - lastTime;
		if (timeline.exists(obj -> obj.tile == tile)) return;
		timeline.push({
			tile: tile,
			delay: delay
		});
		lastTime = stamp;

		final state = tile.getState();
		final outlineC = state.selectionOutline ?? createTileOutlineContainer(tile);
		state.selectionOutline = outlineC;
		DebugPanel.ignoreObject(outlineC);
		outlineC.x = tile.obj.x;
		outlineC.y = tile.obj.y;

		final outline = outlineC.findSpriteByFrame(res.selection_outline);
		outline.alpha = 0;
		final selectionColor:Color = switch tile.getType() {
			case MovingBlockInactive, MovingBlock:
				tile.getState().movingBlock.initialState == Active ? 0xFF888888 : White;
			case JumpPadInactive, JumpPadActive:
				0x880066FF;
			case PlatformCallerInactive:
				0x8800FFCC;
			case PlatformCallerActive:
				0xFF888888;
			case _: White;
		}
		final alpha = selectionColor != White ? 1 : 0.8;
		outline.tint = selectionColor;
		outline.opacityAnim(alpha, 300);

		if (timeline.length == timelineLimit) {
			game.delayed(200, () -> {
				game.exitSelection();
			});
		}
	}

	function createTileOutlineContainer(tile:Tile) {
		final outlineC = tilemap.map.layers[1].addContainer();
		final layer = tilemap.tileset.layers[1];
		// final outline = Sprite.fromTile(res.tileset, layer[42].rect);
		final outline = outlineC.addSprite(res.selection_outline);
		outline.state.name = "outline";
		outlineC.add(outline);
		outlineC.shrink();
		outlineC.state.outline = outline;
		final topOutlineAnim = outlineC.addSprite(res.selection_anim);
		final bottomOutlineAnim = outlineC.addSprite(res.selection_anim);
		bottomOutlineAnim.angle = 180;

		topOutlineAnim.alpha = 0;
		bottomOutlineAnim.alpha = 0;
		topOutlineAnim.showAnim(300);
		bottomOutlineAnim.showAnim(300);

		final outlineAnim = new Anim(selectionOutlineFrames);
		outlineAnim.setType("show");
		outlineC.on(Update, e -> {
			outlineAnim.playType();
			topOutlineAnim.setCrop({
				x: outlineAnim.frameI * 16,
				y: 0,
				w: 16,
				h: 16
			});
			bottomOutlineAnim.setCrop({
				x: outlineAnim.frameI * 16,
				y: 0,
				w: 16,
				h: 16
			});
		});
		return outlineC;
	}

	public function toggleTile(tile:Tile):Void {
		final state = tile.getState();
		final moveState = tile.getTileMoveState();
		if (moveState == Activation || moveState == Deativation) {
			return; // maybe do push/pull anim cancelations later
		}
		switch moveState {
			case Active, Activation:
				final delay = deactivateTile(tile);
				game.delayedTime(delay, () -> {
					final initialState = tile.getTileInitialState();
					switch initialState {
						case Active:
							activateTile(tile);
						case Inactive:
							deactivateTile(tile);
						case _:
					}
				});
			case Deativation, Inactive:
				final delay = activateTile(tile);
				game.delayedTime(delay, () -> {
					deactivateTile(tile);
				});
		}
	}

	public function reset():Void {
		timeline.resize(0);
		isPlaying = false;
	}

	public function playTimeline(?onComplete:() -> Void):Void {
		isPlaying = true;
		final initialDelay = 0.0;
		var delay = initialDelay;
		for (obj in timeline) {
			final tile = obj.tile;
			final outlineC = tile.getState().selectionOutline;
			outlineC.hideAnim(300, () -> {
				outlineC.off();
				tile.getState().selectionOutline = null;
			});
		}
		for (obj in timeline) {
			delay += obj.delay;
			game.delayed(delay, () -> {
				toggleTile(obj.tile);
			});
		}
		timeline.resize(0);
		game.delayed(delay + 3000, () -> {
			isPlaying = false;
			if (onComplete != null) onComplete();
		});
	}

	public function activateTile(tile:Tile, instant = false):Float {
		var deactivationDelay = 3000.0;
		switch tile.getType() {
			case MovingBlockInactive:
			case MovingBlock:
				tile.getState().movingBlock.activate(instant);
			case JumpPadInactive:
				tile.getState().jumpPad.activate(instant);
			case JumpPadActive:
				tile.getState().jumpPad.deactivate(instant);
			case PlatformCallerInactive:
				final tileButton = tile.getState().tileButton;
				deactivationDelay = tileButton.activate(instant);
			case PlatformCallerActive:
			case ElevatorBlock:
		}
		return deactivationDelay;
	}

	public function deactivateTile(tile:Tile, instant = false):Float {
		var deactivationDelay = 3000.0;
		switch tile.getType() {
			case MovingBlockInactive:
			case MovingBlock:
				tile.getState().movingBlock.deactivate(instant);
			case JumpPadInactive:
				tile.getState().jumpPad.activate(instant);
			case JumpPadActive:
				tile.getState().jumpPad.deactivate(instant);
			case PlatformCallerInactive:
			case PlatformCallerActive:
				final tileButton = tile.getState().tileButton;
				deactivationDelay = tileButton.deactivate(instant);
			case ElevatorBlock:
		}
		return deactivationDelay;
	}

	public function update():Void {
		if (timeline.length > 0) {
			final time = App.stamp() - lastTime;
			if (time > 10 * 1000) {
				game.exitSelectionWithError();
			}
		}
	}
}
